/**
 * Dns.h
 *
 * Copyright 2002, 2004 by Anthony Howe. All rights reserved.
 */

#ifndef __com_snert_lib_io_Dns_h__
#define __com_snert_lib_io_Dns_h__	1

#ifndef __com_snert_lib_util_Vector_h__
# include <com/snert/lib/util/Vector.h>
#endif

/*
 * Provide IPPROTO_* definitions.
 */
#if defined(__WIN32__)
# if defined(__VISUALC__)
#  define _WINSOCK2API_
# endif
# include <windows.h>
# include <winsock2.h>
# define ETIMEDOUT	WSAETIMEDOUT
#else
# include <sys/types.h>
# include <netinet/in.h>
#endif /* __WIN32__ */

#ifdef __cplusplus
extern "C" {
#endif

typedef void *Dns;

/*
 * (Re)Load the /etc/resolv.conf file. Currently only nameserver lines
 * are recognised. This function should be called before the first
 * DnsOpen() call.
 *
 * @return
 *	Zero (0) on success otherwise -1 one error.
 */
extern int DnsInit(void);

/*
 * @param servers
 *	A NULL terminate array of C string pointers. Each C string
 * 	is the IP address of a DNS server to consult in order of
 *	preference. This argument can be NULL, in which case the
 *	list of DNS servers from /etc/resolv.conf are consulted.
 *
 * @return 
 *	A Dns object returned by DnsOpen().
 */
extern Dns DnsOpen(const char **servers);

/*
 * @param dns
 *	A Dns object returned by DnsOpen().
 *
 * @param n
 *	Maximum number of times to cycle through the DNS server
 *	list before giving up on a search. Default is 4.
 */
extern void DnsSetAttempts(Dns dns, int n);

/*
 * @param dns
 *	A Dns object returned by DnsOpen().
 *
 * @param ms
 *	The initial timeout in milliseconds to wait for a response
 *	from a DNS server. An exponential backup algorithm is used.
 *	The default is 5000 ms.
 *
 * Using the defaults and assuming a single DNS server, then the
 * exponential backoff algorithm will result in a max. delay of
 * 75 seconds (5+10+20+40) to find an answer before giving up.
 *
 * With more than one server, the timeout each round is divided
 * by the number of servers available. So for example:
 *
 *	2 servers: 5+5+ 5+5+ 10+10+ 20+20        = 80 seconds
 *	3 servers: 5+5+5+ 5+5+5+ 6+6+6+ 13+13+13 = 87 seconds
 */
extern void DnsSetTimeout(Dns dns, long ms);

/*
 * @param dns
 *	A Dns object returned by DnsOpen() to close.
 */
extern void DnsClose(Dns dns);

/*
 * @param dns
 *	A Dns object returned by DnsOpen().
 *
 * @return
 *	A DNS_RCODE_ value. The value DNS_RCODE_NULL and DNS_RCODE_ERRNO
 *	are internal error conditions not returned from a DNS server.
 */
extern int DnsGetReturnCode(Dns dns);

#define DNS_RCODE_NULL			(-2)
#define DNS_RCODE_ERRNO			(-1)
#define DNS_RCODE_OK			0
#define DNS_RCODE_FORMAT		1
#define DNS_RCODE_SERVER		2
#define DNS_RCODE_UNDEFINED		3
#define DNS_RCODE_NOT_IMPLEMENTED	4
#define DNS_RCODE_REFUSED		5

/*
 * @param dns
 *	A Dns object returned by DnsOpen().
 *
 * @return
 *	A C string error message for the corresponding DNS_RCODE_ value.
 */
extern const char *DnsGetError(Dns dns);

/*
 * @param dns
 *	A Dns object returned by DnsOpen().
 *
 * @param type
 *	The DNS resource record type to search for. See DNS_TYPE_*
 *
 * @param name
 *	A domain or host name to search for. In the case of a PTR
 * 	search, its the IP address.
 *
 * @return
 *	A Vector of DnsEntry pointers on success, otherwise a NULL
 *	pointer on error. The Vector may be empty. When finished
 *	with the Vector, pass it to VectorDestroy() to clean up.
 */
extern Vector DnsGet(Dns dns, int type, const char *name);

#define DNS_TYPE_A			1
#define DNS_TYPE_NS			2
#define DNS_TYPE_CNAME			5
#define DNS_TYPE_PTR			12
#define DNS_TYPE_MX			15
#define DNS_TYPE_TXT			16

typedef struct DnsEntry {
	char *name;
	char *value;
	unsigned long ttl;	
	unsigned short type;
	unsigned short preference;
	struct in_addr ip;
} *DnsEntry;

/*
 * @param mask
 *	Set the debug detail according to the bit-wise OR of
 * 	DNS_DEBUG_* values.
 */
extern void DnsSetDebugMask(int mask);

#define DNS_DEBUG_ALL		(~0)
#define DNS_DEBUG_ENTER_EXIT	1
#define DNS_DEBUG_GET_SET	2
#define DNS_DEBUG_TRACE		8

#ifdef  __cplusplus
}
#endif

#endif /* __com_snert_lib_io_Dns_h__ */
