/*
 * MessageStream.h
 *
 * Copyright 2001, 2004 by Anthony Howe. All rights reserved.
 */

#ifndef __com_snert_lib_io_MessageStream_h__
#define __com_snert_lib_io_MessageStream_h__	1

#ifndef __com_snert_lib_io_Message_h__
#include <com/snert/lib/io/Message.h>
#endif

typedef void *MessageStream;

#define MESSAGE_TIMEOUT		((Message) -1)

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Prepare and load any required shared libraries.
 *
 * @return
 *	Zero (0) on success, otherwise -1 on error.
 */
extern int MessageStreamInit(void);

/**
 * <p>
 * Open a bi-direction I/O MessageStream (read, write, both) to
 * a file or socket specified by one of the following supported
 * URI forms:
 * </p>
 * <pre>
 *	file:/absolute/path/to/file
 *	file:relative/path/to/file
 *	connect://host:port		Connect to the remote server.
 *	listen://host:port		Wait for connections from a client.
 * </pre>
 * <p>
 * In the case of the listen URI, the host should be either
 * the local loopback or the prefered IP to listen on in the
 * case of a multihomed host.
 * </p>
 * <p>
 * As an optimisation, C strings can be passed instead of Message
 * objects. This reduces the number of memory allocations, copy, and
 * frees required when the message is always a C string.
 * </p>
 * <pre>
 *	file-text:/absolute/path/to/file
 *	file-text:relative/path/to/file
 *	connect-text://host:port	Connect to the remote server.
 * </pre>
 *
 * @return
 * A MessageStream object.
 */
extern MessageStream MessageStreamOpen(const char *);

/**
 * <p>
 * Close all resources associated with a MessageStream.
 * Once closed, MessageStream cannot be reused.
 * </p>
 *
 * @return
 * Zero (0) on success, otherwise -1 on error.
 */
extern int MessageStreamClose(MessageStream);

/**
 * <p>
 * Read a message from the MessageStream.
 * </p>
 *
 * @return
 * A Message read from the MessageStream on success, MESSAGE_TIMEOUT
 * no message received before the read timer expired,otherwise NULL
 * on error.
 */
extern Message MessageStreamRead(MessageStream);

/**
 * <p>
 * Write a Message to the MessageStream. A C string can be used
 * instead of a Message if the MessageStream was opened with
 * a file-text or connect-text URI.
 * </p>
 *
 * @return
 * Zero (0) on success, otherwise the value of errno.
 */
extern int MessageStreamWrite(MessageStream, void *);

/**
 * <p>
 * Get the current read() timeout in milliseconds.
 * </p>
 *
 * @return
 * The current timeout value in milliseconds.
 */
extern long MessageStreamGetTimeout(MessageStream);

/**
 * <p>
 * Set the read() timeout, in milliseconds, to be used. A value
 * of zero (0) specifies an infinite timeout (blocking I/O).
 * A value greater than zero (0) specifies a how long a read()
 * should wait for input before returning.
 * </p>
 */
extern void MessageStreamSetTimeout(MessageStream, long);

#ifdef  __cplusplus
}
#endif

#endif /* __com_snert_lib_io_MessageStream_h__ */
