/*
 * smf.h
 *
 * Copyright 2002, 2004 by Anthony Howe.  All rights reserved.
 */

#ifndef __com_snert_lib_mail_smf_h__
#define __com_snert_lib_mail_smf_h__	1

#ifdef __cplusplus
extern "C" {
#endif

#include <stdarg.h>
#include <libmilter/mfapi.h>
#include <com/snert/lib/sys/Mutex.h>
#include <com/snert/lib/mail/parsePath.h>

/*@-exportlocal@*/

/***********************************************************************
 *** Constants
 ***********************************************************************/

/*
 * These are things I like to log in assorted milters and felt should be
 * standardised across all my milters. 
 */
#define SMF_LOG_ALL			(~0)
#define SMF_LOG_INFO			1	/* default, general */
#define SMF_LOG_TRACE			2	/* funtions in/out */
#define SMF_LOG_PARSE			4	/* slice & dice */
#define SMF_LOG_DEBUG			8	/* funtions debug */
#define SMF_LOG_DIALOG			16	/* communications */
#define SMF_LOG_STATE			32	/* state transitions */
#define SMF_LOG_DNS			64	/* debug DNS code */
#define SMF_LOG_CACHE			128	/* cache get/put/gc */
#define SMF_LOG_DATABASE		256	/* SMDB lookups */
#define SMF_LOG_SOCKET_FD		512	/* socket open/close */
#define SMF_LOG_SOCKET_ALL		1024	/* socket functions & I/O */
#define SMF_LOG_LIBMILTER		2048	/* libmilter engine */

#ifndef SMF_SOCKET_TIMEOUT
#define SMF_SOCKET_TIMEOUT		1800
#endif

#ifndef MILTER_CHUNK_SIZE
#define MILTER_CHUNK_SIZE		65535
#endif

#define X_SMFIS_UNKNOWN			(-1)

#define PRECEDENCE_SPECIAL_DELIVERY	4
#define PRECEDENCE_FIRST_CLASS		3
#define PRECEDENCE_LIST			2
#define PRECEDENCE_JUNK			1
#define PRECEDENCE_BULK			0

/* See RFC 2821 section 4.5.3.1 Size limits and minimums */
#define SMTP_LINE_SIZE			1000
#define SMTP_REPLY_SIZE			512
#define DOMAIN_LENGTH			255
#define LOCAL_PART_LENGTH		128
#define EMAIL_LENGTH			(LOCAL_PART_LENGTH+1+DOMAIN_LENGTH)
#define IPV6_LENGTH			((128/16)*4+(128/16)-1)

#define SMTP_OK				250
#define SMTP_IS_OK(x)			(200 <= (x) && (x) < 300)
#define SMTP_IS_DEFER(x)		(300 <= (x) && (x) < 400)
#define SMTP_IS_TEMP(x)			(400 <= (x) && (x) < 500)
#define SMTP_IS_PERM(x)			(500 <= (x) && (x) < 600)
#define SMTP_IS_VALID(x)		(200 <= (x) && (x) < 600)

#define SMF_STDIO_CLOSE			0
#define SMF_STDIO_AS_IS			1
#define SMF_STDIO_IGNORE		2

/***********************************************************************
 *** Global Variables
 ***********************************************************************/

typedef struct {
	const char *name;
	const char *version;
	const char *copyright;
	const char *user;
	const char *group;
	const char *stateDir;
	const char *socket;
	int standardIO;				/* SMF_STDIO_ constant */
} smfInfo;

typedef struct {
	SMFICTX *ctx;				/* per connection */
	unsigned short cid;			/* per connection id */
	int skipConnection;			/* per connection */
	int skipMessage;			/* per message */
	const char *qid;			/* per message, $i macro */	
	ParsePath *mail;			/* per message, must free */
	ParsePath *rcpt;			/* per recipient, must free */
	char replyLine[SMTP_REPLY_SIZE+1];	/* smfReply(), smfReplyV() */	
} smfWork;

extern long smfLogDetail;
extern pthread_mutex_t smfMutex;

extern const char smfNoQueue[];
extern const char smfUndefined[];
extern const char * const smfPrecedence[];

extern /* const */ char smMacro_auth_authen[];
extern /* const */ char smMacro_client_addr[];
extern /* const */ char smMacro_client_name[];
extern /* const */ char smMacro_client_resolv[];
extern /* const */ char smMacro_if_addr[];
extern /* const */ char smMacro_if_name[];
extern /* const */ char smMacro_verify[];

/***********************************************************************
 *** Functions
 ***********************************************************************/

extern void smfSetLogDetail(const char *detail);
extern sfsistat smfNullWorkspaceError(const char *where);
extern sfsistat smfReply(smfWork *work, int code, const char *ecode, const char *fmt, ...);
extern sfsistat smfReplyV(smfWork *work, int code, const char *ecode, const char *fmt, va_list args);

extern unsigned short smfOpenProlog(SMFICTX *ctx, char *client_name, _SOCK_ADDR *raw_client_addr, char *client_addr, long length);
extern unsigned short smfCloseEpilog(smfWork *work);

extern int smfAccessHost(smfWork *work, const char *tag, const char *client_name, const char *client_addr, int loopDefault);
extern int smfAccessMail(smfWork *work, const char *tag, const char *mail, int dsnDefault);
extern int smfAccessRcpt(smfWork *work, const char *tag, const char *rcpt);

extern int smfHeaderSet(SMFICTX *ctx, char *field, char *value, int index, int present);

extern void smfAtExitCleanUp(void);
extern void smfSignalExit(int signum);
extern int smfSetProcessOwner(smfInfo *smf);
extern int smfMainStart(smfInfo *smf, struct smfiDesc smfDesc);

/*@=exportlocal@*/

#ifdef  __cplusplus
}
#endif

#endif /* __com_snert_lib_mail_smf_h__ */

