/**
 * Dns.c
 *
 * RFC 1035, 2821
 *
 * Copyright 2002, 2004 by Anthony Howe. All rights reserved.
 */

#undef NDEBUG

#ifndef DNS_DEFAULT_TIMEOUT
#define DNS_DEFAULT_TIMEOUT	5000
#endif

#ifndef DNS_DEFAULT_ATTEMPTS
#define DNS_DEFAULT_ATTEMPTS	4
#endif

#ifndef RESOLV_CONF
#define RESOLV_CONF		"/etc/resolv.conf"
#endif

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>

#if defined(__unix__) && ! defined(__CYGWIN__)
#  include <syslog.h>
#else
#  include <com/snert/lib/io/Log.h>
#endif

#include <com/snert/lib/io/Dns.h>
#include <com/snert/lib/io/posix.h>
#include <com/snert/lib/io/Socket.h>
extern char *ip4ToString(unsigned long);

#include <com/snert/lib/mail/MailSpan.h>
#include <com/snert/lib/util/Text.h>

/***********************************************************************
 *** Constants
 ***********************************************************************/

#define DNS_PORT		53
#define UDP_PACKET_SIZE		512

#define TYPE_A			1
#define TYPE_NS			2
#define TYPE_CNAME		5
#define TYPE_SOA		6
#define TYPE_WKS		11	/* not supported */
#define TYPE_PTR		12
#define TYPE_HINFO		13	/* not supported */
#define TYPE_MINFO		14	/* not supported */
#define TYPE_MX			15
#define TYPE_TXT		16
#define TYPE_ALL		255

#define CLASS_IN		1
#define CLASS_CS		2
#define CLASS_CH		3
#define CLASS_HS		4

#define OP_QUERY		0x0000
#define OP_IQUERY		0x0800
#define OP_STATUS		0x1000

#define BITS_QR			0x8000	/* Query = 0, response = 1 */
#define BITS_OP			0x7800	/* op-code */
#define BITS_AA			0x0400	/* Response is authoritative. */
#define BITS_TC			0x0200	/* Message was truncated. */
#define BITS_RD			0x0100	/* Recursive query desired. */
#define BITS_RA			0x0080	/* Recursion available from server. */
#define BITS_Z			0x0070	/* Reserved - always zero. */
#define BITS_AU			0x0020	/* Answer authenticaed */
#define BITS_RCODE		0x000f	/* Response code */

#define SHIFT_QR		15
#define SHIFT_OP		11
#define SHIFT_AA		10
#define SHIFT_TC		9
#define SHIFT_RD		8
#define SHIFT_RA		7
#define SHIFT_Z			4
#define SHIFT_RCODE		0

#ifdef RCODE_ERRNO
#define RCODE_ERRNO		-1
#define RCODE_OK		0
#define RCODE_FORMAT		1
#define RCODE_SERVER		2
#define RCODE_UNDEFINED		3
#define RCODE_NOT_IMPLEMENTED	4
#define RCODE_REFUSED		5
#endif

#define DOMAIN_LENGTH		255

/***********************************************************************
 *** Internal types.
 ***********************************************************************/

struct header {
	unsigned short id;
	unsigned short bits;
	unsigned short qdcount;
	unsigned short ancount;
	unsigned short nscount;
	unsigned short arcount;
};

struct query {
	int length;
	struct header header;
	unsigned char data[UDP_PACKET_SIZE - sizeof (struct header)];
};

struct dns {
	int rcode;
	int attempts;
	long timeout;
	int nservers;
	InetAddress *servers;
	Socket socket;
	const char *error;
	struct query packet;
	unsigned short counter;
};

static long debug = 0;
static Vector nameServers;

/***********************************************************************
 *** Network value macros.
 ***********************************************************************/

#define UCHAR_BIT		((unsigned) CHAR_BIT)

/* These macros intended to retrieve network numeric data types stored
 * at odd memory addresses, which can cause some bus errors on certain
 * CPU types if the pointer is cast to a particular type.
 */
#define GET_NET_SHORT(v, p)	v = (unsigned short) (                     \
				    ((unsigned char *)(p))[0] << UCHAR_BIT   \
				  | ((unsigned char *)(p))[1]               \
				)

#define GET_NET_LONG(v, p)	v = (unsigned long) (                      \
				    ((unsigned char *)(p))[0] << UCHAR_BIT*3 \
				  | ((unsigned char *)(p))[1] << UCHAR_BIT*2 \
				  | ((unsigned char *)(p))[2] << UCHAR_BIT   \
				  | ((unsigned char *)(p))[3]               \
				)

#define GET_NET_SHORT_ADV(v, p)	v  = (unsigned short) (*((unsigned char *)(p))++ << UCHAR_BIT);   \
				v |= (unsigned short) (*((unsigned char *)(p))++)

#define GET_NET_LONG_ADV(v, p)	v  = (unsigned long) (*((unsigned char *)(p))++ << UCHAR_BIT*3); \
				v |= (unsigned long) (*((unsigned char *)(p))++ << UCHAR_BIT*2); \
				v |= (unsigned long) (*((unsigned char *)(p))++ << UCHAR_BIT);   \
				v |= (unsigned long) (*((unsigned char *)(p))++)

#define NET_SHORT_ADV(p)	(p) += 2

#define NET_LONG_ADV(p)		(p) += 4


/***********************************************************************
 *** Query handling.
 ***********************************************************************/

static void
DnsSetError(Dns dns, int rcode, const char *error)
{
	if (dns == (Dns) 0) {
		errno = EFAULT;
		return "null Dns object pointer";
	}

#ifndef NDEBUG
	if (debug & DNS_DEBUG_GET_SET)
		syslog(LOG_DEBUG, "DnsSetError(%lx, %d, '%s')", (long) dns, rcode, error);
#endif
	((struct dns *) dns)->rcode = rcode;
	((struct dns *) dns)->error = error;
}

/*@null@*/ 
static const char * 
buildQuery(struct dns *dns, const char *name, int type, int opcode)
{
	struct query *q;
	int length, needRoot;
	unsigned char *s, *t, *label;

#ifndef NDEBUG
	if ((debug & DNS_DEBUG_TRACE) == DNS_DEBUG_TRACE)
		syslog(LOG_DEBUG, "buildQuery(%lx, \"%s\", %d, %d)", (long) dns, name, type, opcode);
#endif
	q = &dns->packet;
	length = (int) strlen(name);
	needRoot = 0 < length && name[length-1] != '.' && name[length] == '\0';

	if (DOMAIN_LENGTH < length)
		return "DNS name length exceeds 255 octets";

	/* The header length, length of the name with root segment,
	 * and space for the type and class fields.
	 */
	q->length = sizeof (struct header) + 1 + length + needRoot + 4;

	/* Fill in the header fields that are not zero. */
	q->header.id = htons(++dns->counter);

	/* ASSUMPTION: the DNS servers support recursion. */
	q->header.bits = htons((unsigned short)((opcode & BITS_OP) | BITS_RD));

/* Without recursion I would have to perform alot of queries: first query
 * local DNS, which may return only the NS, then query authorative NS, which
 * may or may not support recursion and theory more queries for other info.
 */
/*	q->header.bits = htons((unsigned short)((opcode & BITS_OP))); */

	q->header.qdcount = htons(1);
	q->header.ancount = 0;
	q->header.nscount = 0;
	q->header.arcount = 0;

	/* Copy labels into question. */
	label = q->data;
	t = label + 1;
	for (s = (unsigned char *) name; *s != '\0' ; ++s, ++t) {
		if (*s == '.') {
			/* A label cannot exceed 63 octets in length. */
			if (63 < t - label - 1)
				return "DNS name segment exceeds 63 octets";

			/* Set label length. */
			*label = (unsigned char)(t - label - 1);
			label = t;
		} else {
			*t = *s;
		}
	}

	/* A label cannot exceed 63 octets in length. */
	if (63 < t - label - 1)
		return "DNS name segment exceeds 63 octets";

	/* Set last label length. */
	*label = (unsigned char)(t - label - 1);
	label = t;

	if (needRoot == 1) {
		/* Set root label length. */
		*label++ = 0;
	}

	/* Complete question, assume the class is the Internet. */
	*label++ = 0;
	*label++ = (unsigned char) type;
	*label++ = 0;
	*label = CLASS_IN;

	return (const char *) 0;
}

static int
sendQuery(struct dns *dns)
{
	int attempt;
	long length, timeout, actualTimeout;
	InetAddress *thisServer, *lastServer;

#ifndef NDEBUG
	if ((debug & DNS_DEBUG_TRACE) == DNS_DEBUG_TRACE)
		syslog(LOG_DEBUG, "sendQuery(dns=%lx)", (long) dns);
#endif

	timeout = dns->timeout;
	thisServer = dns->servers;
	lastServer = &dns->servers[dns->nservers];
	
	/* Using the defaults and assuming a single DNS server, then the
	 * exponential backoff algorithm will result in a max. delay of
	 * 75 seconds (5+10+20+40) to find an answer before giving up.
	 *
	 * With more than one server, the timeout each round is divided
	 * by the number of servers available. So for example:
	 *
	 *	2 servers: 5+5+ 5+5+ 10+10+ 20+20 = 80 seconds
	 *	3 servers: 5+5+5+ 5+5+5+ 6+6+6+ 13+13+13 = 87 seconds
	 */
	for (attempt = 0; attempt < dns->attempts; ++attempt) {
#ifndef NDEBUG
		if ((debug & DNS_DEBUG_TRACE) == DNS_DEBUG_TRACE) {
			char *address = thisServer->getAddress(thisServer);
			syslog(LOG_DEBUG, "sendQuery() attempt %d to %s", attempt, address);
			free(address);
		}
#endif
		length = SocketWriteIp(dns->socket, &dns->packet.header, dns->packet.length, thisServer);

#if defined(__FreeBSD__)
/* sendto() is not defined to return EINVAL under FreeBSD, but under Linux yes.
 * See known issue http://www.freebsd.org/cgi/query-pr.cgi?pr=26506
 */
		if (SocketGetError(dns->socket) == EINVAL)
			continue;
#endif

		if (length != dns->packet.length) {
			DnsSetError(dns, RCODE_ERRNO, "socket write error to DNS");
			return -1;
		}

		/* The base timeout value is divided by the number of DNS
		 * servers avaliable in order to limit the actual amount
		 * of time spent looking for an answer. Maintain a mimimum
		 * timeout value.
		 */
		if ((actualTimeout = timeout/dns->nservers) < DNS_DEFAULT_TIMEOUT)
			actualTimeout= DNS_DEFAULT_TIMEOUT;

		if (SocketWaitForInput(dns->socket, actualTimeout) == 0) {
			/*** Consider checking rcode here for RCODE_SERVER,
			 *** in which case continue loop.
			 ***/
			break;
		}

		/* Cycle through list of DNS servers. */
		if (++thisServer == lastServer) {
			/* Reset cycle to the first server. */
			thisServer = dns->servers;

			/* Double the server timeout value this cycle. */
			timeout += timeout;
		}
	}

	if (dns->attempts <= attempt) {
		DnsSetError(dns, RCODE_ERRNO, "no response from DNS server");
		return -1;
	}

	memset(&dns->packet.header, 0, UDP_PACKET_SIZE);
	dns->packet.length = (int) SocketRead(dns->socket, &dns->packet.header, UDP_PACKET_SIZE);
	if (dns->packet.length < 0) {
		DnsSetError(dns, RCODE_ERRNO, "socket read error from DNS");
		return -1;
	}

	dns->packet.header.id = ntohs(dns->packet.header.id);
	dns->packet.header.bits = ntohs(dns->packet.header.bits);

	if (dns->packet.header.id != dns->counter) {
		DnsSetError(dns, RCODE_FORMAT, "DNS request & response ID mismatch");
		return -1;
	}

	/* TODO: check if the message was truncated and redo the request
	 * over TCP as recommend by RFC 974.
	 */

	switch (dns->packet.header.bits & BITS_RCODE) {
	case RCODE_OK:
		DnsSetError(dns, RCODE_OK, "");
		break;
	case RCODE_FORMAT:
		DnsSetError(dns, RCODE_FORMAT, "DNS request format error");
		return -1;
	case RCODE_SERVER:
		DnsSetError(dns, RCODE_SERVER, "DNS server failure");
		return -1;
	case RCODE_UNDEFINED:
		DnsSetError(dns, RCODE_UNDEFINED, "DNS name not found");
		return -1;
	case RCODE_NOT_IMPLEMENTED:
		DnsSetError(dns, RCODE_NOT_IMPLEMENTED, "DNS operation not implemented");
		return -1;
	case RCODE_REFUSED:
		DnsSetError(dns, RCODE_REFUSED, "DNS request refused");
		return -1;
	default:
		DnsSetError(dns, RCODE_SERVER, "DNS server reported an unknown error");
		return -1;
	}

	dns->packet.header.qdcount = ntohs(dns->packet.header.qdcount);
	dns->packet.header.ancount = ntohs(dns->packet.header.ancount);
	dns->packet.header.nscount = ntohs(dns->packet.header.nscount);
	dns->packet.header.arcount = ntohs(dns->packet.header.arcount);

#ifdef IGNORE_OTHER_PARTS_OF_RESPONSE
	if (dns->packet.header.ancount <= 0) {
		DnsSetError(dns, RCODE_ERRNO, "no answers");
		return -1;
	}
#endif

	return 0;
}

/***********************************************************************
 *** Response Record Name Routines
 ***********************************************************************/

/*
 * Copy the string of the resource record name refered to by ptr
 * within the message into a buffer of length bytes. The name
 * labels are separated by dots and the final root label is added.
 * The terminating null character is appended.
 */
static void
copyName(unsigned char *message, unsigned char *ptr, /*@out@*/ unsigned char *buf, long length)
{
	for ( ; *ptr != 0; ptr += *ptr + 1) {
		if ((*ptr & 0xc0) == 0xc0) {
			unsigned short offset;
			GET_NET_SHORT(offset, ptr);
			ptr = &message[offset & 0x3fff];
		}

		/* Do we still have room in the buffer for the next label. */
		if (length <= *ptr)
			break;

		/* Copy label and advance pointers. */
		(void) strncpy((char *) buf, (char *)(ptr+1), (size_t) *ptr);
		length -= *ptr;
		buf += *ptr;
		*buf++ = '.';
	}

	if (0 < length)
		*buf = '\0';
}

/*
 * Return the string length of the resource record name refered to
 * by ptr within the message. The length includes for name label
 * (dot) delimiters and the final root label.
 */
static long
lengthName(unsigned char *message, unsigned char *ptr)
{
	long length;

	length = 0;
	for ( ; *ptr != 0; ptr += *ptr + 1) {
		if ((*ptr & 0xc0) == 0xc0) {
			unsigned long offset;
			GET_NET_SHORT(offset, ptr);
			ptr = &message[offset & 0x3fff];
		}

		length += *ptr + 1;
	}

	return length;
}

/*
 * Return a pointer to the next field within a resource record that
 * follows after the name field refered to by ptr.
 */
static unsigned char *
skipName(unsigned char *ptr)
{
	for ( ; ; ptr += *ptr + 1) {
		if (*ptr == 0) {
			++ptr;
			break;
		}
		if ((*ptr & 0xc0) == 0xc0) {
			ptr += 2;
			break;
		}
	}

	return ptr;
}

/*
 * Return an allocated C string for the resource record name refered
 * to by the ptr within the message. The pointer to the next unread
 * byte following the name field is passed back..
 */
/*@null@*/
static unsigned char *
getName(unsigned char *message, unsigned char *ptr, unsigned char **stop)
{
	long length;
	unsigned char *buf;

	length = lengthName(message, ptr);
	if (DOMAIN_LENGTH < length)
		return (unsigned char *) 0;

	if ((buf = (unsigned char *) malloc(length + 1)) == (unsigned char *) 0)
		return (unsigned char *) 0;

	copyName(message, ptr, buf, length + 1);

	if (stop != (unsigned char **) 0)
		*stop = skipName(ptr);

	return buf;
}

#ifdef DONT_NEED
/*
 * Return a pointer to the next resource record following the one
 * refered to by ptr.
 */
static unsigned char *
skipRecord(unsigned char *ptr)
{
	unsigned short length;

	ptr = skipName(ptr);
	NET_SHORT_ADV(ptr);		/* type */
	NET_SHORT_ADV(ptr);		/* class */
	NET_LONG_ADV(ptr);		/* ttl */
	GET_NET_SHORT_ADV(length, ptr);
	ptr += length;

	return ptr;
}
#endif

/***********************************************************************
 *** Response parsing.
 ***********************************************************************/

static void
DnsRecordRemoveEntry(void *record,  /*@unused@*/ void *data)
{
	DnsRecord entry = (DnsRecord) record;

	if (entry != (DnsRecord) 0) {
		if (entry->value != (char *) 0)
			free(entry->value);
		free(entry->name);
		free(entry);
	}
}

/*
 * Return true is we successfully allocated and parsed a resource
 * record. The DnsRecord passed should be freed by DnsRecordRemoveEntry().
 */
static int
getRecord(struct query *q, unsigned char **ptr, DnsRecord *record)
{
	DnsRecord entry;
	unsigned short length;

	entry = (DnsRecord) calloc(1, sizeof (struct DnsRecord));
	if (entry == (DnsRecord) 0)
		return 0;

	*record = entry;

	entry->name = getName((unsigned char *) &q->header, *ptr, ptr);

	GET_NET_SHORT_ADV(entry->type, *ptr);
	NET_SHORT_ADV(*ptr); /* class (assume IN) */
	GET_NET_LONG_ADV(entry->ttl, *ptr);
	GET_NET_SHORT_ADV(length, *ptr);

	switch (entry->type) {
	case TYPE_NS:
		free(entry->name);
		entry->name = getName((unsigned char *) &q->header, *ptr, ptr);
		break;
	case TYPE_CNAME:
		entry->value = getName((unsigned char *) &q->header, *ptr, ptr);
		break;
	case TYPE_MX:
		free(entry->name);
		GET_NET_SHORT_ADV(entry->preference, *ptr);
		entry->name = getName((unsigned char *) &q->header, *ptr, ptr);
		break;
	case TYPE_A:
		GET_NET_LONG_ADV(entry->ip.s_addr, *ptr);
		entry->value = ip4ToString(entry->ip.s_addr);
		/*** Leave it in network format ***/
		entry->ip.s_addr = htonl(entry->ip.s_addr);
		break;
	default:
		/* Skip this record. */
#ifndef NDEBUG
		if ((debug & DNS_DEBUG_TRACE) == DNS_DEBUG_TRACE)
			syslog(LOG_DEBUG, "getRecord() entry->name=%s entry->type=%d skipped", entry->name, entry->type);
#endif
		*ptr += length;
		free(entry->name);
		free(entry);
		return 0;
	}

#ifndef NDEBUG
		if ((debug & DNS_DEBUG_TRACE) == DNS_DEBUG_TRACE)
			syslog(LOG_DEBUG, "getRecord() entry->name=%s entry->type=%d", entry->name, entry->type);
#endif
	return 1;
}

static Vector
getRecordList(struct dns *dns)
{
	int i, j;
	char *find;
	Vector list;
	struct query *q;
	unsigned char *ptr;
	unsigned short type;
	DnsRecord entry, extra;

#ifndef NDEBUG
	if ((debug & DNS_DEBUG_TRACE) == DNS_DEBUG_TRACE)
		syslog(LOG_DEBUG, "enter getRecordList(%lx)", (long) dns);
#endif

	/* Allow room for multi-homed hosts. */
	if ((list = VectorCreate(10)) == (Vector) 0)
		goto error0;

	VectorOnRemove(list, DnsRecordRemoveEntry, (void **) 0);
	q = &((struct dns *) dns)->packet;

	/* Skip question section. */
	ptr = q->data;
	for (i = 0; i < q->header.qdcount; i++) {
		ptr = skipName(ptr);
		/* ASSUMPTION: I'm assuming I only asked one question,
		 * so I remember the type of that question in case I
		 * have to repeat it below.
		 */
		GET_NET_SHORT_ADV(type, ptr);
		NET_SHORT_ADV(ptr);
	}

	/* Add all the returned resource-records of interest to the list. */
	j = q->header.ancount + q->header.nscount + q->header.arcount;
	for (i = 0; i < j; i++) {
		if (getRecord(q, &ptr, &entry))
			VectorAdd(list, entry);
	}

	/* Use the extra information provided with the answer to
	 * complete the answers; ie. resolve any CNAME and A records
	 * given with the answer.
	 */
	for (i = 0; i < q->header.ancount; ++i) {
		entry = VectorGet(list, i);
		if (entry == (DnsRecord) 0)
			continue;
			
		if (entry->type == TYPE_A) {
#ifndef NDEBUG
			if ((debug & DNS_DEBUG_TRACE) == DNS_DEBUG_TRACE)
				syslog(LOG_DEBUG, "entry->name=%s entry->type=%d (A)", entry->name, entry->type);
#endif
			continue;
		}

		if (entry->type == TYPE_CNAME) {
#ifndef NDEBUG
			if ((debug & DNS_DEBUG_TRACE) == DNS_DEBUG_TRACE)
				syslog(LOG_DEBUG, "entry->name=%s entry->type=%d (CNAME)", entry->name, entry->type);
#endif
			if (1 < q->header.ancount) {
				/* We have more than one answer, assume that
				 * the other answers are the recusive result
				 * provided by the upstream DNS server. So
				 * just remove the CNAME from the result.
				 */
				VectorRemove(list, i, 1);
				q->header.ancount--;
				i--;
				continue;
			}

			/* RFC 974 "Issuing a Query" paragraph 4 about handling
			 * a CNAME response. We have to repeat the question with
			 * the domain name provided by the CNAME. Annoying, but
			 * legal.
			 *
			 * ASSUMPTION: Applies to NS records as well.
			 */
			find = TextDup(entry->value);
			VectorDestroy(list);
			list = queryAndGetHosts(dns, find, type);
			free(find);

			goto done;
		}

		find = entry->name;
#ifndef NDEBUG
		if ((debug & DNS_DEBUG_TRACE) == DNS_DEBUG_TRACE)
			syslog(LOG_DEBUG, "entry->name=%s entry->type=%d", entry->name, entry->type);
#endif

		/* Loop through the extra records. */
		for (j = q->header.ancount + q->header.nscount; j < VectorLength(list); j++) {
			extra = VectorGet(list, j);
			if (extra == (DnsRecord) 0)
				continue;

			if (TextInsensitiveCompare(find, extra->name) == 0) {
				if (extra->type == TYPE_CNAME) {
#ifndef NDEBUG
					if ((debug & DNS_DEBUG_TRACE) == DNS_DEBUG_TRACE)
						syslog(LOG_DEBUG, "found %s CNAME %s", find, extra->value);
#endif
					find = extra->value;
					continue;
				}
				if (extra->type == TYPE_A) {
#ifndef NDEBUG
					if ((debug & DNS_DEBUG_TRACE) == DNS_DEBUG_TRACE)
						syslog(LOG_DEBUG, "found %s A", find);
#endif
					entry->ip = extra->ip;
					if (entry->value != (char *) 0)
						free(entry->value);
					entry->value = TextDup(extra->value);
					break;
				}
#ifndef NDEBUG
				if ((debug & DNS_DEBUG_TRACE) == DNS_DEBUG_TRACE)
					syslog(LOG_DEBUG, "found %s type=%d", find, extra->type);
#endif
			}
		}
	}

	/* Discard the extra information provided. */
	VectorRemove(list, q->header.ancount, VectorLength(list) -  q->header.ancount);
done:
error0:
#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "exit getRecordList(%lx) list=%lx", (long) dns, (long) list);
#endif
	return list;
}

static Vector
readResolvConf(const char *cf)
{
	FILE *fp;
	char *line, *token;
	Vector servers, tokens;

	servers = (Vector) 0;

	if (cf == (char *) 0 || (fp = fopen(cf, "r")) == (FILE *) 0)
		goto error0;

	if ((line = malloc(BUFSIZ)) == (char *) 0)
		goto error1;

	if ((servers = VectorCreate(10)) == (Vector) 0)
		goto error2;

	VectorOnRemove(servers, VectorRemoveTrivial, (void *) 0);

	while (0 <= TextInputLine(fp, line, BUFSIZ)) {
		if (*line == '\0' || *line == '#')
			continue;
		
		if (TextSensitiveStartsWith(line, "nameserver") < 0)
			continue;

		line[strcspn(line, "\r\n")] = '\0';
		if ((tokens = TextSplit(line, " \t", 0)) == (Vector) 0)
			continue;
		
		if (VectorLength(tokens) == 2) {
			token = VectorGet(tokens, 1);
			VectorSet(tokens, 1, (void *) 0);
			VectorAdd(servers, token);
		}
				
		VectorDestroy(tokens);
	}

	if (!feof(fp)) {
		VectorDestroy(servers);
		servers = (Vector) 0;
	}
error2:
	free(line);
error1:
	fclose(fp);
error0:
	return servers;
}

static Vector
getSingleIP(const char *address, int type)
{
	long length;
	Vector list;
	DnsEntry host;

	if ((list = VectorCreate(1)) == (Vector) 0)
		goto error0;		

	VectorOnRemove(list, DnsRecordRemoveEntry, (void **) 0);

	if ((host = (DnsEntry) calloc(1, sizeof (struct DnsEntry))) == (DnsEntry) 0)
		goto error1;

	/* Is it an "IP as domain" string? ie. [1.2.3.4] */
	if (0 < (length = MailSpanAddressLiteral(address))) {
		address = (const char *) TextSubstring(address, 1, length - 2);
		if (address == (const char *) 0)
			goto error2;
	}
		
	/* Do we have an IP address? */
#if defined(HAVE_INET_PTON)
	if (inet_pton(AF_INET, address, &host->ip) == 0)
		goto error3;	
#elif defined(HAVE_INET_ATON)
	if (inet_aton(address, &host->ip) == 0)
		goto error3;	
#else
	if ((host->ip.s_addr = inet_addr(address)) == INADDR_NONE)
		goto error3;	
#endif

	/* Yeah! We have an IP address. */
	if ((host->name = strdup(address)) == (char *) 0)
		goto error3;

	if ((host->value = strdup(address)) == (char *) 0)
		goto error3;

	host->type = (unsigned short) type;
	host->ttl = 1;
	
	if (VectorAdd(list, host))
		goto error3;

	if (0 < length)
		free((char *) address);

	return list;
error3:
	if (0 < length)
		free((char *) address);
error2:
	DnsRecordRemoveEntry(host, (void *) 0);
error1:
	VectorDestroy(list);
error0:
	return (Vector) 0;
}

static Vector
DnsGetMatchingIP(Dns dns, Vector list)
{
	int i;
	Vector Alist;
	DnsRecord listEntry, Aentry;

#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "enter DnsGetMatchingIP(%lx, %lx)", (long) dns, (long) list);
#endif

	/* Make sure we have the IPs for all the list members. */
	for (i = VectorLength(list); 0 <= --i; ) {
		listEntry = VectorGet(list, i);
#ifndef NDEBUG
		if ((debug & DNS_DEBUG_TRACE) == DNS_DEBUG_TRACE) {
			char addr[50];
			ConvertIpToString(&listEntry->ip, addr, sizeof addr);
			syslog(LOG_DEBUG, "index=%d name=%s ip=%s", i, listEntry->name, addr);
		}
#endif
		if (listEntry->ip.s_addr != 0)
			continue;

		Alist = DnsGetAddresses(dns, listEntry->name);
		if (Alist != (Vector) 0 && 0 < VectorLength(Alist)) {
			Aentry = VectorGet(Alist, 0);
			listEntry->ip = Aentry->ip;
			if (listEntry->value != (char *) 0)
				free(listEntry->value);
			listEntry->value = TextDup(Aentry->value);
		} else {
			/* A record lookup failed. Note that the
			 * Return code and error message will have
			 * been modified. Fix below.
			 */
			VectorRemove(list, i, 1);
		}

		VectorDestroy(Alist);
	}

#ifdef HMM
/* Which is better? An empty list or a null pointer? */
	if (VectorLength(list) <= 0) {
		VectorDestroy(list);
		return (Vector) 0;
	}
#endif

	/* If we have at least one complete result, thats OK. */
	DnsSetError(dns, RCODE_OK, "");

#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "exit DnsGetMatchingIP(%lx, %lx)", (long) dns, (long) list);
#endif
	return list;
}

static int
compareMxPreferences(const void *a, const void *b)
{
	return (*(DnsEntryMx *) a)->preference - (*(DnsEntryMx *) b)->preference;
}

/***********************************************************************
 *** Principal DNS routines.
 ***********************************************************************/

/*
 * @param mask
 *	Set the debug detail according to the bit-wise OR of
 * 	DNS_DEBUG_* values.
 */
void
DnsSetDebugMask(int mask)
{
	debug = mask;
}

/*
 * (Re)Load the /etc/resolv.conf file. Currently only nameserver lines
 * are recognised. This function should be called before the first
 * DnsOpen() call.
 *
 * @return
 *	Zero (0) on success otherwise -1 one error.
 */
int
DnsInit(void)
{
	VectorDestroy(nameServers);

	if ((nameServers = readResolvConf(RESOLV_CONF)) == (Vector) 0)
		return -1;
			
	return 0;
}

/*
 * @param servers
 *	A NULL terminate array of C string pointers. Each C string
 * 	is the IP address of a DNS server to consult in order of
 *	preference. This argument can be NULL, in which case the
 *	list of DNS servers from /etc/resolv.conf are consulted.
 *
 * @return 
 *	A Dns object returned by DnsOpen().
 */
Dns
DnsOpen(const char **servers)
{
	int i;
	struct dns *dns;
	InetAddress addr;

#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "enter DnsOpen(%lx)", (long) servers);
#endif
	if (servers == (const char **) 0 || servers[0] == (const char *) 0)
		servers = (const char **) VectorBase(nameServers);
			
	for (i = 0; servers[i] != (const char *) 0; i++)
		;

	/* Allocate one block for everything. */
	dns = (struct dns *) malloc(sizeof *dns + i * sizeof (InetAddress));
	if (dns == (struct dns *) 0)
		goto error0;

	if (InetAddressInitIP(&addr, "127.0.0.1", DNS_PORT))
		goto error1;

	/* Open a UDP socket. We don't care about the host or port. */
	if ((dns->socket = SocketOpen(&addr, SOCK_DGRAM)) == (Socket) 0)
		goto error1;

	dns->servers = (InetAddress *)(dns + 1);
	dns->nservers = i;

	for (i = 0; i < dns->nservers; i++) {
		if (InetAddressInitIP(&dns->servers[i], servers[i], DNS_PORT))
			goto error1;
	}

	dns->counter = (unsigned short) rand();
	DnsSetTimeout(dns, DNS_DEFAULT_TIMEOUT);
	DnsSetAttempts(dns, DNS_DEFAULT_ATTEMPTS);

#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "exit DnsOpen(%lx) Dns=%lx", (long) servers, (long) dns);
#endif
	return (Dns) dns;
error1:
	DnsClose(dns);
error0:
#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "error exit DnsOpen(%lx) Dns=0 errno=%d, %s", (long) servers, errno, strerror(errno));
#endif
	return (Dns) 0;
}

/*
 * @param dns
 *	A Dns object returned by DnsOpen() to close.
 */
void
DnsClose(Dns dns)
{
	struct dns *d = (struct dns *) dns;

#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "enter DnsClose(%lx)", (long) dns);
#endif
	if (dns != (Dns) 0) {
		SocketClose(d->socket);		
		free(d);
	}
#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "exit DnsClose(%lx)", (long) dns);
#endif
}

/*
 * @param dns
 *	A Dns object returned by DnsOpen().
 *
 * @param type
 *	The DNS resource record type to search for. See DNS_TYPE_*
 *
 * @param name
 *	A domain or host name to search for. In the case of a PTR
 * 	search, its the IP address.
 *
 * @return
 *	A Vector of DnsEntry pointers on success, otherwise a NULL
 *	pointer on error. The Vector may be empty. When finished
 *	with the Vector, pass it to VectorDestroy() to clean up.
 */
Vector
DnsGet(Dns dns, int type, const char *name)
{
	Vector list;
	const char *error;
	
#ifndef NDEBUG
	const char *typeName;

	switch (type) {
	case DNS_TYPE_A:	typeName = "A"; 	break;
	case DNS_TYPE_NS:	typeName = "NS";	break;
	case DNS_TYPE_CNAME:	typeName = "CNAME"; 	break;
	case DNS_TYPE_PTR:	typeName = "PTR"; 	break;
	case DNS_TYPE_MX:	typeName = "MX"; 	break;
	case DNS_TYPE_TXT:	typeName = "TXT"; 	break;
	default:		typeName = "(unknown)"; break;
	}

	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "enter DnsGet(%lx, %s=%d, %s)", (long) dns, typeName, type, name);
#endif
	if (name == (const char *) 0) {
		DnsSetError(dns, RCODE_ERRNO, "null name argument");
		list = (Vector) 0;
		errno = EINVAL;
		goto error0;
	}

	DnsSetError(dns, DNS_RCODE_OK, "");

	switch (type) {
	case DNS_TYPE_PTR:
		break;
	case DNS_TYPE_A:
	case DNS_TYPE_MX:
		/* Handle an for "IP as domain" form. */
		if ((list = getSingleIP(name, type)) != (Vector) 0)
			goto done;
		break;
	case DNS_TYPE_NS:
	case DNS_TYPE_TXT:
	case DNS_TYPE_CNAME:
		if ((list = getSingleIP(name, type)) != (Vector) 0) {
			VectorDestroy(list);
			DnsSetError(dns, RCODE_ERRNO, "cannot lookup NS, CNAME, or TXT with an IP address");
			list = (Vector) 0;
			errno = EINVAL;
			goto done;
		}
		break;
	default:
		DnsSetError(dns, RCODE_ERRNO, "unknown resource record type");
		list = (Vector) 0;
		errno = EINVAL;
		goto error0;
	}
	
	if ((error = buildQuery(dns, name, type, OP_QUERY)) != (const char *) 0) {
		DnsSetError(dns, RCODE_FORMAT, error);
		goto error0;
	}

	if (sendQuery(dns))
		goto error0;

	if ((list = parseResult(dns)) == (Vector) 0) {
		DnsSetError(dns, RCODE_ERRNO, "out of memory for result Vector");
		goto error0;
	}
	
	if (DnsGetReturnCode(dns) != RCODE_OK)
		goto error0;
			
	switch (type) {
	case DNS_TYPE_MX:
		if (VectorLength(list) <= 0) {
			/* RFC 974 section "Interpreting the List of MX RRs"
			 * paragraph 2 special case.
			 */
			VectorDestroy(list);
			list = DnsGet(dns, DNS_TYPE_A, name);
			break;			
		}
		
		VectorSort(list, compareMxPreferences);
		/*@fallthrough@*/
	case DNS_TYPE_NS:
	case DNS_TYPE_CNAME:
		/* Fill in the blanks. */
		DnsGetMatchingIP(dns, list);
		break;
	}
done:
error0:	
#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "exit DnsGet(%lx, %s=%d, %s) Vector=%lx rc=%d error=%s", (long) dns, typeName, type, name, (long) list, DnsGetReturnCode(dns), DnsGetError(dns));
#endif
	return list;
}

/*
 * @param dns
 *	A Dns object returned by DnsOpen().
 *
 * @return
 *	A C string error message for the corresponding DNS_RCODE_ value.
 */
const char *
DnsGetError(Dns dns)
{
	if (dns == (Dns) 0)
		return (const char *) "null pointer to DNS structure";

#ifndef NDEBUG
	if (debug & DNS_DEBUG_GET_SET)
		syslog(LOG_DEBUG, "DnsGetError(%lx) rcode=%d error=\"%s\"", (long) dns, ((struct dns *) dns)->rcode, ((struct dns *) dns)->error);
#endif

	return ((struct dns *) dns)->error;
}

/*
 * @param dns
 *	A Dns object returned by DnsOpen().
 *
 * @return
 *	A DNS_RCODE_ value. The value DNS_RCODE_NULL and DNS_RCODE_ERRNO
 *	are internal error conditions not returned from a DNS server.
 */
int
DnsGetReturnCode(Dns dns)
{
	if (dns == (Dns) 0) {
		errno = EFAULT;
		return RCODE_NULL;
	}

#ifndef NDEBUG
	if (debug & DNS_DEBUG_GET_SET)
		syslog(LOG_DEBUG, "DnsGetReturnCode(%lx) rcode=%d error=\"%s\"", (long) dns, ((struct dns *) dns)->rcode, ((struct dns *) dns)->error);
#endif

	return ((struct dns *) dns)->rcode;
}

/*
 * @param dns
 *	A Dns object returned by DnsOpen().
 *
 * @param n
 *	Maximum number of times to cycle through the DNS server
 *	list before giving up on a search. Default is 4.
 */
void
DnsSetAttempts(Dns dns, int attempts)
{
	if (attempts < 1)
		attempts = 1;
#ifndef NDEBUG
	if (debug & DNS_DEBUG_GET_SET)
		syslog(LOG_DEBUG, "DnsSetAttempts(%lx, %d)", (long) dns, attempts);
#endif
	((struct dns *) dns)->attempts = attempts * ((struct dns *) dns)->nservers;
}

/*
 * @param dns
 *	A Dns object returned by DnsOpen().
 *
 * @param ms
 *	The initial timeout in milliseconds to wait for a response
 *	from a DNS server. An exponential backup algorithm is used.
 *	The default is 5000 ms.
 *
 * Using the defaults and assuming a single DNS server, then the
 * exponential backoff algorithm will result in a max. delay of
 * 75 seconds (5+10+20+40) to find an answer before giving up.
 *
 * With more than one server, the timeout each round is divided
 * by the number of servers available. So for example:
 *
 *	2 servers: 5+5+ 5+5+ 10+10+ 20+20        = 80 seconds
 *	3 servers: 5+5+5+ 5+5+5+ 6+6+6+ 13+13+13 = 87 seconds
 */
void
DnsSetTimeout(Dns dns, long ms)
{
	if (ms < 1000)
		ms = 1000;

#ifndef NDEBUG
	if (debug & DNS_DEBUG_GET_SET)
		syslog(LOG_DEBUG, "DnsSetTimeout(%lx, %ld)", (long) dns, ms);
#endif
	((struct dns *) dns)->timeout = ms;
}
