/**
 * DnsTest.c
 *
 * RFC 1035, 2821
 *
 * Copyright 2002, 2004 by Anthony Howe. All rights reserved.
 */

#undef NDEBUG

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <com/snert/lib/version.h>
#include <com/snert/lib/io/Dns.h>
#include <com/snert/lib/io/Log.h>
#include <com/snert/lib/io/Socket.h>

#undef DUMP

int
main(int argc, char **argv)
{
	const char *error;
	Vector hosts;
	DnsRecord host;
	Dns dns;
	int i;

#ifndef NDEBUG
	LogSetLevel(LOG_DEBUG);
	LogSetProgramName("DnsTest");
	if (LogOpen("DnsTest.log")) {
		fprintf(stderr, "Failed to open \"DnsTest.log\": %s (%d)\n", strerror(errno), errno);
		return 2;
	}

	DnsSetDebugMask(DNS_DEBUG_ALL);
#endif

	if (argc < 2) {
		fprintf(stderr, "usage: DnsTest host [dns ...]\n");
		fprintf(stderr, LIBSNERT_VERSION " " LIBSNERT_COPYRIGHT "\n");
		return 1;
	}

#ifndef NDEBUG
	LogDebug("SocketInit()...");
#endif
	if (SocketInit())
		return 1;

	if (DnsInit())
		return 1;

#ifndef NDEBUG
	LogDebug("DnsOpen(%lx)...", &argv[2]);
#endif
	dns = DnsOpen((const char **) &argv[2]);

/**********************************************************************
 * Find A reconds.
 **********************************************************************/
#ifndef NDEBUG
	LogInfo("looking up A records...");
#endif

#ifndef NDEBUG
	LogDebug("DnsGetAddresses(%lx, %s)...", (long) dns, argv[1]);
#endif
	hosts = DnsGetAddresses(dns, argv[1]);
#ifndef NDEBUG
	LogDebug("DnsGetAddresses(%lx, %s) = %lx", (long) dns, argv[1], hosts);
#endif

	if (DnsGetReturnCode(dns) != RCODE_OK) {
		error = DnsGetError(dns);
		fprintf(stderr, "DnsGetAddresses() failed: %s: %s (%d)\n", error, strerror(errno), errno);
		return 2;
	}

	printf("A records\n");
	if (hosts == NULL) {
		printf("Result vector is null.\n");
	} else if (VectorLength(hosts) <= 0) {
		printf("Result vector is empty.\n");
	} else {
		for (i = 0; i < VectorLength(hosts); i++) {
			host = VectorGet(hosts, i);
			printf("%s %s\n", host->name, host->value);
		}

		VectorDestroy(hosts);
	}
	fflush(stdout);
#ifdef DUMP
	DnsDumpLast(dns);
#endif

/**********************************************************************
 * Find MX reconds.
 **********************************************************************/
#ifndef NDEBUG
	LogInfo("looking up MX records...");
#endif

	hosts = DnsGetMailServers(dns, argv[1]);

	if (DnsGetReturnCode(dns) != RCODE_OK) {
		error = DnsGetError(dns);
		fprintf(stderr, "DnsGetMailServers() failed: %s: %s (%d)\n", error, strerror(errno), errno);
		return 2;
	}

	printf("MX records\n");
	if (hosts == NULL) {
		printf("Result vector is null.\n");
	} else if (VectorLength(hosts) <= 0) {
		printf("Result vector is empty.\n");
	} else {
		for (i = 0; i < VectorLength(hosts); i++) {
			host = VectorGet(hosts, i);
			printf("%s %s preference=%d\n", host->name, host->value, host->preference);
		}

		VectorDestroy(hosts);
	}
	fflush(stdout);
#ifdef DUMP
	DnsDumpLast(dns);
#endif

/**********************************************************************
 * Find NS reconds.
 **********************************************************************/
#ifndef NDEBUG
	LogInfo("looking up NS records...");
#endif

	hosts = DnsGetNameServers(dns, argv[1]);

	if (DnsGetReturnCode(dns) != RCODE_OK) {
		error = DnsGetError(dns);
		fprintf(stderr, "DnsGetNameServers() failed: %s: %s (%d)\n", error, strerror(errno), errno);
		return 2;
	}

	printf("NS records\n");
	if (hosts == NULL) {
		printf("Result vector is null.\n");
	} else if (VectorLength(hosts) <= 0) {
		printf("Result vector is empty.\n");
	} else {
		for (i = 0; i < VectorLength(hosts); i++) {
			host = VectorGet(hosts, i);
			printf("%s %s\n", host->name, host->value);
		}
		VectorDestroy(hosts);
	}
	fflush(stdout);
#ifdef DUMP
	DnsDumpLast(dns);
#endif

/**********************************************************************
 * Clean-Up
 **********************************************************************/

#ifndef NDEBUG
	LogDebug("DnsClose(%lx)...", dns);
#endif
	DnsClose(dns);

#ifndef NDEBUG
	LogDebug("LogClose() and exit...");
#endif

#ifndef NDEBUG
	LogClose();
#endif

	return 0;
}

