/*
 * Log.c
 *
 * Copyright 2002, 2004 by Anthony Howe.  All rights reserved.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>

#include <com/snert/lib/io/Log.h>
#include <com/snert/lib/io/posix.h>

#if defined(__WIN32__)
extern long getpid(void);
#else
#include <unistd.h>
#endif

static FILE *logFile;
static int logLevel = LOG_INFO;
static char *logLevels[] = { "PANIC", "ALERT", "FATAL", "ERROR", "WARN", "NOTICE", "INFO", "DEBUG" };

static const char *programName = (char *) 0;

const char *
LogGetProgramName(void)
{
	return programName;
}

void
LogSetProgramName(const char *name)
{
	if (programName == (char *) 0) {
		programName = name;
		errno = 0;
	}
}

static void
LogNewline(FILE *fp)
{
	if (fp != (FILE *) 0) {
		(void) fprintf(fp, "\n");
		(void) fflush(fp);
	}
}

/*
 * Return true if successfully logged, otherwise false if skipped.
 */
void
LogV(int level, const char *msg, va_list args)
{
	time_t now;
	struct tm *gmt;
	char *nl, stamp[21]; /* yyyy-mm-dd HH:MM:SS */

	if (logLevel < level || logFile == (FILE *) 0)
		return;

	now = time((time_t *) 0);
	gmt = gmtime(&now);
	(void) strftime(stamp, sizeof stamp-1, "%Y-%m-%d %H:%M:%S", gmt);
	(void) fprintf(
		logFile, "%s %s:%04lu %s ",
		stamp, programName, (unsigned long) getpid(), logLevels[level]
	);

	(void) vfprintf(logFile, msg, args);
	
	if ((nl = strrchr(msg, '\n')) == NULL || nl[1] != '\0')
		LogNewline(logFile);
}

void
Log(int level, const char *msg, ...)
{
	va_list args;
	va_start(args, msg);

	LogV(level, msg, args);

	va_end(args);
}

void
LogSetLevel(int level)
{
	if (LOG_FATAL <= level)
		logLevel = level;
}

void
LogDebug(const char *msg, ...)
{
	va_list args;
	va_start(args, msg);

	LogV(LOG_DEBUG, msg, args);

	va_end(args);
}

void
LogInfo(const char *msg, ...)
{
	va_list args;
	va_start(args, msg);

	LogV(LOG_INFO, msg, args);

	va_end(args);
}

void
LogWarn(const char *msg, ...)
{
	va_list args;
	va_start(args, msg);

	LogV(LOG_WARN, msg, args);

	va_end(args);
}

void
LogStderrV(int level, const char *msg, va_list args)
{
	if (logFile != (FILE *) 0) {
		(void) LogV(level, msg, args);
		if (errno != 0)
			(void) fprintf(logFile, ": (%d) %s", errno, strerror(errno));
		LogNewline(logFile);
	}

	(void) fprintf(stderr, "%s: ", programName);
	(void) vfprintf(stderr, msg, args);
	if (errno != 0)
		(void) fprintf(stderr, ": (%d) %s", errno, strerror(errno));
	LogNewline(stderr);
}

void
LogStderr(int level, const char *msg, ...)
{
	va_list args;
	va_start(args, msg);

	LogStderrV(level, msg, args);

	va_end(args);
}

void
LogError(const char *msg, ...)
{
	va_list args;
	va_start(args, msg);

	LogStderrV(LOG_ERROR, msg, args);

	va_end(args);
}

void
LogFatal(const char *msg, ...)
{
	va_list args;
	va_start(args, msg);

	LogStderrV(LOG_FATAL, msg, args);

	va_end(args);

	exit(9);
}

int
LogOpen(const char *fname)
{
	if ((logFile = fopen(fname, "a")) == (FILE *) 0) {
		LogError("failed to open log file \"%s\"", fname);
		return -1;
	}
#if defined(__BORLANDC__)
	setmode(fileno(logFile), O_BINARY);
#endif	
	return 0;
}

void
LogClose(void)
{
	(void) fclose(logFile);
}

/*
 * openlog() replacement for systems without <syslog.h>, like Windows
 */
void
LogOpenLog(const char *ident, int option, int facility)
{
	char *filename;

	if ((filename = malloc(strlen(ident) + 5)) != (char *) 0) {
		LogSetProgramName(ident);
		(void) sprintf(filename, "%s.log", ident);
		(void) LogOpen(filename);
		free(filename);
	}
}

