/*
 * Message.c
 *
 * Copyright 2001, 2004 by Anthony Howe. All rights reserved.
 */

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <com/snert/lib/io/posix.h>
#include <com/snert/lib/io/Message.h>
#include <com/snert/lib/io/Socket.h>

/**
 * <p>
 * Message consists of a variable length array of bytes passed
 * between processes.
 * </p><p>
 * Below is the format used for Version 1 messages:
 * </p>
 * <pre>
 *	------------------------------------------------
 *	| VN | CREATED | SENT    | LENGTH | CONTENT... |
 *	------------------------------------------------
 *	   1      8        8         4      LENGTH
 * </pre>
 * <p>
 * The VN is the version number byte. CREATED and SENT
 * are milliseconds since 1 Jan. 1970 00:00:00 GMT. All
 * values are in network byte order (highest byte first).
 * </p>
 *
 * @see snert.lib.io.MessageStream
 *
 * @author Anthony Howe
 * @version 1.0
 */

#define MESSAGE_VERSION			0x01

#define MESSAGE_HEADER_LENGTH		21

/*
 * Opaque data type.
 */
struct message {
	JavaTime _created;
	JavaTime _sent;
	JavaTime _received;
	long _length;			/* Length of content. */
	unsigned char *_content;	/* Message, may contain binary. */
};

/*
 * Returns a message of a given size with undefined content;
 * otherwise null on allocation error.
 */
Message
MessageCreate(long length)
{
	struct message *m = calloc(1, sizeof *m);

	if (m == (Message) 0)
		return (Message) 0;

	m->_content = malloc((size_t) length + 1);
	if (m->_content == (unsigned char *) 0) {
		free(m);
		return (Message) 0;
	}

	m->_length = length;
	m->_content[length] = '\0';
	m->_created = JavaTimeCreate(time((time_t *) 0));

	return (Message) m;
}

/**
 * <p>
 * Create a new message that can be passed later to
 * MessageStream.write().
 * </p>
 *
 * @param s
 *	An array of bytes.
 *
 * @parem length
 *	Length of byte array.
 *
 * @see snert.lib.io.MessageStream#write(Message)
 */
Message
MessageFromBytes(unsigned char *s, long length)
{
	struct message *copy = MessageCreate(length);
	memcpy(copy->_content, s, length);
	return (Message) copy;
}

/**
 * <p>
 * Create a new message that can be passed later to
 * MessageStream.write().
 * </p>
 *
 * @param content
 *	A string that will be converted to bytes.
 *
 * @see snert.lib.io.MessageStream#write(Message)
 */
Message
MessageFromString(const char *s)
{
	size_t length = strlen(s);
	struct message *copy = MessageCreate(length);
	memcpy(copy->_content, s, length);
	return (Message) copy;
}

/**
 * <p>
 * Create a new message that can be passed later to
 * MessageStream.write(). This message is a duplicate
 * of the original, with the exception that the sent
 * and received timestamps are initially null.
 * </p>
 *
 * @param orig
 *	A message.
 *
 * @see snert.lib.io.MessageStream#write(Message)
 */
Message
MessageFromMessage(Message orig)
{
	struct message *m = (struct message *) orig;
	struct message *copy = MessageCreate(m->_length);
	memcpy(copy->_content, m->_content, m->_length);
	copy->_created = m->_created;
	return (Message) copy;
}

/**
 * <p>
 * Create an exact copy of this Message, including all
 * timestamp fields.
 * </p>
 *
 * @return
 *	A copy of the Message.
 */
Message
MessageClone(Message msg)
{
	struct message *orig = (struct message *) msg;
	struct message *copy = MessageFromMessage(msg);
	copy->_received = orig->_received;
	copy->_sent = orig->_sent;
	return (Message) copy;
}

/*
 * Releases memory used by a message.
 */
void
MessageFree(Message msg)
{
	struct message *m = (struct message *) msg;

	if (m->_content != (unsigned char *) 0)
		free(m->_content);
	if (m->_created != (JavaTime) 0)
		free(m->_created);
	if (m->_sent != (JavaTime) 0)
		free(m->_sent);
	if (m->_received != (JavaTime) 0)
		free(m->_received);
	free(m);
}

/**
 * <p>
 * Get the date-time this message was created.
 * </p>
 *
 * @return
 *	The date-time this message was created.
 */
JavaTime
MessageGetDateCreated(Message msg)
{
	return ((struct message *) msg)->_created;
}

/**
 * <p>
 * Get the date-time this message was sent.
 * </p>
 *
 * @return
 *	The date-time this message was sent.
 *	May be null if not yet sent.
 */
JavaTime
MessageGetDateSent(Message msg)
{
	return ((struct message *) msg)->_sent;
}

/**
 * <p>
 * Get the date-time this message was received.
 * </p>
 *
 * @return
 *	The date-time this message was received.
 *	May be null if never sent.
 */
JavaTime
MessageGetDateReceived(Message msg)
{
	return ((struct message *) msg)->_received;
}

/*
 * Return the length of the message.
 */
long
MessageLength(Message msg)
{
	struct message *m = (struct message *) msg;
	return m->_content == (unsigned char *) 0 ? 0 : m->_length;
}

/**
 * <p>
 * Get a reference to the Message content.
 * </p>
 *
 * @return
 *	A reference to a byte array.
 */
unsigned char *
MessageGetContent(Message msg)
{
	return ((struct message *) msg)->_content;
}

/**
 * <p>
 * Set (copy) the content for this Message.
 * </p>
 *
 * @param content
 *	An array of bytes to be copied into the message.
 *
 * @return
 *	Zero (0) on success or -1 on error.
 */
int
MessageSetContent(Message msg, unsigned char *content, long length)
{
	struct message *m = (struct message *) msg;
	unsigned char *replace = realloc(m->_content, length+1);

	if (replace == (unsigned char *) 0)
		return -1;

	memcpy(replace, content, length);
	replace[length] = '\0';
	m->_content = replace;
	m->_length = length;

	return 0;
}

/**
 * <p>
 * Converts the array of bytes into a C string.
 * </p>
 *
 * @param msg
 *	The Message object.
 *
 * @return
 *	A string.
 */
char *
MessageToString(Message msg)
{
	struct message *m = (struct message *) msg;
	m->_content[m->_length] = '\0';
	return (char *) strdup((char *) m->_content);
}

/**
 * <p>
 * Test if this Message equals another object.
 * </p>
 *
 * @return
 *	True if this Message is the same as the object. To be equal
 *	the same creation timestamp, content length, and content
 *	must be the same.
 *
 * @see #Message(Message)
 * @see #clone()
 */
int
MessageEquals(Message A, Message B)
{
	struct message *a = (struct message *) A;
	struct message *b = (struct message *) B;

	if (JavaTimeCompare(a->_created, b->_created) != 0)
		return 0;

	if (a->_length != b->_length)
		return 0;

	return memcmp(a->_content, b->_content, b->_length) == 0;
}

/**
 * <p>
 * Computes a hash value for this Message.
 * </p>
 *
 * @return
 *	A hash value.
 */
long
MessageHashCode(Message msg)
{
	struct message *m = (struct message *) msg;

	long i, hash = m->_created->seconds;

	for (i = 0; i < m->_length; ++i)
		hash = 37 * hash + m->_content[i];

	hash += m->_length;

	return hash;
}

static void
MessageFillHeader(struct message *m, unsigned char *header)
{
	header[0] = MESSAGE_VERSION;

	if (JavaTimeFillJavaLong(m->_created, header+1))
		goto error0;

	if (JavaTimeFillJavaLong(m->_sent, header+9))
		goto error0;

	header[17] = (unsigned char) ((m->_length >> 24) & 0xff);
	header[18] = (unsigned char) ((m->_length >> 16) & 0xff);
	header[19] = (unsigned char) ((m->_length >> 8) & 0xff);
	header[20] = (unsigned char) ((m->_length) & 0xff);

	return;
error0:
	memset(header, 0, MESSAGE_HEADER_LENGTH);
}

static struct message *
MessageParseHeader(struct message *m, unsigned char *header)
{
	long length;

	length  = header[17] << 24;
	length |= header[18] << 16;
	length |= header[19] << 8;
	length |= header[20];

	if (m == (struct message *) 0)
		m = MessageCreate(length);
	else
		m->_length = length;

	m->_created = JavaTimeFromJavaLong(header + 1);
	m->_sent = JavaTimeFromJavaLong(header + 9);

	return m;
}

Message
MessageRead(Message msg, FILE *fp)
{
	size_t n;
	unsigned char *replace;
	struct message *m = (struct message *) msg;
	unsigned char header[MESSAGE_HEADER_LENGTH];

	n = fread(header, 1, 1, fp);
	if (n != 1 || header[0] != MESSAGE_VERSION)
		return (Message) 0;

	n = fread(header + 1, sizeof header - 1, 1, fp);
	if (n != 1)
		return (Message) 0;

	m = MessageParseHeader(m, header);
	m->_received = JavaTimeCreate(time((time_t *) 0));

	replace = realloc(m->_content, m->_length+1);
	if (replace == (unsigned char *) 0)
		return (Message) 0;
	m->_content = replace;
	m->_content[m->_length] = '\0';

	n = fread(m->_content, sizeof *m->_content, m->_length, fp);
	if ((long) n != m->_length)
		return (Message) 0;

	return m;
}

int
MessageWrite(Message msg, FILE *fp)
{
	size_t n;
	struct message *m = (struct message *) msg;
	unsigned char header[MESSAGE_HEADER_LENGTH];

	m->_sent = JavaTimeCreate(time((time_t *) 0));

	MessageFillHeader(m, header);
	n = fwrite(header, sizeof header, 1, fp);
	if (n != 1)
		return -1;

	n = fwrite(m->_content, sizeof *m->_content, m->_length, fp);
	if ((long) n != m->_length)
		return -1;

	return 0;
}

int
MessageWriteString(unsigned char *s, FILE *fp)
{
	int rc;
	struct message m;
	struct JavaTime c;

	JavaTimeSet(&c, time((time_t *) 0));

	m._content = s;
	m._length = (long) strlen(s);
	m._sent = m._received = (JavaTime) 0;
	m._created = &c;

	rc = MessageWrite(&m, fp);

	free(m._sent);

	return rc;
}

Message
MessageReadFd(Message msg, int fd)
{
	unsigned char *replace;
	struct message *m = (struct message *) msg;
	unsigned char header[MESSAGE_HEADER_LENGTH];

	if (read(fd, header, sizeof header) != (long) sizeof header)
		return (Message) 0;

	if (header[0] != MESSAGE_VERSION)
		return (Message) 0;

	m = MessageParseHeader(m, header);
	m->_received = JavaTimeCreate(time((time_t *) 0));

	replace = realloc(m->_content, m->_length+1);
	if (replace == (unsigned char *) 0)
		return (Message) 0;

	m->_content = replace;
	m->_content[m->_length] = '\0';

	if (read(fd, m->_content, m->_length) != m->_length)
		return (Message) 0;

	return m;
}

int
MessageWriteFd(Message msg, int fd)
{
	long n;
	struct message *m = (struct message *) msg;
	unsigned char header[MESSAGE_HEADER_LENGTH];

	m->_sent = JavaTimeCreate(time((time_t *) 0));

	MessageFillHeader(m, header);
	n = write(fd, header, sizeof header);
	if (n != (long) sizeof header)
		return -1;

	n = (long) write(fd, m->_content, m->_length);
	if (n != m->_length)
		return -1;

	return 0;
}

int
MessageWriteStringFd(unsigned char *s, int fd)
{
	int rc;
	struct message m;
	struct JavaTime c;

	JavaTimeSet(&c, time((time_t *) 0));

	m._content = s;
	m._length = (long) strlen(s);
	m._sent = m._received = (JavaTime) 0;
	m._created = &c;

	rc = MessageWriteFd(&m, fd);

	free(m._sent);

	return rc;
}

int
MessageIsWaitingFd(int fd)
{
	char ch;

	return 0 < recv(fd, &ch, sizeof ch, MSG_PEEK);
}

Message
MessageReceiveFd(Message msg, int fd)
{
	unsigned char *replace;
	struct message *m = (struct message *) msg;
	unsigned char header[MESSAGE_HEADER_LENGTH];

	if (recv(fd, header, sizeof header, 0) != (long) sizeof header)
		/* EOF when 0 and errno = 0, otherwise error. */
		return (Message) 0;

	if (header[0] != MESSAGE_VERSION)
		return (Message) 0;

	m = MessageParseHeader(m, header);
	m->_received = JavaTimeCreate(time((time_t *) 0));

	replace = realloc(m->_content, m->_length+1);
	if (replace == (unsigned char *) 0)
		return (Message) 0;

	m->_content = replace;
	m->_content[m->_length] = '\0';

	if (recv(fd, m->_content, m->_length, 0) != m->_length)
		/* EOF when 0 and errno = 0, otherwise error. */
		return (Message) 0;

	return m;
}

int
MessageSendFd(Message msg, int fd)
{
	long n;
	struct message *m = (struct message *) msg;
	unsigned char header[MESSAGE_HEADER_LENGTH];

	m->_sent = JavaTimeCreate(time((time_t *) 0));

	MessageFillHeader(m, header);
	n = send(fd, header, sizeof header, 0);
	if (n != (long) sizeof header)
		return -1;

	n = (long) send(fd, m->_content, m->_length, 0);
	if (n != m->_length)
		return -1;

	return 0;
}

int
MessageSendStringFd(unsigned char *s, int fd)
{
	int rc;
	struct message m;
	struct JavaTime c;

	JavaTimeSet(&c, time((time_t *) 0));

	m._content = s;
	m._length = (long) strlen(s);
	m._sent = m._received = (JavaTime) 0;
	m._created = &c;

	rc = MessageSendFd(&m, fd);

	free(m._sent);

	return rc;
}



