/*
 * MessageStream.c
 *
 * Copyright 2001, 2004 by Anthony Howe. All rights reserved.
 */

#define MESSAGE_STREAM_THREADED

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <com/snert/lib/io/posix.h>
#include <com/snert/lib/io/MessageStream.h>
#include <com/snert/lib/io/Socket.h>
#include <com/snert/lib/io/URI.h>
#include <com/snert/lib/sys/Thread.h>

struct message_stream;

typedef Message (*msg_read)(struct message_stream *);
typedef int (*msg_write_fd)(void *, int);
typedef int (*msg_write)(struct message_stream *, void *);
typedef int (*msg_open)(struct message_stream *);

struct message_stream {
	URI _uri;
	int _in;
	int _out;
	long _timeout;
	msg_open _open;
	msg_read _read;
	msg_write _write;
	msg_write_fd _write_fd;
#ifdef MESSAGE_STREAM_THREADED
	Socket _client;
	Socket _server;
	Thread _listener;
#else
	Socket _socket;
#endif
};

int
MessageStreamInit(void)
{
	return SocketInit();
}

static Message
MessageStreamFileRead(struct message_stream *ms)
{
	if (ms->_in < 0) {
		ms->_in = open(ms->_uri->path, O_RDONLY|O_BINARY);
		if (ms->_in < 0)
			return (Message) 0;
	}

	return MessageReadFd((Message) 0, ms->_in);
}

static int
MessageStreamFileWrite(struct message_stream *ms, void *msg)
{
	if (ms->_out < 0) {
		ms->_out = open(ms->_uri->path, O_CREAT|O_WRONLY|O_APPEND|O_BINARY, 0666);
		if (ms->_out < 0)
			return -1;
	}

	return (*ms->_write_fd)(msg, ms->_out);
/*	return MessageWriteFd(msg, ms->_out); */
}

#ifdef MESSAGE_STREAM_THREADED
/*
 * A MessageStream server only waits for one client connection
 * then closes the server socket.
 */
static void
ServerThread(void *data)
{
	struct message_stream *ms = (struct message_stream *) data;

	ms->_server = SocketOpenTcpServer(ms->_uri->host, uriGetPort(ms->_uri), 1);
	if (ms->_server == (Socket) 0)
		return;

	ms->_client = SocketAccept(ms->_server);
	ms->_out = ms->_in = SocketGetFd(ms->_client);
}
#endif

/*
 * A MessageStream server only waits for one client connection
 * then closes the server socket.
 */
static int
MessageStreamServerOpen(struct message_stream *ms)
{
#ifdef MESSAGE_STREAM_THREADED
	if (0 < ms->_timeout) {
		ThreadWaitOn(ms->_listener, ms->_timeout);
		if (ms->_client == (Socket) 0)
			return ETIMEDOUT;
	} else {
		/* Wait for a client to connect to our server. */
		while (ms->_client == (Socket) 0)
			ThreadYield();
	}
#else
	if (ms->_socket == (Socket) 0)
		return EFAULT;

	if (SocketWaitForInput(ms->_socket, ms->_timeout) == ETIMEDOUT)
		return ETIMEDOUT;

	ms->_out = ms->_in = SocketGetFd(ms->_socket);
#endif

	return 0;
}

static int
MessageStreamClientOpen(struct message_stream *ms)
{
#ifdef MESSAGE_STREAM_THREADED
	if (ms->_client != (Socket) 0)
		return 0;

	/* Establish client to server connection. */
	ms->_client = SocketOpenTcpClient(ms->_uri->host, uriGetPort(ms->_uri));
	if (ms->_client == (Socket) 0)
		return errno;

	ms->_out = ms->_in = SocketGetFd(ms->_client);
#else
	if (ms->_socket != (Socket) 0)
		return 0;

	/* Establish client to server connection. */
	ms->_socket = SocketOpenTcpClient(ms->_uri->host, uriGetPort(ms->_uri));
	if (ms->_socket == (Socket) 0)
		return errno;

	ms->_out = ms->_in = SocketGetFd(ms->_socket);
#endif
	return 0;
}

static Message
MessageStreamSocketRead(struct message_stream *ms)
{
	switch ((ms->_open)(ms)) {
	case 0: /* OK */
		break;
	case ETIMEDOUT:
		return MESSAGE_TIMEOUT;
	default: /* Some other error. */
		return (Message) 0;
	}

#ifdef MESSAGE_STREAM_THREADED
	if (SocketWaitForInput(ms->_client, ms->_timeout) == ETIMEDOUT)
#else
	if (SocketWaitForInput(ms->_socket, ms->_timeout) == ETIMEDOUT)
#endif
		return MESSAGE_TIMEOUT;

	return MessageReceiveFd((Message) 0, ms->_in);
}

static int
MessageStreamSocketWrite(struct message_stream *ms, void *msg)
{
#ifdef MESSAGE_STREAM_THREADED
	if (ms->_client == (Socket *) 0 && (ms->_open)(ms))
#else
	if (ms->_socket == (Socket *) 0 && (ms->_open)(ms))
#endif
		return -1;

	return (*ms->_write_fd)(msg, ms->_out);
/*	return MessageSendFd(msg, ms->_out); */
}

MessageStream
MessageStreamOpen(const char *uri)
{
	struct message_stream *ms = malloc(sizeof *ms);

	if (ms == (struct message_stream *) 0)
		goto error0;

	ms->_uri = uriCreate(uri);
#ifdef MESSAGE_STREAM_THREADED
	ms->_client = (Socket) 0;
	ms->_server = (Socket) 0;
	ms->_listener = (Thread) 0;
#else
	ms->_socket = (Socket) 0;
#endif
	if (ms->_uri == (URI) 0)
		goto error1;

	if (strcmp(ms->_uri->scheme, "file") == 0) {
		ms->_open = (msg_open) 0;
		ms->_read = MessageStreamFileRead;
		ms->_write = MessageStreamFileWrite;
		ms->_write_fd = (msg_write_fd) MessageWriteFd;
	} else if (strcmp(ms->_uri->scheme, "file-text") == 0) {
		ms->_open = (msg_open) 0;
		ms->_read = MessageStreamFileRead;
		ms->_write = MessageStreamFileWrite;
		ms->_write_fd = (msg_write_fd) MessageWriteStringFd;
	} else if (strcmp(ms->_uri->scheme, "connect") == 0) {
		ms->_open = MessageStreamClientOpen;
		ms->_read = MessageStreamSocketRead;
		ms->_write = MessageStreamSocketWrite;
		ms->_write_fd = (msg_write_fd) MessageSendFd;
	} else if (strcmp(ms->_uri->scheme, "connect-text") == 0) {
		ms->_open = MessageStreamClientOpen;
		ms->_read = MessageStreamSocketRead;
		ms->_write = MessageStreamSocketWrite;
		ms->_write_fd = (msg_write_fd) MessageSendStringFd;
	} else if (strcmp(ms->_uri->scheme, "listen") == 0) {
#ifdef MESSAGE_STREAM_THREADED
		ms->_listener = ThreadCreate(ServerThread, ms);
		if (ms->_listener == (Thread) 0)
			goto error2;
#else
		ms->_socket = SocketOpenTcpServer(ms->_uri->host, uriGetPort(ms->_uri), 1);
		if (ms->_socket == (Socket) 0)
			goto error2;
#endif
		ms->_open = MessageStreamServerOpen;
		ms->_read = MessageStreamSocketRead;
		ms->_write = MessageStreamSocketWrite;
		ms->_write_fd = (msg_write_fd) MessageSendFd;
	} else {
		goto error2;
	}

	ms->_timeout = 0;
	ms->_out = -1;
	ms->_in = -1;

	return (MessageStream) ms;
error2:
	uriFree(ms->_uri);
error1:
	free(ms);
error0:
	return (MessageStream) 0;
}

int
MessageStreamClose(MessageStream mstream)
{
	int rc = 0;
	struct message_stream *ms = (struct message_stream *) mstream;

	if (ms->_read == MessageStreamFileRead) {
		if (0 <= ms->_out && close(ms->_out))
			rc = -1;
		if (0 <= ms->_in && close(ms->_in))
			rc = -1;
	} else {
#ifdef MESSAGE_STREAM_THREADED
		SocketClose(ms->_client);
		SocketClose(ms->_server);
		ThreadDestroy(ms->_listener);
#else
		SocketClose(ms->_socket);
#endif
	}

	uriFree(ms->_uri);
	free(ms);

	return rc;
}

Message
MessageStreamRead(MessageStream mstream)
{
	struct message_stream *ms = (struct message_stream *) mstream;
	return (*ms->_read)(ms);
}

int
MessageStreamWrite(MessageStream mstream, Message msg)
{
	struct message_stream *ms = (struct message_stream *) mstream;
	return (*ms->_write)(ms, msg);
}

int
MessageStreamWriteString(MessageStream mstream, unsigned char *str)
{
	struct message_stream *ms = (struct message_stream *) mstream;
	return (*ms->_write)(ms, (void *) str);
}

long
MessageStreamGetTimeout(MessageStream mstream)
{
	return ((struct message_stream *) mstream)->_timeout;
}

void
MessageStreamSetTimeout(MessageStream mstream, long milliseconds)
{
	((struct message_stream *) mstream)->_timeout = milliseconds;
}
