#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <com/snert/lib/io/MessageStream.h>

#define MARKER		__FILE__,__LINE__

static void
isNull(void *x, const char *file, int line)
{
	if (x != (void *) 0) {
		printf("\nFAILURE %s:%d\n", file, line);
		exit(1);
	}
}

static void
expectMessage(void *x, const char *file, int line)
{
	if (x == (void *) 0) {
		printf("\nFAILURE %s:%d null message: %s\n", file, line, strerror(errno));
		exit(1);
	}
	if (x == MESSAGE_TIMEOUT) {
		printf("\nFAILURE %s:%d message timeout: %s\n", file, line, strerror(errno));
		exit(1);
	}
}

static void
expectTimeout(void *x, const char *file, int line)
{
	if (x == (void *) 0) {
		printf("\nFAILURE %s:%d null message: %s\n", file, line, strerror(errno));
		exit(1);
	}
	if (x != MESSAGE_TIMEOUT) {
		printf("\nFAILURE %s:%d expected timeout: %s\n", file, line, strerror(errno));
		exit(1);
	}
}

static void
success(int rc, const char *file, int line)
{
	if (rc != 0) {
		printf("\nFAILURE %s:%d error: %s\n", file, line, strerror(errno));
		exit(1);
	}
}

static void
failure(int rc, const char *file, int line)
{
	if (rc == 0) {
		printf("\nFAILURE %s:%d succeed: %s\n", file, line, strerror(errno));
		exit(1);
	}
}

static void
sameMessage(Message a, Message b, const char *file, int line)
{
	if (!MessageEquals(a, b)) {
		printf("\nFAILURE %s:%d messages not equal\n", file, line);
		exit(1);
	}
}

static void
sameMessageString(Message a, char *b, const char *file, int line)
{
	if (strcmp(MessageGetContent(a), b) != 0) {
		printf("\nFAILURE %s:%d messages not equal\n", file, line);
		exit(1);
	}
}

unsigned testNumber = 0;

int
main(int argc, char **argv)
{
	Message m;
	char s1[] = "Hello world!";
	char s2[] = "Beat me with a wet noodle.";
	char s3[] = "Oh you would like that sort of thing.";
	Message m1 = MessageFromString(s1);
	Message m2 = MessageFromString(s2);
	Message m3 = MessageFromString(s3);

	MessageStream ms, c, s;

	printf("MessageStreamTest...\n");

	if (MessageStreamInit()) {
		printf("FAILED to initialise\n");
		exit(1);
	}

	/**************************************************************
	 * Write everything to a file then read everything back from the file.
	 **************************************************************/

	/* Remove test output files. */
	if (unlink("output1.dat") && errno != ENOENT && errno != 0) {
		printf("FAILURE %s:%d (%s)\n", MARKER, strerror(errno));
		exit(1);
	}

	printf("TEST %u (msg):    Write 3 messages then read 3 messages from a file.\r", ++testNumber);

	expectMessage(ms = MessageStreamOpen("file:output1.dat"), MARKER);

	success(MessageStreamWrite(ms, m1), MARKER);
	success(MessageStreamWrite(ms, m2), MARKER);
	success(MessageStreamWrite(ms, m3), MARKER);
	success(MessageStreamClose(ms), MARKER);

	expectMessage(ms = MessageStreamOpen("file:output1.dat"), MARKER);

	expectMessage(m = MessageStreamRead(ms), MARKER);
	sameMessage(m, m1, MARKER);
	free(m);

	expectMessage(m = MessageStreamRead(ms), MARKER);
	sameMessage(m, m2, MARKER);
	free(m);

	expectMessage(m = MessageStreamRead(ms), MARKER);
	sameMessage(m, m3, MARKER);
	free(m);

	success(MessageStreamClose(ms), MARKER);

	printf("TEST %u (msg): OK\n", testNumber);

	/**************************************************************
	 * Alternating write and reads from same stream on the same file.
	 **************************************************************/

	/* Remove test output files. */
	if (unlink("output2.dat") && errno != ENOENT && errno != 0) {
		printf("FAILURE %s:%d (%s)\n", MARKER, strerror(errno));
		exit(1);
	}

	printf("TEST %u (msg):    Alternate write & read from same stream & file.\r", ++testNumber);

	expectMessage(ms = MessageStreamOpen("file:output2.dat"), MARKER);

	success(MessageStreamWrite(ms, m1), MARKER);
	expectMessage(m = MessageStreamRead(ms), MARKER);
	sameMessage(m, m1, MARKER);

	success(MessageStreamWrite(ms, m2), MARKER);
	expectMessage(m = MessageStreamRead(ms), MARKER);
	sameMessage(m, m2, MARKER);

	success(MessageStreamWrite(ms, m3), MARKER);
	expectMessage(m = MessageStreamRead(ms), MARKER);
	sameMessage(m, m3, MARKER);

	success(MessageStreamClose(ms), MARKER);

	printf("TEST %u (msg): OK\n", testNumber);

	/**************************************************************
	 * Read from non-existant file should throw exception.
	 **************************************************************/

	printf("TEST %u (msg):    Read from non-existant file should throw exception.\r", ++testNumber);

	expectMessage(ms = MessageStreamOpen("file:noexist.dat"), MARKER);
	isNull(MessageStreamRead(ms), MARKER);

	printf("TEST %u (msg): OK\n", testNumber);

	/**************************************************************
	 * Open & close a server socket.
	 **************************************************************/

	printf("TEST %u (msg):    Open & close a server socket.\r", ++testNumber);

	expectMessage(s = MessageStreamOpen("listen://127.0.0.1:9090"), MARKER);
	success(MessageStreamClose(s), MARKER);

	printf("TEST %u (msg): OK\n", testNumber);

	/**************************************************************
	 * Open & close a client socket when no server present.
	 **************************************************************/

	printf("TEST %u (msg):    Open & close a client socket when no server present.\r", ++testNumber);

	expectMessage(c = MessageStreamOpen("connect://127.0.0.1:9090"), MARKER);
	success(MessageStreamClose(c), MARKER);

	printf("TEST %u (msg): OK\n", testNumber);

	/**************************************************************
	 * Open, write, close on client socket when no server.
	 **************************************************************/

	printf("TEST %u (msg):    Open, write, close on client socket when no server.\r", ++testNumber);

	expectMessage(c = MessageStreamOpen("connect://127.0.0.1:9090"), MARKER);

	failure(MessageStreamWrite(c, m1), MARKER);

	success(MessageStreamClose(c), MARKER);

	printf("TEST %u (msg): OK\n", testNumber);

	/**************************************************************
	 * Alternating write and reads to sockets.
	 **************************************************************/

	printf("TEST %u (msg):    Alternating write & reads to sockets.\r", ++testNumber);

	expectMessage(c = MessageStreamOpen("connect://127.0.0.1:9090"), MARKER);
	expectMessage(s = MessageStreamOpen("listen://127.0.0.1:9090"), MARKER);

	success(MessageStreamWrite(c, m1), MARKER);
	expectMessage(m = MessageStreamRead(s), MARKER);
	sameMessage(m, m1, MARKER);

	success(MessageStreamWrite(c, m2), MARKER);
	expectMessage(m = MessageStreamRead(s), MARKER);
	sameMessage(m, m2, MARKER);

	success(MessageStreamWrite(c, m3), MARKER);
	expectMessage(m = MessageStreamRead(s), MARKER);
	sameMessage(m, m3, MARKER);

	success(MessageStreamClose(c), MARKER);
	success(MessageStreamClose(s), MARKER);

	printf("TEST %u (msg): OK\n", testNumber);

	/**************************************************************
	 * Write everything to a socket then read everything back from the socket.
	 **************************************************************/

	printf("TEST %u (msg):    Write a few messages to a socket then read them back.\r", ++testNumber);

	expectMessage(c = MessageStreamOpen("connect://127.0.0.1:9090"), MARKER);
	expectMessage(s = MessageStreamOpen("listen://127.0.0.1:9090"), MARKER);

	success(MessageStreamWrite(c, m1), MARKER);
	success(MessageStreamWrite(c, m2), MARKER);
	success(MessageStreamWrite(c, m3), MARKER);

	expectMessage(m = MessageStreamRead(s), MARKER);
	sameMessage(m, m1, MARKER);

	expectMessage(m = MessageStreamRead(s), MARKER);
	sameMessage(m, m2, MARKER);

	expectMessage(m = MessageStreamRead(s), MARKER);
	sameMessage(m, m3, MARKER);

	success(MessageStreamClose(c), MARKER);
	success(MessageStreamClose(s), MARKER);

	printf("TEST %u (msg): OK\n", testNumber);

	/**************************************************************
	 * Listen on a socket using a timeout with no input ready.
	 **************************************************************/

	printf("TEST %u (msg):    Listen on a socket using a timeout with no input ready.\r", ++testNumber);

	expectMessage(s = MessageStreamOpen("listen://127.0.0.1:9090"), MARKER);

	MessageStreamSetTimeout(s, 3000);
	expectTimeout(MessageStreamRead(s), MARKER);

	success(MessageStreamClose(s), MARKER);

	printf("TEST %u (msg): OK\n", testNumber);

	/**************************************************************
	 * Listen on a socket using a timeout with input ready.
	 **************************************************************/

	printf("TEST %u (msg):    Listen on a socket using a timeout with input ready.\r", ++testNumber);

	expectMessage(s = MessageStreamOpen("listen://127.0.0.1:9090"), MARKER);
	expectMessage(c = MessageStreamOpen("connect://127.0.0.1:9090"), MARKER);

	MessageStreamSetTimeout(s, 3000);

	success(MessageStreamWrite(c, m1), MARKER);
	expectMessage(m = MessageStreamRead(s), MARKER);
	sameMessage(m, m1, MARKER);

	expectTimeout(MessageStreamRead(s), MARKER);

	success(MessageStreamClose(c), MARKER);
	success(MessageStreamClose(s), MARKER);

	printf("TEST %u (msg): OK\n", testNumber);


/**************************************************************
 ************** Test optimised interface for writing only text.
 **************************************************************/

	/**************************************************************
	 * Write 3 messages then read 3 messages from a file.
	 **************************************************************/

	/* Remove test output files. */
	if (unlink("output3.dat") && errno != ENOENT && errno != 0) {
		printf("FAILURE %s:%d (%s)\n", MARKER, strerror(errno));
		exit(1);
	}

	printf("TEST %u (text):    Write 3 messages then read 3 messages from a file\r", ++testNumber);

	expectMessage(ms = MessageStreamOpen("file-text:output3.dat"), MARKER);

	success(MessageStreamWrite(ms, s1), MARKER);
	success(MessageStreamWrite(ms, s2), MARKER);
	success(MessageStreamWrite(ms, s3), MARKER);
	success(MessageStreamClose(ms), MARKER);

	expectMessage(ms = MessageStreamOpen("file:output3.dat"), MARKER);

	expectMessage(m = MessageStreamRead(ms), MARKER);
	sameMessageString(m, s1, MARKER);
	free(m);

	expectMessage(m = MessageStreamRead(ms), MARKER);
	sameMessageString(m, s2, MARKER);
	free(m);

	expectMessage(m = MessageStreamRead(ms), MARKER);
	sameMessageString(m, s3, MARKER);
	free(m);

	success(MessageStreamClose(ms), MARKER);

	printf("TEST %u (text): OK\n", testNumber);

	/**************************************************************
	 * Alternate write & read from same stream & file.
	 **************************************************************/

	/* Remove test output files. */
	if (unlink("output4.dat") && errno != ENOENT && errno != 0) {
		printf("FAILURE %s:%d (%s)\n", MARKER, strerror(errno));
		exit(1);
	}

	printf("TEST %u (text):    Alternate write & read from same stream & file.\r", ++testNumber);

	expectMessage(ms = MessageStreamOpen("file-text:output4.dat"), MARKER);

	success(MessageStreamWrite(ms, s1), MARKER);
	expectMessage(m = MessageStreamRead(ms), MARKER);
	sameMessageString(m, s1, MARKER);

	success(MessageStreamWrite(ms, s2), MARKER);
	expectMessage(m = MessageStreamRead(ms), MARKER);
	sameMessageString(m, s2, MARKER);

	success(MessageStreamWrite(ms, s3), MARKER);
	expectMessage(m = MessageStreamRead(ms), MARKER);
	sameMessageString(m, s3, MARKER);

	success(MessageStreamClose(ms), MARKER);

	printf("TEST %u (text): OK\n", testNumber);

	/**************************************************************
	 * Read from non-existant file should throw exception.
	 **************************************************************/

	printf("TEST %u (text):    Read from non-existant file should throw exception.\r", ++testNumber);

	expectMessage(ms = MessageStreamOpen("file-text:noexist.dat"), MARKER);
	isNull(MessageStreamRead(ms), MARKER);

	printf("TEST %u (text): OK\n", testNumber);

	/**************************************************************
	 * Open & close a server socket.
	 **************************************************************/

	printf("TEST %u (text):    Open & close a server socket.\r", ++testNumber);

	expectMessage(s = MessageStreamOpen("listen://127.0.0.1:9090"), MARKER);
	success(MessageStreamClose(s), MARKER);

	printf("TEST %u (text): OK\n", testNumber);

	/**************************************************************
	 * Open & close a client socket when no server present.
	 **************************************************************/

	printf("TEST %u (text):    Open & close a client socket when no server present.\r", ++testNumber);

	expectMessage(c = MessageStreamOpen("connect-text://127.0.0.1:9090"), MARKER);
	success(MessageStreamClose(c), MARKER);

	printf("TEST %u (text): OK\n", testNumber);

	/**************************************************************
	 * Open, write, close on client socket when no server.
	 **************************************************************/

	printf("TEST %u (text):    Open, write, close on client socket when no server.\r", ++testNumber);

	expectMessage(c = MessageStreamOpen("connect-text://127.0.0.1:9090"), MARKER);

	failure(MessageStreamWrite(c, s1), MARKER);

	success(MessageStreamClose(c), MARKER);

	printf("TEST %u (text): OK\n", testNumber);

	/**************************************************************
	 * Alternating write and reads to sockets.
	 **************************************************************/

	printf("TEST %u (text):    Alternating write & reads to sockets.\r", ++testNumber);

	expectMessage(c = MessageStreamOpen("connect-text://127.0.0.1:9090"), MARKER);
	expectMessage(s = MessageStreamOpen("listen://127.0.0.1:9090"), MARKER);

	success(MessageStreamWrite(c, s1), MARKER);
	expectMessage(m = MessageStreamRead(s), MARKER);
	sameMessageString(m, s1, MARKER);

	success(MessageStreamWrite(c, s2), MARKER);
	expectMessage(m = MessageStreamRead(s), MARKER);
	sameMessageString(m, s2, MARKER);

	success(MessageStreamWrite(c, s3), MARKER);
	expectMessage(m = MessageStreamRead(s), MARKER);
	sameMessageString(m, s3, MARKER);

	success(MessageStreamClose(c), MARKER);
	success(MessageStreamClose(s), MARKER);

	printf("TEST %u (text): OK\n", testNumber);

	/**************************************************************
	 * Write everything to a socket then read everything back from the socket.
	 **************************************************************/

	printf("TEST %u (text):    Write a few messages to a socket then read them back.\r", ++testNumber);

	expectMessage(c = MessageStreamOpen("connect-text://127.0.0.1:9090"), MARKER);
	expectMessage(s = MessageStreamOpen("listen://127.0.0.1:9090"), MARKER);

	success(MessageStreamWrite(c, s1), MARKER);
	success(MessageStreamWrite(c, s2), MARKER);
	success(MessageStreamWrite(c, s3), MARKER);

	expectMessage(m = MessageStreamRead(s), MARKER);
	sameMessageString(m, s1, MARKER);

	expectMessage(m = MessageStreamRead(s), MARKER);
	sameMessageString(m, s2, MARKER);

	expectMessage(m = MessageStreamRead(s), MARKER);
	sameMessageString(m, s3, MARKER);

	success(MessageStreamClose(c), MARKER);
	success(MessageStreamClose(s), MARKER);

	printf("TEST %u (text): OK\n", testNumber);

	/**************************************************************
	 * Listen on a socket using a timeout with no input ready.
	 **************************************************************/

	printf("TEST %u (text):    Listen on a socket using a timeout with no input ready.\r", ++testNumber);

	expectMessage(s = MessageStreamOpen("listen://127.0.0.1:9090"), MARKER);

	MessageStreamSetTimeout(s, 3000);
	expectTimeout(MessageStreamRead(s), MARKER);

	success(MessageStreamClose(s), MARKER);

	printf("TEST %u (text): OK\n", testNumber);

	/**************************************************************
	 * Listen on a socket using a timeout with input ready.
	 **************************************************************/

	printf("TEST %u (text):    Listen on a socket using a timeout with input ready.\r", ++testNumber);

	expectMessage(s = MessageStreamOpen("listen://127.0.0.1:9090"), MARKER);
	expectMessage(c = MessageStreamOpen("connect-text://127.0.0.1:9090"), MARKER);

	MessageStreamSetTimeout(s, 3000);

	success(MessageStreamWrite(c, s1), MARKER);
	expectMessage(m = MessageStreamRead(s), MARKER);
	sameMessageString(m, s1, MARKER);

	expectTimeout(MessageStreamRead(s), MARKER);

	success(MessageStreamClose(c), MARKER);
	success(MessageStreamClose(s), MARKER);

	printf("TEST %u (text): OK\n", testNumber);


	return 0;
}
