/**
 * Socket.c
 *
 * Copyright 2001, 2004 by Anthony Howe. All rights reserved.
 */

#undef NDEBUG

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>

#include <com/snert/lib/version.h>
#include <com/snert/lib/io/Dns.h>
#include <com/snert/lib/io/Socket.h>

#if defined(__WIN32__)

# if defined(__BORLANDC__)
#  include <io.h>

extern unsigned int sleep(unsigned int);

# endif
# if defined(__VISUALC__)
#  define _WINSOCK2API_
# endif

# include <windows.h>
# include <winsock2.h>

# define UPDATE_ERRNO	if (errno == 0)	errno = WSAGetLastError()

# define EINPROGRESS	WSAEINPROGRESS
# define EISCONN        WSAEISCONN

# define USE_GETHOSTBYNAME

#include <com/snert/lib/io/Log.h>

#else /* a sensible OS */

# include <syslog.h>
# include <sys/time.h>
# include <netdb.h>
# include <arpa/inet.h>
# include <netinet/tcp.h>
# include <fcntl.h>
# include <unistd.h>

int
WSAGetLastError(void)
{
	return errno;
}

# define UPDATE_ERRNO

# ifndef INADDR_NONE
#  define INADDR_NONE	((in_addr_t) -1)
# endif

# if defined(__bsdi__) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__) || defined(__CYGWIN__)
#  define USE_GETHOSTBYNAME
# endif

#endif /* a sensible OS */

#define SHUTDOWN_RD		1
#define SHUTDOWN_WR		2
#define SHUTDOWN_RDWR		(SHUTDOWN_RD|SHUTDOWN_WR)

/***********************************************************************
 *** Global Variables, Constants, Macros
 ***********************************************************************/

#define GETHOSTBYNAME_BUFFER_SIZE	2048

struct socket {
	int fd;
	int error;
	int readFlags;
	int writeFlags;
	int onShutdown;
	int hasShutdown;
	InetAddress send;
	InetAddress recv;

	int isNonBlocking;
	int readOffset;
	int readLength;
	char readBuffer[BUFSIZ];
};

#define GET_FD(s)		(((struct socket *) s)->fd)

static long debug = 0;

/***********************************************************************
 *** Global Functions
 ***********************************************************************/

void
SocketSetDebugMask(int mask)
{
	debug = mask;
}

/*
 * Called before any SocketOpen() in the parent thread or process.
 */
int
SocketInit(void)
{
	static int SocketSupportLoaded = 0;

	if (!SocketSupportLoaded) {
#if defined(__WIN32__)
/*
 * WinSock 2.2 provides a subset of the BSD Sockets API.
 */
		int rc;
		WORD version;
		WSADATA wsaData;

		version = MAKEWORD(2, 2);
		if ((rc = WSAStartup(version, &wsaData)) != 0) {
			fprintf(stderr, "WSAStartup() failed: %d\n", rc);
			return -1;
		}

		if (HIBYTE( wsaData.wVersion ) < 2 || LOBYTE( wsaData.wVersion ) < 2) {
			fprintf(stderr, "WinSock API must be version 2.2 or better.\n");
			(void) WSACleanup();
			return -1;
		}

		if (atexit((AtExitFunction) WSACleanup)) {
			fprintf(stderr, "atexit(WSACleanup) failed: %s\n", strerror(errno));
			return -1;
		}
#endif
		if (DnsInit()) {
			fprintf(stderr, "DnsInit() failed: %s\n", strerror(errno));
			return -1;
		}

		SocketSupportLoaded = -1;
	}

	return 0;
}

/*
 * Convert ASCII IP address or host name into an in_addr structure.
 * This function is similar to inet_aton, which is not available on
 * all platforms. The difference here is that it will also do
 * hostname lookup, similar to gethostbyname(). 
 *
 * Return 0 on success otherwise -1 on error.
 */
int
SocketAddressToIP(const char *address, struct in_addr *ip)
{
	int rc;
	Dns dns;
	Vector hosts;
	DnsRecord host;
	
	rc = -1;
	
	if (ip == (struct in_addr *) 0)
		goto error0;
		
	/* Open DNS session using the system's name server list. */
	if ((dns = DnsOpen((const char **) 0)) == (Dns) 0)
		goto error0;
	
	if (address == NULL)
		address = "127.0.0.1";
	
	/* Get the address of the host name or IP address. This avoids
	 * gethostbyname(), which is not thread safe; and gethostbyname_r(),
	 * which appears to have problems on older Linux kernels or glibc
	 * libraries.
	 */
	if ((hosts = DnsGetAddresses(dns, address)) == NULL)
		goto error1;

	/* Fetch the first result only. */		
	if ((host = VectorGet(hosts, 0)) == (DnsRecord) 0)
		goto error2;

	/* Passback the IP address. */
	*ip = host->ip;

	rc = 0;	
error2:
	VectorDestroy(hosts);
error1:
	DnsClose(dns);
error0:
	return rc;
}

/***********************************************************************
 *** InetAddress Functions
 ***********************************************************************/

void
getIp4Octets(unsigned long ip, int *a, int *b, int *c, int *d)
{
	*a = (ip >> 24) & 0xff;
	*b = (ip >> 16) & 0xff;
	*c = (ip >> 8 ) & 0xff;
	*d = (ip      ) & 0xff;
}

char *
ip4ToString(unsigned long ip)
{
	char *s;
	int a, b, c, d;

	if ((s = malloc(16)) != (char *) 0) {
		getIp4Octets(ip, &a, &b, &c, &d);
		(void) snprintf(s, 16, "%u.%u.%u.%u", a,b,c,d);
	}

	return s;
}

#ifndef HAVE_INET_NTOP
const char *
inet_ntop(int af, const void *in_addr, char *buffer, size_t size)
{
	int a, b, c, d;

	getIp4Octets(ntohl(((struct in_addr *) in_addr)->s_addr), &a, &b, &c, &d);
	(void) snprintf(buffer, size, "%u.%u.%u.%u", a,b,c,d);

	return buffer;
}
#endif

int
ConvertIpToString(const void *ip, char *buffer, long length)
{
#if defined(HAS_INET_NTOP) && defined(AF_INET6)
{
	/* struct in6_addr *ip */
	return -(inet_ntop(AF_INET6, ip, buffer, length) == (char *) 0);
}
#else
{
	/* struct in_addr *ip */
	struct in_addr from;
	(void) memcpy(&from, (unsigned char *) ip, sizeof from);
	(void) strncpy(buffer, inet_ntoa(from), length);
	return 0;
}
#endif
}

static char *
InetAddressGetAddress(InetAddress *ip)
{
	char addr[50];
	
#if ! defined(__WIN32__)
	if (ip->domain == PF_UNIX)
		return strdup(ip->address.un.sun_path);
#endif

#ifdef PF_INET6
	if (ip->domain != PF_INET && ip->domain != PF_INET6)
#else
	if (ip->domain != PF_INET)
#endif
		return (char *) 0;
		
	if (ConvertIpToString(&ip->address.in.sin_addr, addr, sizeof addr))
		return (char *) 0;

	return strdup(addr);
}

static int
InetAddressGetPort(InetAddress *ip)
{
#ifdef PF_INET6
	if (ip->domain == PF_INET || ip->domain == PF_INET6)
#else
	if (ip->domain == PF_INET)
#endif
		return ntohs(ip->address.in.sin_port);

	return -1;			
}

#ifdef NOT_IMPLEMENTED_YET
static char *
InetAddressGetHostName(InetAddress *ip)
{
	/* TODO */
	
	return (char *) 0;
}
#endif

static void
InetAddressDestroyNothing(InetAddress *addr)
{
	/* Do nothing. */
}

static int
InetAddressInitFields(InetAddress *ip, const char *host, int port)
{
	if (ip == (InetAddress *) 0)
		return -1;

	if (host == (char *) 0)
		return -1;

	if (port < 0)
		return -1;

	/* Assume static memory structure that can't be freed. */
	ip->destroy = InetAddressDestroyNothing;
	ip->getAddress = InetAddressGetAddress;
	ip->getPort = InetAddressGetPort;
	
	/* Other fields. */
	ip->domain = PF_INET;
	ip->length = sizeof(ip->address.in);
	ip->address.in.sin_port = htons((unsigned short) port);
	ip->address.in.sin_family = AF_INET;

	return 0;
}

int
InetAddressInitIP(InetAddress *ip, const char *host, int port)
{
	if (InetAddressInitFields(ip, host, port))
		return -1;
		
#if defined(HAVE_INET_PTON)
	if (inet_pton(AF_INET, host, &ip->address.in.sin_addr) <= 0)
		return -1;
#elif defined(HAVE_INET_ATON)
	if (inet_aton(host, &ip->address.in.sin_addr) == 0)
		return -1;
#else
	if ((ip->address.in.sin_addr.s_addr = inet_addr(host)) == INADDR_NONE)
		return -1;
#endif
	return 0;
}

int
InetAddressInitHost(InetAddress *ip, const char *host, int port)
{
	if (InetAddressInitFields(ip, host, port))
		return -1;
		
#if ! defined(__WIN32__)
	/* This shouldn't probably be here, but its handy to think of a
	 * unix domain socket as an internet address.
	 */
	if (*host == '/') {
		ip->domain = PF_UNIX;
        	(void) strncpy(ip->address.un.sun_path, host, sizeof (ip->address.un.sun_path));
		ip->length = sizeof(ip->address.un);
		ip->address.un.sun_family = AF_UNIX;

		return 0;
	}
#endif
	if (SocketAddressToIP(host, &ip->address.in.sin_addr) != 0)
		return -1;

	return 0;
}

InetAddress *
InetAddressCreate(const char *host, int port)
{
	InetAddress *ip;
	
	if ((ip = calloc(1, sizeof *ip)) == (InetAddress *) 0)
		goto error0;
		
	if (InetAddressInitHost(ip, host, port))
		goto error1;

	/* Dynamic memory structure can be freed. */
	ip->destroy = (void (*)(InetAddress *)) free;
		
	return ip;
error1:
	free(ip);
error0:
	return (InetAddress *) 0;	
}

/***********************************************************************
 *** Socket Functions
 ***********************************************************************/

Socket
SocketOpen(InetAddress *addr, int type)
{
	struct socket *sock;

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "enter SocketOpen(%lx, %d)", (long) addr, type);
#endif
	errno = 0;

	if ((sock = calloc(1, sizeof *sock)) == (Socket) 0)
		goto error0;

	/* Copy contents of the send InetAddress into the recv InetAddress,
	 * since they will be in the same domain (PF_UNIX or PF_INET).
	 */
	sock->recv = *addr;
	sock->send = *addr;	
	sock->readOffset = 0;
	sock->readLength = 0;
	sock->isNonBlocking = 0;

	switch (type) {
	case SOCK_STREAM:
		/* By default close only one-half of the connection. */
		SocketSetShutdown(sock, SHUT_WR);
		break;
	case SOCK_DGRAM:
		/* You can't shutdown a UDP connection, duh. */
		SocketSetShutdown(sock, -1);
		break;
	default:
		errno = EINVAL;
		goto error1;
	}

	/* Get us the socket. */
	if ((sock->fd = socket(addr->domain, type, 0)) < 0) {
		UPDATE_ERRNO;
		goto error1;
	}
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_FD)
		syslog(LOG_DEBUG, "SocketOpen() socket() fd=%d", sock->fd);
#endif
#ifdef USE_IP_TOS
	(void) SocketSetTypeOfService(sock, IPTOS_LOWDELAY);
#endif
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "exit SocketOpen(%lx, %d) Socket=%lx fd=%d", (long) addr, type, (long) sock, sock->fd);
#endif
	return (Socket) sock;
error1:
	free(sock);
error0:
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "exit SocketOpen(%lx, %d) Socket=0 errno=%d", (long) addr, type, errno);
#endif
	return (Socket *) 0;
}

Socket
SocketFromFd(int fd)
{
	struct socket *sock;

	sock = calloc(1, sizeof *sock);
	if (sock == (Socket) 0)
		goto error0;

	if (sock->fd < 0)
		goto error1;

	sock->fd = fd;

	return (Socket) sock;
error1:
	free(sock);
error0:
	return (Socket *) 0;
}

Socket
SocketOpenUdp(const char *host, int port)
{
	struct socket *sock;
	struct inet_address addr;

	sock = (struct socket *) 0;
	
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "enter SocketOpenUdp(%s, %d)", host, port);
#endif
	if (InetAddressInitHost(&addr, host, port))
		goto error0;

	if ((sock = SocketOpen(&addr, SOCK_DGRAM)) == (Socket) 0)
		goto error0;
		
	(void) SocketSetLinger((Socket) sock, 1);
error0:
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "exit SocketOpenUdp(%s, %d) Socket=%lx", host, port, (long) sock);
#endif
	return sock;
}

Socket
SocketOpenTcpClient(const char *host, int port)
{
	struct socket *sock;
	struct inet_address addr;

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "enter SocketOpenTcpClient(%s, %d)", host, port);
#endif
	if (InetAddressInitHost(&addr, host, port))
		return (Socket) 0;

	if ((sock = SocketOpen(&addr, SOCK_STREAM)) == (Socket) 0)
		return (Socket) 0;

	/* Establish the client to server connection using the "liberal
	 * behaviour" technique for blocking connect() suggested by:
	 *
	 *  http://www.eleves.ens.fr:8080/home/madore/computers/connect-intr.html
	 */
	while (connect(sock->fd, (struct sockaddr *) &sock->send.address, sock->send.length) != 0) {
		(void) SocketGetError(sock);		
		if (errno == EISCONN)
			break;
		if (errno != EINTR) {
			SocketClose(sock);
			sock = (struct socket *) 0;
			break;
		}
		SocketClearError(sock);
	}

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "exit SocketOpenTcpClient(%s, %d) Socket=%lx", host, port, (long) sock);
#endif
	return (Socket) sock;
}

Socket
SocketOpenTcpClientWait(const char *host, int port, long timeout)
{
	int rc;
	struct socket *sock;
	struct inet_address addr;

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "enter SocketOpenTcpClientWait(%s, %d, %ld)", host, port, timeout);
#endif
	if (InetAddressInitHost(&addr, host, port))
		goto error0;

	if ((sock = SocketOpen(&addr, SOCK_STREAM)) == (Socket) 0)
		goto error0;

/* man connect
 *
 * EINPROGRESS
 * 	The  socket  is  non-blocking  and the connection cannot be com-
 * 	pleted immediately.  It is possible to select(2) or poll(2)  for
 * 	completion  by  selecting  the  socket for writing. After select
 * 	indicates writability, use getsockopt(2) to  read  the  SO_ERROR
 * 	option  at  level  SOL_SOCKET  to determine whether connect com-
 * 	pleted  successfully  (SO_ERROR  is  zero)   or   unsuccessfully
 * 	(SO_ERROR  is one of the usual error codes listed here, explain-
 * 	ing the reason for the failure).
 */
	rc = SocketSetNonBlocking(sock, 1);
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "result SocketSetNonBlocking(%d, 1) rc=%d", sock->fd, rc);
#endif
	if (rc != 0)
		goto error1;

	rc = connect(sock->fd, (struct sockaddr *) &sock->send.address, sock->send.length);

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "result connect(%d, ip, len) rc=%d errno=%d", sock->fd, rc, errno);
#endif
	if (rc != 0 && errno != 0 && errno != EINTR && errno != EINPROGRESS)
		goto error1;

	rc = SocketWaitForOutput(sock, timeout);
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "result SocketWaitForOutput(%d, %ld) rc=%d", sock->fd, timeout, rc);
#endif
	if (rc != 0)
		goto error1;

	rc = SocketSetNonBlocking(sock, 0);
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "result SocketSetNonBlocking(%d, 0) rc=%d", sock->fd, rc);
#endif
	if (rc != 0)
		goto error1;

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "exit SocketOpenTcpClientWait(%s, %d, %ld) Socket=%lx fd=%d", host, port, timeout, (long) sock, sock->fd);
#endif
	return (Socket) sock;
error1:
	(void) SocketGetError(sock);
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "closing socket %d on error", sock->fd);
#endif
	SocketClose(sock);
error0:
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "exit SocketOpenTcpClientWait(%s, %d, %ld) Socket=0", host, port, timeout);
#endif
	return (Socket) 0;
}

Socket
SocketOpenTcpServer(const char *host, int port, int nqueue)
{
	struct socket *sock;
	struct inet_address addr;

	if (InetAddressInitHost(&addr, host, port))
		goto error0;

	if ((sock = SocketOpen(&addr, SOCK_STREAM)) == (Socket) 0)
		goto error0;

	(void) SocketSetReuseAddr((Socket) sock, 1);

	/* Setup a server socket. Once the socket is open, its the
	 * responsibilty of the application to wait for connections
	 * by calling SocketAccept() or SocketWaitForInput().
	 *
	 * We cannot do that here since the SocketAccept() is often
	 * used within a loop to spawn threads or processes to handle
	 * the client connection.
	 */
	sock->send.address.in.sin_addr.s_addr = htonl(INADDR_ANY);

	if (bind(sock->fd, (struct sockaddr *) &sock->send.address, sock->send.length) < 0)
		goto error1;

	if (listen(sock->fd, nqueue) < 0)
		goto error1;

	return (Socket) sock;
error1:
	UPDATE_ERRNO;
	SocketClose(sock);
error0:
	return (Socket *) 0;
}

static int
getShutMask(int shut) 
{
	switch (shut) {
	case SHUT_RD:		return SHUTDOWN_RD;
	case SHUT_WR:		return SHUTDOWN_WR;
	case SHUT_RDWR:		return SHUTDOWN_RDWR;
	}
	
	return 0;
}

static int
getShutConstant(int mask) 
{
	switch (mask) {
	case SHUTDOWN_RD:	return SHUT_RD;
	case SHUTDOWN_WR:	return SHUT_WR;
	case SHUTDOWN_RDWR:	return SHUT_RDWR;
	}
	
	return -1;
}

void
SocketShutdownNow(Socket sock, int shut)
{
	int rc, bits;
	struct socket *s = (struct socket *) sock;
	
	/* Convert SHUT_ constant to a bit mask. */
	bits = getShutMask(shut);	
	
	/* Mask off what has been shutdown and convert what remains
	 * back into a SHUT_ constant.
	 */
	if ((shut = getShutConstant(bits & ~s->hasShutdown)) == -1)
		return;

	if ((rc = shutdown(s->fd, shut)) == 0)
		/* Remember what was shutdown. */
		s->hasShutdown |= bits;
	
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE) {
		UPDATE_ERRNO;
		syslog(LOG_DEBUG, "immediate socket shutdown(%d, %d) rc=%d errno=%d", GET_FD(sock), shut, rc, errno);
	}
#endif
}

void
SocketClose(Socket sock)
{
	int rc;
	struct socket *s = (struct socket *) sock;

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "enter SocketClose(%lx)", (long) s);
#endif
	if (s == (Socket) 0)
		goto error0;
	
	if (0 <= s->fd) {
		int shut, bits;
		
		/* Shutdown the socket in a clean manner only if there
		 * were no errors of any kind. The intent is to avoid
		 * performing any socket operation on a socket in an
		 * undefined state.
		 */

		/* Convert SHUT_ constant to a bit mask. */
		bits = getShutMask(s->onShutdown);	

		/* Mask off what has been shutdown and convert what remains
		 * back into a SHUT_ constant.
		 */
		shut = getShutConstant(bits & ~s->hasShutdown);
		
		if (SocketGetError(s) == 0 && 0 <= shut) {
			rc = shutdown(s->fd, shut);
#ifndef NDEBUG
			if (debug & SOCKET_DEBUG_OPEN_CLOSE) {
				UPDATE_ERRNO;
				syslog(LOG_DEBUG, "SocketClose() shutdown(%d, %d) rc=%d errno=%d", s->fd, s->onShutdown, rc, errno);
			}
#endif
		}

		/* Clear errno before the socket close operation. No point
		 * in SocketClearError(), since the structure goes away just
		 * below.
		 */
		errno = 0;

		/* This assumes that closing a socket is a safe operation
		 * whether in an error state or not.
		 */
#if defined(__WIN32__)
		rc = closesocket(s->fd);
#else
		rc = close(s->fd);
#endif
#ifndef NDEBUG
		if (debug & (SOCKET_DEBUG_FD|SOCKET_DEBUG_OPEN_CLOSE)) {
			UPDATE_ERRNO;
			syslog(LOG_DEBUG, "SocketClose() close(%d) rc=%d errno=%d", s->fd, rc, errno);
		}
#endif
	}

	free(s);
error0:
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "exit SocketClose(%lx)", (long) s);
#endif
	return;
}

char *
SocketGetAddress(Socket sock)
{
	if (sock == (Socket) 0)
		return (char *) 0;

	return ((struct socket *) sock)->send.getAddress(&((struct socket *) sock)->send);
}

char *
SocketGetSourceAddress(Socket sock)
{
	if (sock == (Socket) 0)
		return (char *) 0;

	return ((struct socket *) sock)->recv.getAddress(&((struct socket *) sock)->recv);
}

int
SocketGetPort(Socket sock)
{
	struct socket *s = (struct socket *)sock;

	if (s == (struct socket *) 0)
		return -1;

	return s->send.getPort(&s->send);
}

int
SocketGetFd(Socket sock)
{
	return sock == (Socket) 0 ? -1 : GET_FD(sock);
}

Socket
SocketAccept(Socket server)
{
	struct socket *c = calloc(1, sizeof *c);
	struct socket *s = (struct socket *) server;

	SocketClearError(server);

	if (c == (Socket) 0) {
		s->error = ENOMEM;
		goto error0;
	}

	/* Copy contents of the server socket InetAddress into the client
	 * InetAddress, since they will be in the same domain (PF_UNIX or
	 * PF_INET).
	 */
	*c = *s;

	if ((c->fd = accept(s->fd, (struct sockaddr *) &c->send.address, &c->send.length)) < 0)
		goto error1;

	if (SocketSetKeepAlive((Socket) c, 1) < 0)
		goto error1;

	return (Socket) c;
error1:
	SocketClose(c);
error0:
	return (Socket) 0;
}

static int
SocketWait(Socket sock, long milliseconds, int waitForInput)
{
	int rc;
	struct timeval timeout;
	fd_set fdset, *readset, *writeset;
	struct socket *s = (struct socket *) sock;

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_READ_WRITE)
		syslog(LOG_DEBUG, "enter SocketWait(%d, %ld, %d=%s) waiting...", s->fd, milliseconds, waitForInput, waitForInput ? "input" : "output");
#endif

	FD_ZERO(&fdset);
	FD_SET((unsigned) s->fd, &fdset);

	if (waitForInput) {
		readset = &fdset;
		writeset = (fd_set *) 0;

		/* Is there data already waiting in the read buffer? */
		if (s->readOffset < s->readLength)
			return 0;
	} else {
		readset = (fd_set *) 0;
		writeset = &fdset;
	}
	
	timeout.tv_sec = milliseconds / 1000L;
	timeout.tv_usec = (milliseconds % 1000L) * 1000L;

	for (;;) {
		SocketClearError(s);
		rc = select(
			s->fd + 1, readset, writeset, (fd_set *) 0,
			0 < milliseconds ? &timeout : (struct timeval *) 0
		);

#ifndef NDEBUG
		if (debug & SOCKET_DEBUG_READ_WRITE)
			syslog(LOG_DEBUG, "select(%d+1, %lx, %lx, ...) rc=%d errno=%d", s->fd, (long) readset, (long) writeset, rc, errno);
#endif
		if (0 < rc) {
			/* Its alive!. */
			break;
		}

		if (rc == 0) {
			/* Nothing interesting happened. Explicity set
			 * the error code since the man page does not
			 * state whether errno will be set ETIMEDOUT.
			 */
			s->error = ETIMEDOUT;
			break;
		}

		if (errno != EINTR) {
			/* "Oh no! Not again." said the pot of petunias. */
			break;
		}
	}

	(void) SocketGetError(s);
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_READ_WRITE)
		syslog(LOG_DEBUG, "exit SocketWait(%d, %ld, %d=%s) error=%d s->error=%d", s->fd, milliseconds, waitForInput, waitForInput ? "input" : "output", errno, s->error);
#endif
	return s->error;
}

/*
 * Wait for input on a given socket until input is ready or the time
 * to wait given in milliseconds, expires. A zero (0) wait time causes
 * the function to block indefinitely.
 *
 * Return 0 if input ready, else ETIMEDOUT for a timeout, otherwise
 * the value of errno.
 */
int
SocketWaitForInput(Socket sock, long milliseconds)
{
	return SocketWait(sock, milliseconds, 1);
}

int
SocketWaitForOutput(Socket sock, long milliseconds)
{
	return SocketWait(sock, milliseconds, 0);
}

#ifdef USE_IP_TOS
int
SocketSetTypeOfService(Socket sock, int tos)
{
	return setsockopt(GET_FD(sock), IPPROTO_IP, IP_TOS, (char *) &tos, sizeof tos);
}
#endif

int
SocketSetTcpNoDelay(Socket sock, int flag)
{
	return setsockopt(GET_FD(sock), IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof flag);
}

int
SocketSetBroadcast(Socket sock, int flag)
{
	return setsockopt(GET_FD(sock), SOL_SOCKET, SO_BROADCAST, (char *) &flag, sizeof flag);
}

int
SocketSetKeepAlive(Socket sock, int flag)
{
	return setsockopt(GET_FD(sock), SOL_SOCKET, SO_KEEPALIVE, (char *) &flag, sizeof flag);
}

int
SocketSetReuseAddr(Socket sock, int flag)
{
	return setsockopt(GET_FD(sock), SOL_SOCKET, SO_REUSEADDR, (char *) &flag, sizeof flag);
}

int
SocketSetLinger(Socket sock, int seconds)
{
	struct linger setlinger = { 1, 0 };

#if defined(__WIN32__)
	setlinger.l_onoff = (u_short) (0 < seconds);
	setlinger.l_linger = (u_short) seconds;
#else
	setlinger.l_onoff = 0 < seconds;
	setlinger.l_linger = seconds;
#endif

	return setsockopt(GET_FD(sock), SOL_SOCKET, SO_LINGER, (char *) &setlinger, sizeof setlinger);
}

void
SocketSetShutdown(Socket sock, int shutdown)
{
	((struct socket *) sock)->onShutdown = shutdown;
}

/*
 * Return zero (0) on success, otherwise -1 on error. Remember
 * that setting this affects both socket input and output.
 */
int
SocketSetNonBlocking(Socket sock, int flag)
{
	int rc;
	long flags;

#if defined(__WIN32__)
	flags = flag;
	rc = ioctlsocket(GET_FD(sock), FIONBIO, (unsigned long *) &flags);
#else
	flags = (long) fcntl(GET_FD(sock), F_GETFL);
# ifndef NDEBUG
	if (debug & SOCKET_DEBUG_GET_SET)
		syslog(LOG_DEBUG, "fcntl(%d, F_GETFL) flags=%lx errno=%d", GET_FD(sock), flags, errno);
# endif

	if (flag)
		flags |= O_NONBLOCK;
	else
		flags &= ~O_NONBLOCK;

	rc = fcntl(GET_FD(sock), F_SETFL, flags);
# ifndef NDEBUG
	if (debug & SOCKET_DEBUG_GET_SET)
		syslog(LOG_DEBUG, "fcntl(%d, F_SETFL, %lx) rc=%d errno=%d", GET_FD(sock), flags, rc, errno);
# endif
#endif
	if (rc != 0)
		return -1;
	
	/* Remember this state for SocketReadLine(). */
	((struct socket *) sock)->isNonBlocking = flag;

	return 0;
}

void
SocketSetWriteFlags(Socket sock, int flags)
{
	((struct socket *) sock)->writeFlags = flags;
}

void
SocketSetReadFlags(Socket sock, int flags)
{
	((struct socket *) sock)->readFlags = flags;
}

void
SocketClearError(Socket sock)
{
	errno = 0;
	((struct socket *) sock)->error = 0;
}

int
SocketGetError(Socket sock)
{
	struct socket *s = (struct socket *) sock;

	UPDATE_ERRNO;

	if (s->error == 0) {
		int so_error;
		int saved_error = errno;
		size_t length = sizeof so_error;

		/* getsockopt(SO_ERROR) resets the socket's copy of the error code,
		 * so we keep it around in our own structure until the next Socket*
		 * operation.
		 */
		if (getsockopt(s->fd, SOL_SOCKET, SO_ERROR, (void *) &so_error, &length) == 0)
			s->error = so_error;

		/* SocketClearError() clears both errno and so_error. So if
		 * so_error is zero and errno is not zero after a socket operation,
		 * then assume that errno is our socket error. This assumes that
		 * errno is part if the thread's context.
		 */
		if (s->error == 0 && saved_error != 0)
			s->error = saved_error;
	}

#ifndef NDEBUG
	/* There used to be enter/exit log lines, but thats just too much output. */
	if (debug & SOCKET_DEBUG_GET_SET)
		syslog(LOG_DEBUG, "SocketGetError(%lx) s->fd=%d errno=%d s->error=%d", (long) s, s->fd, errno, s->error);
#endif
	return s->error;
}

long
SocketWriteIp(Socket sock, void *buffer, long length, InetAddress *ip)
{
	long nbytes;
	struct socket *s = (struct socket *) sock;

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_READ_WRITE)
		syslog(LOG_DEBUG, "enter SocketWriteIp(%lx, %lx, %ld, %lx) s->fd=%d", (long) s, (long) buffer, length, (long) ip, s->fd);
#endif
	SocketClearError(sock);

	nbytes = sendto(
		s->fd, buffer, (size_t) length, s->writeFlags,
		(const struct sockaddr *) &ip->address, ip->length
	);

	if (nbytes < 0)
		(void) SocketGetError(sock);

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_READ_WRITE)
		syslog(LOG_DEBUG, "exit SocketWriteIp(%lx, %lx, %ld, %lx) s->fd=%d nbytes=%ld errno=%d so_error=%d", (long) s, (long) buffer, length, (long) ip, s->fd, nbytes, errno, s->error);
#endif
	return nbytes;
}

long
SocketWrite(Socket sock, void *buffer, long length)
{
	return SocketWriteIp(sock, buffer, length, &((struct socket *) sock)->send);
}

long
SocketRead(Socket sock, void *buffer, long length)
{
	long nbytes;
	struct socket *s = (struct socket *) sock;

	SocketClearError(sock);

	/* Do we have bytes still in the read buffer? */
	if (s->readOffset < s->readLength) {
		long readSize = s->readLength - s->readOffset;
		nbytes = readSize < length ? readSize : length;
		(void) memcpy(buffer, s->readBuffer + s->readOffset, nbytes);
		s->readOffset += nbytes;		
		return nbytes;
	}

	nbytes = recvfrom(
		s->fd, buffer, length, s->readFlags,
		(struct sockaddr *) &s->recv.address, &s->recv.length
	);
	if (nbytes < 0)
		(void) SocketGetError(sock);

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_READ_WRITE) {
		int lastbyte = 0 < nbytes ? ((unsigned char *) buffer)[nbytes-1] : -1;
		syslog(
			LOG_DEBUG, "SocketRead(%d, %lx, %ld) nbytes=%ld errno=%d so_error=%d lastbyte=%d %c",
			s->fd, (long) buffer, length, nbytes, errno, s->error,
			lastbyte, isprint(lastbyte) ? lastbyte : ' '
		);
	}
#endif

	return nbytes;
}

#ifdef SINGLE_BYTE_READ_VERSION
/*
 * Read in a line of at most size-1 bytes from the socket, stopping on a
 * newline (LF) or when the buffer is full. Return the number of bytes 
 * read; -1 for an error or -2 for EOF. The ASCII LF byte, and if 
 * immediately preceed by a ASCII CR, are always removed. The buffer is 
 * always '\0' terminated.
 */
long
SocketReadLine(Socket sock, char *line, long size)
{
	long i;
	long nbytes;
	struct socket *s = (struct socket *) sock;

	if (line == (char *) 0 || size <= 0)
		return -1;

	/* Leave room in the buffer for a terminating null byte. */
	line[--size] = '\0';

	for (i = 0; i < size; ++i) {
		/* Keep the buffer null terminated. */
		line[i] = '\0';

		nbytes = SocketRead(sock, &line[i], 1);
		if (nbytes < 0 && s->error != EINTR && s->error != EAGAIN) {
			/* Read error. */
			SocketClearError(s);
			return -1;
		}

		if (nbytes == 0) {
			/* Buffer underflow (EOF before newline) or EOF. */
			return 0 < i ? i : -2;
		}

		if (line[i] == '\n') {
			/* Newline found. */
			line[i] = '\0';
			if (0 < i && line[i-1] == '\r')
				line[--i] = '\0';

			return i;
		}
	}

	/* Buffer overflow. */
	return i;
}
#else
/*
 * Read in a line of at most size-1 bytes from the socket, stopping on a
 * newline (LF) or when the buffer is full. Return the number of bytes 
 * read; -1 for an error or -2 for EOF. The ASCII LF byte, and if 
 * immediately preceed by a ASCII CR, are always removed. The buffer is 
 * always '\0' terminated.
 */
long
SocketReadLine(Socket sock, char *line, long size)
{
	long i;
	struct socket *s = (struct socket *) sock;

	SocketClearError(s);

	if (line == (char *) 0 || size <= 0)
		return -1;

	/* Leave room in the buffer for a terminating null byte. */
	line[--size] = '\0';
	
	for (i = 0; i < size; ) {
		/* Keep the buffer null terminated in case we exit prematurely,
		 * because of an error or EOF condition. "Where ever you go
		 * there you are." - Buckaroo Bonzai
		 */
		line[i] = '\0';

		/* Get a new chunk of data off the wire? */
		if (s->readLength <= s->readOffset) {
			/* Non-blocking reads will be more efficient, trying
			 * to read more bytes at a time. When blocking reads
			 * are used, fallback on a byte at a time reads. I 
			 * choose not to set non-blocking mode at the top of
			 * the function, leaving it up to the application
			 * design to set it up before hand.
			 */
			s->readLength = SocketRead(sock, s->readBuffer, s->isNonBlocking ? sizeof s->readBuffer : 1);
			if (s->readLength < 0) {
				if (s->error == EAGAIN || s->error == EINTR) {
					/* Pause before trying again. */
					sleep(1);
					SocketClearError(s);
					continue;
				}

				/* Read error. */
				return -1;
			}
			
			if (s->readLength == 0) {
				/* Buffer underflow (EOF before newline) or EOF. */
				return 0 < i ? i : -2;				
			}
			
			s->readOffset = 0;
		}
		
		/* Copy from our read buffer into the line buffer. */
		line[i] = s->readBuffer[s->readOffset++];
		
		if (line[i] == '\n') {
			/* Newline found. */
			line[i] = '\0';
			if (0 < i && line[i-1] == '\r')
				line[--i] = '\0';

			return i;
		}
		
		/* This is not part of the loop control for a reason; avoids
		 * the need to decrement the index before the continue after
		 * EAGAIN in order to stay put.
		 */
		i++;
	}

	/* Buffer overflow before newline. */
	return i;
}
#endif

/*
 * Return zero (0) if the entire line was sent,
 * otherwise -1 and check SocketGetError();
 */
int
SocketPrint(Socket sock, char *msg)
{
	size_t length = strlen(msg);
	return SocketWrite(sock, msg, length) == (long) length ? 0 : -1;
}

/*
 * Return zero (0) if the entire line was sent,
 * otherwise -1 and check SocketGetError();
 */
int
SocketPrintLine(Socket sock, char *msg)
{
	if (SocketPrint(sock, msg))
		return -1;

	return SocketWrite(sock, "\r\n", 2) == 2 ? 0 : -1;
}


