/*
 * URI uri.c
 *
 * RFC 2396 - Uniform Resource Identifier (URI uri)
 *
 * Copyright 2001, 2004 by Anthony Howe. All rights reserved.
 */

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include <com/snert/lib/io/URI.h>
#include <com/snert/lib/util/Text.h>

/**
 * <p>
 * This class parses a Uniform Resource Identifier (URI) as per RFC 2396.
 * It is similar to java.net.URL, without the I/O methods. This class
 * is ment only to be a parser and informational value object (nothing
 * else).
 * </p>
 *
 * @see "RFC 2396"
 * @see java.net.URL
 */

void
uriFree(URI uri)
{
	if (uri != (URI) 0) {
		free(uri->uri);
		free(uri->scheme);
		free(uri->value);
		free(uri->userinfo);
		free(uri->host);
		free(uri->port);
		free(uri->path);
		free(uri->query);
		free(uri->fragment);
		free(uri);
	}
}

static int
indexof(const char *s, int offset, int ch)
{
	char *p = strchr(s + (offset < 0 ? 0 : offset), ch);
	return (p == (char *) 0 ? -1 : p - s);
}

static char *
substring(const char *s, int start, int stop)
{
	int len = stop - start;
	char *sub = malloc(len + 1);

	if (sub != (char *) 0) {
		(void) strncpy(sub, s + start, len);
		sub[len] = '\0';
	}

	return sub;
}

int
uriDecode(unsigned char *t, long tsize, unsigned char *s)
{
	int hex;
	unsigned char *tstop;

	if (t == NULL || tsize < 1 || s == NULL)
		return -1;
		
	for (tstop = t + tsize - 1; t < tstop && *s != '\0'; t++, s++) {
		switch (*s) {
		case '+':
			*t = ' ';
			break;
		case '%':
			if (sscanf(s+1, "%2x", &hex) == 1) {
				*t = (unsigned char) hex;
				s += 2;
			}
			break;
		default:
			*t = *s;
		}
	}
	
	*t = '\0';
	
	return 0;
}

/* Two common URI formats used:
 *
 *	<scheme>://[<userinfo>@]<host>[:<port>][/<path>][?<query>][#<fragment>]
 *	<scheme>:/<path>[?<query>][#<fragment>]
 */
URI
uriCreate(const char *s)
{
	URI uri;
	int len, colon, fragment, query;

	if (s == (const char *) 0)
		goto error0;

	uri = calloc(1, sizeof *uri);
	if (uri == (URI) 0)
		goto error0;

	uri->uri = (char *) strdup(s);
	if (uri->uri == (char *) 0)
		goto error1;

	len = strlen(s);

	/* <scheme>:<value> */
	colon = indexof(s, 0, ':');
	if (colon < 0)
		goto error1;

	uri->scheme = substring(s, 0, colon);
	if (!uriIsValidScheme(uri->scheme))
		goto error1;

	uri->value = substring(s, colon+1, len);
	if (uri->value == (char *) 0)
		goto error1;

	/* ...[#<fragment>] */
	len -= colon+1;
	fragment = indexof(uri->value, 0, '#');
	if (0 <= fragment && fragment < len) {
		uri->fragment = substring(uri->value, fragment+1, len);
		if (uri->fragment == (char *) 0)
			goto error1;
	} else {
		fragment = len;
	}

	/* ...[?<query>]... */
	query = indexof(uri->value, 0, '?');
	if (0 <= query && query < fragment) {
		uri->query = substring(uri->value, query+1, fragment);
		if (uri->query == (char *) 0)
			goto error1;
	} else {
		query = fragment;
	}

	if (uri->value[0] == '/' && uri->value[1] == '/') {
		/* ...//[<userinfo>@][<host>][:<port>][/<path>]... */
		int host, port, path;

		path = indexof(uri->value, 2, '/');
		if (2 <= path && path < query) {
			uri->path = substring(uri->value, path, query);
		} else {
			uri->path = (char *) strdup("/");
			path = query;
		}

		host = indexof(uri->value, 2, '@');
		if (2 <= host && host < path)
			uri->userinfo = substring(uri->value, 2, host);
		else
			host = 1;

		port = indexof(uri->value, host, ':');
		if (host < port && port < path) {
			uri->port = substring(uri->value, port + 1, path);
		} else {
			uri->port = (char *) 0;
			port = path;
		}

		uri->host = substring(uri->value, host + 1, port);

		/* Allow "file:///", disallow "http:///" */
		if (strcmp(uri->scheme, "file") != 0 && !uriIsValidHost(uri->host))
			goto error1;
	} else {
		/* .../<path>... */
		uri->path = substring(uri->value, 0, query);
	}

	if (uri->path == (char *) 0)
		goto error1;

	return uri;
error1:
	uriFree(uri);
error0:
	return (URI) 0;
}

/**
 * <p>
 * Validate a URI scheme syntax.
 * </p>
 *
 * @param scheme
 *	The scheme portion of a URI.
 *
 * @return
 *	True if the scheme is valid according to RFC 2396.
 *
 * @see "RFC 2396"
 */
int
uriIsValidScheme(const char *scheme)
{
	size_t i, len;

	if (scheme == (const char *) 0)
		return 0;

	len = strlen(scheme);

	if (len < 1)
		return 0;

	if (!isalpha(scheme[0]))
		return 0;

	for (i = 1; i < len; ++i) {
		char ch = scheme[i];
		switch (ch) {
		case '.': case '-': case '+':
			break;
		default:
			if (!isalnum(ch))
				return 0;
		}
	}

	return 1;
}

/**
 * <p>
 * Validate a URI host or IP syntax.
 * </p>
 *
 * @param host
 *	The host portion of a URI.
 *
 * @return
 *	True if the host is valid according to RFC 2396.
 *
 * @see "RFC 2396"
 */
int
uriIsValidHost(const char *host)
{
	if (uriIsValidDomain(host))
		return 1;

	return uriIsValidIP(host);
}

/**
 * <p>
 * Validate domain name syntax.
 * </p>
 *
 * @param domain
 *	The host domain names portion of a URI.
 *
 * @return
 *	True if the domain is valid according to RFC 2396.
 *
 * @see "RFC 2396"
 */
int
uriIsValidDomain(const char *domain)
{
	size_t len = strlen(domain);
	unsigned i, dot, lastCharAfterDot;

	if (domain == (const char *) 0 || len < 1)
		return 0;

	if (!isalpha(domain[0]))
		return 0;

	dot = 0;
	lastCharAfterDot = 'A';

	for (i = 1; i < len; ++i) {
		char ch = domain[i];

		if (dot) {
			lastCharAfterDot = ch;
			dot = 0;
		}

		switch (ch) {
		case '.':
			dot = 1;
			break;
		case '-': case '_':
			break;
		default:
			if (!isalnum(ch))
				return 0;
		}
	}

	return isalpha(lastCharAfterDot);
}

/**
 * <p>
 * Validate IP syntax.
 * </p>
 *
 * @param ip
 *	The host ip address portion of a URI.
 *
 * @return
 *	True if the IP is valid according to RFC 2396.
 *
 * @see "RFC 2396"
 */
int
uriIsValidIP(const char *ip)
{
	int dots;
	long octet;
	char *stop;

	if (ip == (const char *) 0 || *ip == '\0')
		return 0;

	for (dots = 0; *ip != '\0'; ip = stop) {
		octet = strtol(ip, &stop, 10);

		/* Must be an octet between 0..255. */
		if (ip == stop || octet < 0 || 255 < octet)
			return 0;

		/* Count the dot separators. */
		if (*stop == '.') {
			++stop;
			++dots;
		}
	}

	/* Minimum 3 dot separators to be at least an IP4 address. */
	return 3 <= dots;
}

/**
 * <p>
 * Test if this object equals another object.
 * </p>
 *
 * @return
 *	True if this object is the same as the given object.
 */
int
uriEquals(URI a, URI b)
{
	if (TextInsensitiveCompare(a->scheme, b->scheme) != 0)
		return 0;

	if (TextInsensitiveCompare(a->userinfo, b->userinfo) != 0)
		return 0;

	if (TextInsensitiveCompare(a->host, b->host) != 0)
		return 0;

	if (uriGetPort(a) != uriGetPort(b))
		return 0;

	if (TextSensitiveCompare(a->path, b->path) != 0)
		return 0;

	if (TextSensitiveCompare(a->query, b->query) != 0)
		return 0;

	if (TextSensitiveCompare(a->fragment, b->fragment) != 0)
		return 0;

	return 1;
}

/**
 * <p>
 * Computes a hash value for this URI.
 * </p>
 *
 * @return
 *	A hash value.
 */
long
uriHashCode(URI uri)
{
	return TextHash(0, uri->uri);
}

/**
 * <p>
 * Get the port number that may appear in a network URI. When no
 * port is explicitly specified, then return the default port for
 * know schemes.  If no port is specified or the default port is
 * not known, return -1.
 * </p>
 *
 * @return
 *	The port number or -1 if unknown.
 */
int
uriGetPort(URI uri)
{
	if (uri->port != (char *) 0 && *uri->port != '\0')
		return strtol(uri->port, (char **) 0, 10);

	/* Default protocol ports for convenience. */
	if (strcmp(uri->scheme, "ftp") == 0)
		return 21;

	if (strcmp(uri->scheme, "gopher") == 0)
		return 70;

	if (strcmp(uri->scheme, "http") == 0)
		return 80;

	if (strcmp(uri->scheme, "https") == 0)
		return 443;

	if (strcmp(uri->scheme, "imap") == 0)
		return 143;

	if (strcmp(uri->scheme, "mailto") == 0)
		return 25;

	if (strcmp(uri->scheme, "smtp") == 0)
		return 25;

	if (strcmp(uri->scheme, "nntp") == 0)
		return 119;

	if (strcmp(uri->scheme, "pop3") == 0)
		return 110;

	if (strcmp(uri->scheme, "telnet") == 0)
		return 23;

	if (strcmp(uri->scheme, "ssh") == 0)
		return 22;

	return -1;
}

#ifdef TEST

static void
dump(URI uri)
{
	printf("URI=[%s]\n", uri->uri);
	printf("  scheme\t[%s]\n", uri->scheme);
	printf("  value\t\t[%s]\n", uri->value);
	printf("  userinfo\t[%s]\n", uri->userinfo);
	printf("  host\t\t[%s]\n", uri->host);
	printf("  port\t\t[%d]\n", uriGetPort(uri));
	printf("  path\t\t[%s]\n", uri->path);
	printf("  query\t\t[%s]\n", uri->query);
	printf("  fragment\t[%s]\n", uri->fragment);
	printf("\n");
}

static void
success(const char *url)
{
	URI uri = uriCreate(url);
	if (uri == (URI) 0) {
		printf("FAILURE\n");
		exit(1);
	}
	dump(uri);
	uriFree(uri);
}

static void
failure(const char *url)
{
	URI uri = uriCreate(url);
	if (uri != (URI) 0) {
		dump(uri);
		uriFree(uri);
		printf("FAILURE\n");
		exit(1);
	}
}

int
main(int argc, char **argv)
{
	success("http:");
	success("http:*");

	success("http://host");
	success("http://127.0.0.1");
	success("http://1.2.127.0.0.1");
	success("http://abc.def.ghi");
	success("http://abc.def.ghi.");

	success("http://host:8080");
	success("http://userinfo@host:8080");
	success("http://userinfo:something@host");

	success("http://host/");
	success("http://host/path");
	success("http://host:8080/path");
	success("http://host/path/segments");
	success("http://host/path/segments?query");
	success("http://host/path/segments?query#fragment");

	success("http:/path/segment1/segment2?query");

	success("mailto:someone@domain.com");
	success("telnet://www.snert.lib.com");
	success("telnet://www.snert.lib.com:110");

	success("file:///");
	success("file:///C|/jdk1.3/docs/api/index.html");
	success("file:/a/b/c");

	failure(0);
	failure("");
	failure(":");
	failure("http:///");
	failure("http://userinfo@");
	failure("http://userinfo@:8080");
	failure("http://userinfo@:ff");
	failure("http://www.~.com");
	failure("http://www.foo.1com");
	failure("h^tp://localhost");
	failure("3pop://localhost");

	printf("OK\n");
}

#endif /* TEST */
