/*
 * MailSpan.c
 *
 * Assorted span functions for validation.
 *
 * Copyright 2002, 2004 by Anthony Howe.  All rights reserved.
 */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include <com/snert/lib/mail/MailSpan.h>

/*
 * RFC 2821 section 4.1.3 Address Literals
 *
 *	IPv4-address-literal = Snum 3("." Snum)
 *
 *	Snum = 1*3DIGIT         ; representing a decimal integer
 *				; value in the range 0 through 255
 */
long
MailSpanIPv4(const char *ip)
{
	int dots;
	long octet;
	const char *start, *stop;

	dots = 0;
	for (start = ip; *ip != '\0'; ip = stop) {
		octet = strtol(ip, (char **) &stop, 10);

		/* Must be an octet between 0..255. */
		if (ip == stop)
			break;

		if (octet < 0 || 255 < octet)
			return ip - start;

		/* Count the dot separators. */
		if (*stop == '.') {
			++stop;
			++dots;
		}
	}

	if (dots != 3)
		return 0;

	return ip - start;
}

long
MailSpanAddressLiteral(const char *s)
{
	const char *t = s;

	if (*t++ != '[')
		return 0;

	if (isdigit(*t)) {
		t += MailSpanIPv4(t);
	} else {
		/*** TODO IPv6 address forms. ***/
	}

	if (*t++ != ']')
		return 0;
	
	return t - s;
}

/*
 * RFC 2821 section 4.1.2 Command Argument Syntax
 *
 * Validate the characters and syntax.
 *
 *     Mailbox = Local-part "@" Domain
 *     Domain = (sub-domain 1*("." sub-domain)) / address-literal
 *     sub-domain = Let-dig [Ldh-str]
 *     Let-dig = ALPHA / DIGIT
 *     Ldh-str = *( ALPHA / DIGIT / "-" ) Let-dig
 */
long
MailSpanDomainName(const char *s, int minimumDots)
{
	int dots;
	const char *t;

	if (*s == '[')
		return MailSpanAddressLiteral(s);

	dots = 0;

	for (t = s; *t != '\0'; t++) {
		/* First charcter of a domain segment must be alpha-numeric. */
		if (!isalnum(*t))
			break;

		/* Rest of domain segment can be alpha-numeric or hypen. */
		for (t++; *t != '\0'; t++) {
			if (!isalnum(*t) && *t != '-')
				break;
		}

		/* Something other than the domain segment delimiter? */
		if (*t != '.')
			break;

		/* Last character of domain segment must be alpha-numeric. */
		if (t[-1] == '-') {
			t--;
			break;
		}

		dots++;
	}

	/* 0	simple machine name		local@machine
	 * 1	a domain name			local@domain.tld
	 * 2	fully qualified domain name	local@machine.domain.tld
	 */
	if (dots < minimumDots)
		return 0;

	return t - s;
}

/*
 * RFC 2821 section 4.1.2 Local-part and RFC 2822 section 3.2.4 Atom
 *
 * Validate only the characters.
 *
 *    Local-part = Dot-string / Quoted-string
 *    Dot-string = Atom *("." Atom)
 *    Atom = 1*atext
 *    Quoted-string = DQUOTE *qcontent DQUOTE
 */
long
MailSpanLocalPart(const char *s)
{
	const char *t;
	/*@-type@*/
	char x[2] = { 0, 0 };
	/*@-type@*/

	if (*s == '"') {
		/* Quoted-string = DQUOTE *qcontent DQUOTE */
		for (t = s+1; *t != '\0' && *t != '"'; t++) {
			switch (*t) {
			case '\\':		
				if (t[1] != '\0')
					t++;
				break;
			case '\t':
			case '\r':
			case '\n':
			case '#':
				return t - s;
			}
		}
		
		if (*t == '"')
			t++;
			
		return t - s;
	}

	/* Dot-string = Atom *("." Atom) */
	for (t = s; *t != '\0'; t++) {
		if (isalnum(*t))
			continue;

		/* atext does not include '.', but we do here to 
		 * simplify scanning dot-atom.
		 */
		*x = *t;
		if (strspn(x, "!#$%&'*+-/=?^_`{|}~.") != 1)
			break;
	}

	return t - s;
}

