/*
 * parsePath.c
 *
 * Copyright 2002, 2004 by Anthony Howe. All rights reserved.
 */

#define NDEBUG

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#include <stdlib.h>
#include <string.h>

#if defined(__unix__)
# include <syslog.h>
#endif

#include <com/snert/lib/util/Text.h>
#include <com/snert/lib/mail/MailSpan.h>
#include <com/snert/lib/mail/parsePath.h>

/***********************************************************************
 *** Global Variables & Constants
 ***********************************************************************/

#define TEST_RFC_2821_LENGTH_CONFORMANCE

#define DOMAIN_LENGTH		255
#define LOCAL_PART_LENGTH	128
#define EMAIL_LENGTH		(LOCAL_PART_LENGTH+1+DOMAIN_LENGTH)

/* NOTE that the LOCAL_PART_LENGTH in RFC 2821 is limited to 64
 * octets, but some stupid mailing list management software use 
 * Variable Envelope Return Paths (VERP) or the Sender Rewriting
 * Scheme (SRS), in order to encode another address within the
 * local-part of an address. If the address being encoded is 
 * already EMAIL_LENGTH octets in length, then the new encoded
 * address will NOT be conformant with RFC 2821 size limits for
 * the local-part.
 *
 * The LOCAL_PART_LENGTH here is doubled as a partial concesion
 * to these popular yet broken schemes. It appears to work for
 * majority of cases.
 */

/***********************************************************************
 *** Routines
 ***********************************************************************/

/*
 * Find the right inner most path. Return 0 for success, passing back
 * the pointer to the start of the path and the pointer just after the
 * end of the path (stop - start = length); otherwise return -1 for
 * imbalanced angle brackets.
 *
 *	<>				<>
 *	<<<>>>				<>
 *	<<<a>>>				<a>
 *	a <b> c				<b>
 *	<a <b> c>			<b>
 *	<a <b <c>>>			<c>
 *	<a> <b>				<b>
 *	<a <b> <c> d>			<c>
 *	<a <b> <c <d <e> f> g> h>	<e>
 *
 * Strange but valid:
 *
 *	<"<>"@plab.ku.dk>		<"<>"@plab.ku.dk>
 */
int
findInnerPath(const char *path, const char **start, const char **stop)
{
	int depth = 0;
	int quote = 0;

	*start = path;
	*stop = path;
	
	for ( ; *path != '\0'; path++) {
		switch (*path) {
		case '"':
			quote = !quote;		
			break;
		case '\\':
			if (quote && path[1] != '\0')
				path++;
			break;
		case '<':
			if (!quote) {
				depth++;
				*start = path;		
			}
			break;		
		case '>':
			if (!quote) {
				if (*stop <= *start)
					*stop = path+1;
				depth--;
			}
			break;		
		}
	}

	return -(depth != 0);
}

/*
 * Parse an envelope path into its component parts. Return NULL on success
 * and pass back an allocated ParsePath structure. This structure can later
 * be freed with a single call to free(). 
 *
 * Otherwise on error, return a pointer to an SMTP error reply message, in
 * which case the reply code should be 553. NULL will be passed back for the 
 * ParsePath structure pointer.
 */
const char *
parsePath(const char *path, int dots, ParsePath **out)
{
	ParsePath *p;	
	const char *start, *stop;
	int hasAtSign, hasPlusSign;

#if !defined(NDEBUG) && defined(__unix__)
	syslog(LOG_DEBUG, "enter parsePath(%s, %d, %lx)", path, dots, (long) out);
#endif

	/*@-mustdefine@*/	
	if (path == NULL || out == NULL)
		return "5.0.0 internal error: invalid arguments in parsePath()";
	/*@=mustdefine@*/	

	*out = NULL;
		
	/* Apparently there is some old mail server software that specifies:
	 *
	 *	MAIL FROM:<john smith <jsmith@domain>>
	 *
	 * Which is so non-conformant with RFC 2821. Since sendmail supports
	 * this archaic form, I've opted to support it for now on the grounds
	 * that the necessary information can still be found within.
	 */
	if (findInnerPath(path, &start, &stop))
		return "5.1.0 imbalanced angle brackets in path";

#if !defined(NDEBUG) && defined(__unix__)
	syslog(LOG_DEBUG, "*start=%c", *start);
#endif

	/* TEST that the path, as defined by the RFC 2821 grammar, begins
	 * and ends with angles brackets. Some mail servers relax this 
	 * requirement allowing for MAIL FROM:address or RCPT TO:address
	 * instead of MAIL FROM:<address> or RCPT TO:<address>.
	 */
	if (*start != '<' || (start < stop && stop[-1] != '>'))
		return "5.1.0 address does not conform to RFC 2821 syntax";

	/* Remove the angle brackets. */
	start++;
	stop--;
	
#if !defined(NDEBUG) && defined(__unix__)
	syslog(LOG_DEBUG, "call calloc(1, %lu)", sizeof(ParsePath) + (stop - start + 1) * 2);
#endif
	/* Allocate enough space for the structure and the string data.
	 * While more complex to setup, it allows for a single call to
	 * free() to release the structure and the strings it points to.
	 */
	if ((p = calloc(1, sizeof(ParsePath) + (stop - start + 1) * 2)) == (ParsePath *) 0)
		return "5.0.0 internal error: out of memory in parsePath()";	

#if !defined(NDEBUG) && defined(__unix__)
	syslog(LOG_DEBUG, "start equals stop=%d", start == stop);
#endif

	/* Check for the null address. */
	if (start == stop) {
		/*@-observertrans@*/
		p->address.string 	=
		p->localLeft.string 	=
		p->localRight.string 	=
		p->domain.string 	= "";
		/*@=observertrans@*/
		*out = p;
		return NULL;
	}

#if !defined(NDEBUG) && defined(__unix__)
	syslog(LOG_DEBUG, "Place a copy of the address string after the structure.");
#endif
	/* Place a copy of the address string after the structure. */
	p->address.length = stop - start;

#if !defined(NDEBUG) && defined(__unix__)
	syslog(LOG_DEBUG, "address-length=%ld", p->address.length);
#endif

	p->address.string = (char *)(&p[1]);

	/*@-modobserver -observertrans@*/
	(void) strncpy(p->address.string, start, (size_t) p->address.length);
	/*@=modobserver =observertrans@*/

#if !defined(NDEBUG) && defined(__unix__)
	syslog(LOG_DEBUG, "address-string='%s'", p->address.string, start);
#endif

/*	TextLower(p->address.string, 0, -1); */

	/* Place a second copy following the previous. */	
	p->localLeft.string = &p->address.string[p->address.length + 1];
	/*@-modobserver -observertrans */
	(void) strncpy(p->localLeft.string, p->address.string, (size_t) p->address.length);
	TextLower(p->localLeft.string, 0, -1);
	/*@=modobserver =observertrans@*/
	
#if !defined(NDEBUG) && defined(__unix__)
	syslog(LOG_DEBUG, "Split the local-part at a plus-sign or at-sign.");
#endif
	/* Split the local-part at a plus-sign or at-sign. */
	p->localLeft.length = (long) strcspn(p->localLeft.string, "+@");
	p->localRight.length = MailSpanLocalPart(p->localLeft.string);

#if !defined(NDEBUG) && defined(__unix__)
	syslog(LOG_DEBUG, "left-length=%ld right-length=%ld", p->localLeft.length, p->localRight.length);
#endif
#ifdef TEST_RFC_2821_LENGTH_CONFORMANCE
	if (LOCAL_PART_LENGTH < p->localRight.length) {
		free(p);
		return "5.1.0 local-part too long";
	}
#endif	
	hasPlusSign = p->localLeft.length < p->localRight.length;
	hasAtSign = p->localLeft.string[p->localRight.length] == '@';

	/* TEST for the presence of an at-sign or that the local part is 
	 * terminated at the null byte (we stripped the angle brackets
	 * above), which allows for <local> form.
	 */
	if (!hasAtSign && p->localLeft.string[p->localRight.length] != '\0') {
		/* An illegal character was found. */
		free(p);
		
		/* NOTE that dots == 0 implies a RCPT otherwise a MAIL. */
		return dots <= 0 ? "5.1.3 invalid local part" : "5.1.7 invalid local part";
	}

	p->domain.string = &p->localLeft.string[p->localRight.length + hasAtSign];	
	p->localRight.string = &p->localLeft.string[p->localLeft.length + hasPlusSign];	

	/* TEST that the domain name is within RFC 2821 limits and has
	 * at least the demanded number of dots. For example we would
	 * disallow MAIL FROM:<local>, but allow RCPT TO:<local>.
	 */
	p->domain.length = MailSpanDomainName(p->domain.string, dots);
/*	p->domain.length = p->address.length - p->localRight.length - hasAtSign; */

#if !defined(NDEBUG) && defined(__unix__)
	syslog(LOG_DEBUG, "domain-length=%ld", p->domain.length);
#endif
#ifdef TEST_RFC_2821_LENGTH_CONFORMANCE
	if (DOMAIN_LENGTH < p->domain.length) {
		free(p);
		return "5.1.0 domain name too long";
	}
#endif
	if (0 < dots && p->domain.length <= 0) {
		/* We want some dots in the domain after an @-sign. */
		free(p);
		return "5.1.7 address incomplete";
	}

	if (p->domain.string[p->domain.length] != '\0') {
		/* An illegal character was found. */
		free(p);
		
		/* NOTE that dots == 0 implies a RCPT otherwise a MAIL. */
		return dots <= 0 ? "5.1.3 invalid domain name" : "5.1.7 invalid domain name";
	}

	/* Adjust local-right-side length. */
	p->localRight.length -= p->localLeft.length + hasPlusSign;	

	/*@-modobserver -observertrans */
	/* Split the string into local-part and domain at the at-sign. */	
	p->localRight.string[p->localRight.length] = '\0';
	
	/* Split the string into local-left and local-right at the plus-sign if present. */		
	p->localLeft.string[p->localLeft.length] = '\0';
	/*@=modobserver =observertrans */

	/* Check for ip-as-domain surrounded by square brackets. */
	p->isDomainAnIP = (p->domain.string[0] == '[' && p->domain.string[p->domain.length - 1] == ']');
	
	if (p->isDomainAnIP) {
		/* Remove the square brackets from around the IP address. */
		p->domain.length -= 2;
		/*@-modobserver -observertrans */
		(void) memmove(p->domain.string, p->domain.string+1, (size_t) p->domain.length);
		p->domain.string[p->domain.length] = '\0';
		/*@=modobserver =observertrans */
	}

	*out = p;
	
#if !defined(NDEBUG) && defined(__unix__)
	syslog(LOG_DEBUG, "exit parsePath(%s, %d, %lx) error=NULL", path, dots, (long) out);
#endif
	
	return NULL;
}

/*
 * Given a path-format string, fill the given buffer. Return the number
 * of bytes copied excluding the null byte. The buffer is always null
 * terminated, unless -1 is returned to indicate an argument error.
 *
 * The format string comprises of literal character and percent-sign (%)
 * prefixed format characters:
 *
 *	%%		A literal percent-sign (%)
 *	%A		The whole address.
 *	%D		The domain name.
 *	%L		The left-hand-side of a plus-detailed 
 *			address or the entire local part if
 *			there was no plus-detail. See parsePath().
 *	%R		The right-hand-side of a plus detailed
 *			address or the empty string.
 *	%P		The entire local-part. Equivalent to
 *			"%L+%R" or "%L" depending if %R is
 *			empty or not.
 *
 * Some examples:
 *
 *	Source			Format		Result
 *
 *	foo@domain		%P@%D		foo@domain
 *	foo+bar@domain		%P@%D		foo+bar@domain
 *	foo+bar@domain		%L+bulk@%D	foo+bulk@domain
 *	foo+bar@domain		bulk+%L@%D	bulk+foo@domain
 *	foo+bar@domain		bulk+%R@%D	bulk+bar@domain
 *	foo+bar@domain		%P@bulk.%D	foo+bar@bulk.domain
 *	
 */
long
formatPath(char *buffer, long length, const char *fmt, ParsePath *p)
{
	long i;
	
	if (buffer == (char *) 0 || length <= 0 || fmt == (char *) 0 || p == (ParsePath *) 0)
		return -1;
	
	if (p->localLeft.string == (char *) 0)
		p->localLeft.string = "";
		
	if (p->localRight.string == (char *) 0)
		p->localRight.string = "";

	if (p->domain.string == (char *) 0)
		p->domain.string = "";
	
	/* Leave room for the null byte. */
	length--;
	
	for (i = 0; *fmt != '\0' && i < length; fmt++) {
		if (*fmt != '%') {
			/* Copy a literal character into the buffer.
			 */
			buffer[i++] = *fmt;
			continue;
		}

		/* Percent format character. */
		switch (*++fmt) {	
		case '%':
			/* A percent literal. */
			buffer[i++] = *fmt;
			continue;
		case 'A':
			i += TextCopy(buffer + i, length - i, p->address.string);
			break;			
		case 'P':
			/* Copy local-part into the buffer. If the 
			 * right-hand-string is not empty then this
			 * is equivalent to "%L+%R", else just "%L".
			 */
			i += TextCopy(buffer + i, length - i, p->localLeft.string);
			if (i+1 < length && *p->localRight.string != '\0') {
				buffer[i++] = '+';
				i += TextCopy(buffer + i, length - i, p->localRight.string);
			}
			break;
		case 'L':
			/* Copy the left-hand-string into the buffer.
			 * This is either the left-hand-side of a plus
			 * separated email address or the entire local-
			 * part.
			 */
			i += TextCopy(buffer + i, length - i, p->localLeft.string);
			break;
		case 'R':
			/* Copy the right-hand-string into the buffer.
			 * This is the right-hand-side of a plus 
			 * separated email address or the empty string.
			 */
			i += TextCopy(buffer + i, length - i, p->localRight.string);
			break;
		case 'D':
			/* Copy the domain name into the buffer.
			 */
			if (i+1 < length && p->isDomainAnIP)
				buffer[i++] = '[';
			
			i += TextCopy(buffer + i, length - i, p->domain.string);

			if (i+1 < length && p->isDomainAnIP)
				buffer[i++] = ']';
			break;		
		}		
	}
	
	buffer[i] = '\0';
	
	return i;
}

#ifdef TEST
#include <stdio.h>

int
main(int argc, char **argv) 
{
	int i;
	ParsePath *p;
	const char *error;
	static char buffer[1024];
	
	if (argc <= 1) {
		printf("usage: parsePath email-path ...\n");
		return 2;
	}
	
	for (i = 1; i < argc; i++) {
		error = parsePath(argv[i], 0, &p);
		if (error != (char *) 0) {
			printf("%s\n", error);
			continue;
		}

		printf(
			"address='%s' localleft='%s' localright='%s' domain='%s'\n",
			p->address.string, p->localLeft.string, p->localRight.string, p->domain.string
		);			

		if (strlen(p->address.string) != (size_t) p->address.length) {
			printf("strlen(p->address.string) != p->address.length\n");
			return 1;
		}
		
		if (strlen(p->localLeft.string) != (size_t) p->localLeft.length) {
			printf("strlen(p->localLeft.string) != p->localLeft.length\n");
			return 1;
		}

		if (strlen(p->localRight.string) != (size_t) p->localRight.length) {
			printf("strlen(p->localRight.string) != p->localRight.length\n");
			return 1;
		}

		if (strlen(p->domain.string) != (size_t) p->domain.length) {
			printf("strlen(p->domain.string) != p->domain.length\n");
			return 1;
		}

		formatPath(buffer, sizeof buffer, "<%A>\n", p);
		printf(buffer); 

		formatPath(buffer, sizeof buffer, "%P@%D\n", p);
		printf(buffer); 
		
		formatPath(buffer, sizeof buffer, "%L+bulk@%D\n", p);
		printf(buffer); 

		formatPath(buffer, sizeof buffer, "bulk+%L@%D\n", p);
		printf(buffer); 

		formatPath(buffer, sizeof buffer, "bulk%%%%+%R@%D\n", p);
		printf(buffer); 

		formatPath(buffer, sizeof buffer, "%P@bulk.%D\n", p);
		printf(buffer); 

		free(p);
	}

	return 0;
}
#endif
