/*
 * smf.c
 *
 * Sendmail Filter Support
 *
 * Copyright 2004 by Anthony Howe. All rights reserved.
 */

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#define _REENTRANT	1

#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/stat.h>

#include <com/snert/lib/version.h>
#include <com/snert/lib/io/Socket.h>
#include <com/snert/lib/type/Data.h>
#include <com/snert/lib/util/Text.h>
#include <com/snert/lib/util/setBitWord.h>
#include <com/snert/lib/mail/smdb.h>
#include <com/snert/lib/mail/smf.h>

#define	TAG_FORMAT		"%05d %s: "
#define	TAG_ARGS		work->cid, work->qid

/***********************************************************************
 *** Global Variables
 ***********************************************************************/

pthread_mutex_t smfMutex;

long smfLogDetail = SMF_LOG_INFO;

const char smfNoQueue[] = "NOQUEUE";
const char smfUndefined[] = "";
const char * const smfPrecedence[] = {
	"bulk", "junk", "list", "first-class", "special-delivery", NULL
};

/* const */ char smMacro_auth_authen[] = "{auth_authen}";
/* const */ char smMacro_client_addr[] = "{client_addr}";
/* const */ char smMacro_client_name[] = "{client_name}";
/* const */ char smMacro_client_resolv[] = "{client_resolv}";
/* const */ char smMacro_if_addr[] = "{if_addr}";
/* const */ char smMacro_if_name[] = "{if_name}";
/* const */ char smMacro_verify[] = "{verify}";

static smfInfo *smfDesc;
static volatile unsigned short smfCount;

/***********************************************************************
 *** Error Handling routines.
 ***********************************************************************/

static struct bitword logBits[] = {
	{ SMF_LOG_ALL,		"all" },
	{ SMF_LOG_INFO,		"info" },
	{ SMF_LOG_TRACE,	"trace" },
	{ SMF_LOG_PARSE,	"parse" },
	{ SMF_LOG_DEBUG,	"debug" },
	{ SMF_LOG_DIALOG,	"dialog" },
	{ SMF_LOG_STATE,	"state" },
	{ SMF_LOG_DNS,		"dns" },
	{ SMF_LOG_CACHE,	"cache" },
	{ SMF_LOG_DATABASE,	"db" },
	{ SMF_LOG_DATABASE,	"database" },
	{ SMF_LOG_SOCKET_FD,	"socket-fd" },
	{ SMF_LOG_SOCKET_ALL,	"socket-all" },
	{ SMF_LOG_LIBMILTER,	"libmilter" },
	{ 0, NULL }
};

void
smfSetLogDetail(const char *detail)
{
	smfLogDetail = setBitWord(logBits, detail);	
}

sfsistat
smfNullWorkspaceError(const char *where)
{
	/* Without the workspace we can't do a hell of a lot. */
	syslog(LOG_ERR, "%s(): internal error, null workspace", where);

	/* In the case of an internal error, we can reject the message or 
	 * accept it. Accepting it would by-pass the milter, but allow the
	 * message to be received and maybe filtered by something else.
	 */
	return SMFIS_ACCEPT;
}

sfsistat
smfReplyV(smfWork *work, int code, const char *ecode, const char *fmt, va_list args)
{
	sfsistat rc;
	char rcode[4], xcode[6];		

	work->replyLine[0] = '\0';
	
	if (SMTP_IS_PERM(code))
		rc = SMFIS_REJECT;
	else if (SMTP_IS_TEMP(code))
		rc = SMFIS_TEMPFAIL;
	else
		rc = SMFIS_CONTINUE;

	/* Make sure the return code is present and in an acceptable range. */
	if (!SMTP_IS_VALID(code))
		code = 550;

	/* Convert int value to 3-digit string for smfi_smfReply(). */
	rcode[3] = '\0';
	rcode[2] = '0' + (code / 1   % 10);
	rcode[1] = '0' + (code / 10  % 10);
	rcode[0] = '0' + (code / 100 % 10);

	if (ecode == (const char *) 0) {
		if (isdigit(fmt[0])) {
			ecode = strncpy(xcode, fmt, sizeof(xcode)-1);
			xcode[sizeof(xcode)-1] = '\0';
			fmt += sizeof(xcode);
		} else {
			ecode = strncpy(xcode, "x.7.1", sizeof xcode);
			xcode[0] = rcode[0];
		}
	}

	/* Build reply message. */
	(void) vsnprintf(work->replyLine, sizeof (work->replyLine), fmt, args);

	if (smfLogDetail & SMF_LOG_TRACE)
		syslog(LOG_DEBUG, TAG_FORMAT "reply %s %s %s", TAG_ARGS, rcode, ecode, work->replyLine);

	/* Tell sendmail what error to report to the sender. */
	if (smfi_setreply(work->ctx, rcode, (char *) ecode, work->replyLine) == MI_FAILURE)
		syslog(LOG_ERR, TAG_FORMAT "smfReplyV(): smfi_smfReply() failed: %s (%d)", TAG_ARGS, strerror(errno), errno);

	return rc;
}

sfsistat
smfReply(smfWork *work, int code, const char *ecode, const char *fmt, ...)
{
	sfsistat rc;

	va_list args;
	va_start(args, fmt);

	rc = smfReplyV(work, code, ecode, fmt, args);

	va_end(args);

	return rc;
}

/***********************************************************************
 *** Access Database White List Tests
 ***********************************************************************/

/*
 * Return one of SMDB_ACCESS_OK, SMDB_ACCESS_REJECT, or SMDB_ACCESS_UNKNOWN.
 */
int 
smfAccessHost(smfWork *work, const char *tag, const char *client_name, const char *client_addr, int loopDefault)
{
	int access;
	
	/* Check this once per connection instead of repeatly each
	 * time we access the DB. If the database changes after
	 * this, its not to critical (I hope), that we'll catch
	 * it the next time through.
	 */
	(void) smdbReload(smdbAccess);

#ifdef PREVIOUS
 	if (smfi_getsymval(work->ctx, smMacro_if_addr) == NULL) {	
#else
	if (client_addr == NULL || client_addr[0] == '\0') {
#endif	
		access = loopDefault;
	} else {
		/* BY-PASS or OVERRIDE
		 *
		 * Black and white lists IP handled through Sendmail's access
		 * database. The lookup order is:
		 *
		 *	tag:a.b.c.d
		 *	tag:a.b.c
		 *	tag:a.b
		 *	tag:a
		 *
		 * This Sendmail syntax limits the netmasks to /32, /24, /16, /8,
		 * which are the most common cases, but not so flexible as full
		 * range netmasks. Doing that would require a separate preloaded
		 * network/mask firewall-like list from another file since the
		 * access database syntax does not express this information well.
		 */
		if ((access = smdbAccessIp(smdbAccess, tag, client_addr, NULL, NULL)) != SMDB_ACCESS_NOT_FOUND)
			;

		else if ((access = smdbAccessDomain(smdbAccess, tag, client_name, NULL, NULL)) != SMDB_ACCESS_NOT_FOUND)
			;

		/* BY-PASS
		 *
		 * Black and white lists IP handled through Sendmail's access
		 * database. The lookup order is:
		 *
		 *	Connect:a.b.c.d
		 *	Connect:a.b.c
		 *	Connect:a.b
		 *	Connect:a
		 *
		 *	a.b.c.d
		 *	a.b.c
		 *	a.b
		 *	a
		 *
		 * This Sendmail syntax limits the netmasks to /32, /24, /16, /8,
		 * which are the most common cases, but not so flexible as full
		 * range netmasks. Doing that would require a separate preloaded
		 * network/mask firewall-like list from another file since the
		 * access database syntax does not express this information well.
		 */
		else if ((access = smdbAccessIp2(smdbAccess, "connect:", client_addr, NULL, NULL)) != SMDB_ACCESS_NOT_FOUND)
			;

		/* BY-PASS
		 *
		 * If sendmail successfully resolves the client connection (PTR-A-PTR 
		 * lookup) and there is a Connect:domain OK or RELAY entry, then white
		 * list the connection. We don't reject on this since Sendmail will act
		 * on it itself before calling the smf.
		 *
		 * If the client IP resolved and matched, then the lookup order is:
		 *
		 *	Connect:some.sub.domain.tld
		 *	Connect:sub.domain.tld
		 *	Connect:domain.tld
		 *	Connect:tld
		 *
		 *	some.sub.domain.tld
		 *	sub.domain.tld
		 *	domain.tld
		 *	tld	 
		 *
		 * If the client IP did not resolve or match, then the lookup order is:
		 *
		 *	Connect:[ip]
		 *	[ip]
		 */
		else if ((access = smdbAccessDomain2(smdbAccess, "connect:", client_name, NULL, NULL)) != SMDB_ACCESS_NOT_FOUND)
			;
	}

	if ((access = smdbAccessIsOk(access)) == SMDB_ACCESS_OK) {
		if (smfLogDetail & SMF_LOG_INFO)
			syslog(LOG_INFO, TAG_FORMAT "connection white listed, skipping", TAG_ARGS);
		work->skipConnection = 1;
	}

	return access;
}

/*
 * Return one of SMDB_ACCESS_OK, SMDB_ACCESS_REJECT, SMDB_ACCESS_UNKNOWN, or
 * SMDB_ACCESS_ERROR for a parse error, in which case the SMTP reponse will
 * also have been set.
 */
int 
smfAccessMail(smfWork *work, const char *tag, const char *mail, int dsnDefault)
{
	int access;
	ParsePath *path;
	char *auth_authen;
	const char *error;

	free(work->mail);
	work->mail = NULL;

	if ((error = parsePath(mail, 1, &path)) != NULL) {		
		(void) smfReply(work, 553, NULL, error);
		return SMDB_ACCESS_ERROR;
	}
 
	if (smfLogDetail & SMF_LOG_PARSE)
		syslog(
			LOG_DEBUG, TAG_FORMAT "address='%s' localleft='%s' localright='%s' domain='%s'",
			TAG_ARGS, path->address.string, path->localLeft.string,
			path->localRight.string, path->domain.string
		);

	if ((auth_authen = smfi_getsymval(work->ctx, smMacro_auth_authen)) != NULL)
		access = SMDB_ACCESS_OK;

	/* How to handle the DSN address. */
	else if (*path->address.string == '\0')
		access = dsnDefault;

	/* BY-PASS or OVERRIDE
	 *
	 * Lookup
	 *
	 *	tag:account@some.sub.domain.tld
	 *	tag:some.sub.domain.tld
	 *	tag:sub.domain.tld
	 *	tag:domain.tld
	 *	tag:tld
	 *	tag:account@
	 */
	else if ((access = smdbAccessMail(smdbAccess, tag, path->address.string, NULL, NULL)) != SMDB_ACCESS_NOT_FOUND)
		;

	/* BY-PASS
	 *
	 * Lookup
	 *
	 *	from:account@some.sub.domain.tld
	 *	from:some.sub.domain.tld
	 *	from:sub.domain.tld
	 *	from:domain.tld
	 *	from:tld
	 *	from:account@
	 *
	 *	account@some.sub.domain.tld
	 *	some.sub.domain.tld
	 *	sub.domain.tld
	 *	domain.tld
	 *	tld
	 *	account@
	 */
	else if ((access = smdbAccessMail2(smdbAccess, "from:", path->address.string, NULL, NULL)) != SMDB_ACCESS_NOT_FOUND)
		;

	/* This behaviour is similar to FEATURE(`delay_checks')
	 * in that MAIL white-listing can override connection 
	 * black-listing. However, MAIL black-listing will
	 * not override connection white-listing.
	 *
	 * For a black-listing or unknown result, continue.
	 */
	if ((access = smdbAccessIsOk(access)) == SMDB_ACCESS_OK) {
		if (smfLogDetail & SMF_LOG_INFO)
			syslog(
				LOG_INFO, TAG_FORMAT "sender <%s> %s, skipping", TAG_ARGS, path->address.string,
				auth_authen == NULL ? "white listed" : "authenticated" 
			);
		work->skipMessage = 1;
	}

	work->mail = path;
	
	return access;
}

/*
 * Return one of SMDB_ACCESS_OK, SMDB_ACCESS_REJECT, SMDB_ACCESS_UNKNOWN, or
 * SMDB_ACCESS_ERROR for a parse error, in which case the SMTP reponse will
 * also have been set.
 */
int 
smfAccessRcpt(smfWork *work, const char *tag, const char *rcpt)
{
	int access;
	ParsePath *path;
	const char *error;

	free(work->rcpt);
	work->rcpt = NULL;

	if ((error = parsePath(rcpt, 0, &path)) != NULL) {		
		(void) smfReply(work, 553, NULL, error);
		return SMDB_ACCESS_ERROR;
	}

	if (smfLogDetail & SMF_LOG_PARSE)
		syslog(
			LOG_DEBUG, TAG_FORMAT "address='%s' localleft='%s' localright='%s' domain='%s'",
			TAG_ARGS, path->address.string, path->localLeft.string,
			path->localRight.string, path->domain.string
		);

	/* BY-PASS or OVERRIDE
	 *
	 * Lookup
	 *
	 *	tag:account@some.sub.domain.tld
	 *	tag:some.sub.domain.tld
	 *	tag:sub.domain.tld
	 *	tag:domain.tld
	 *	tag:tld
	 *	tag:account@
	 */
	if ((access = smdbAccessMail(smdbAccess, tag, path->address.string, NULL, NULL)) != SMDB_ACCESS_NOT_FOUND)
		;

	/* BY-PASS
	 *
	 * Lookup
	 *
	 *	spam:account@some.sub.domain.tld	FRIEND
	 *	spam:some.sub.domain.tld		FRIEND
	 *	spam:sub.domain.tld			FRIEND
	 *	spam:domain.tld				FRIEND
	 *	spam:tld				FRIEND
	 *	spam:account@				FRIEND
	 */
	else if (smdbAccessMail(smdbAccess, "spam:", path->address.string, NULL, NULL) == SMDB_ACCESS_FRIEND)
		access = SMDB_ACCESS_OK;
	
	/* BY-PASS
	 *
	 * Lookup
	 *
	 *	to:account@some.sub.domain.tld
	 *	to:some.sub.domain.tld
	 *	to:sub.domain.tld
	 *	to:domain.tld
	 *	to:tld
	 *	to:account@
	 *
	 *	account@some.sub.domain.tld
	 *	some.sub.domain.tld
	 *	sub.domain.tld
	 *	domain.tld
	 *	tld
	 *	account@
	 */
	else if ((access = smdbAccessMail2(smdbAccess, "to:", path->address.string, NULL, NULL)) != SMDB_ACCESS_NOT_FOUND)
		;

	/* This behaviour is similar to FEATURE(`delay_checks') in 
	 * that RCPT white-listing can override MAIL or connection 
	 * black-listing. I think this provides for maximum possible
	 * B/W listing flexibilty. However, RCPT black-listing will
	 * not override MAIL or connection white-listing.
	 *
	 * For a black-listing or unknown result, the test proceeds.
	 */
	if ((access = smdbAccessIsOk(access)) == SMDB_ACCESS_OK) {
		if (smfLogDetail & SMF_LOG_INFO)
			syslog(LOG_INFO, TAG_FORMAT "recipient <%s> white listed, skipping", TAG_ARGS, path->address.string);
		work->skipMessage = 1;
	}

	work->rcpt = path;

	return access;
}

/***********************************************************************
 *** Support Routines
 ***********************************************************************/

int
smfHeaderSet(SMFICTX *ctx, char *field, char *value, int index, int present)
{
	if (present)
		return smfi_chgheader(ctx, field, index, value);
		
	return smfi_addheader(ctx, field, value);
}

unsigned short
smfOpenProlog(SMFICTX *ctx, char *client_name, _SOCK_ADDR *raw_client_addr, char *client_addr, long length)
{
	char *ipv6;	
	void *addr;
	unsigned short cid;
	
	client_addr[0] = '\0';
	
	if (pthread_mutex_lock(&smfMutex))
		syslog(LOG_ERR, "mutex lock in smfOpenProlog() failed: %s (%d)", strerror(errno), errno);	

	if (++smfCount == 0)
		smfCount = 1;	
	cid = smfCount;
		
	if (raw_client_addr != NULL) {		
		ipv6 = client_addr;
		addr = &((struct sockaddr_in *) raw_client_addr)->sin_addr;

#if defined(AF_INET6) && defined(INET6_ADDRSTRLEN)
		if (raw_client_addr->sa_family == AF_INET6) {
			addr = &((struct sockaddr_in6 *) raw_client_addr)->sin6_addr;
			ipv6 = &client_addr[sizeof ("IPv6")-1];
			strcpy(client_addr, "IPv6:");
		}
#endif
		if (inet_ntop(raw_client_addr->sa_family, addr, ipv6, length) == NULL)
			syslog(LOG_ERR, "inet_ntop() failed: %s (%d)", strerror(errno), errno);
	}

	if (pthread_mutex_unlock(&smfMutex))
		syslog(LOG_ERR, "mutex unlock in smfOpenProlog() failed: %s (%d)", strerror(errno), errno);	

	return cid;
}

unsigned short
smfCloseEpilog(smfWork *work)
{
	unsigned short cid = 0;
	
	if (work != NULL) {
		cid = work->cid;
		free(work->mail);
		free(work->rcpt);

		/* We have to do this, otherwise we get a warning. */
		(void) smfi_setpriv(work->ctx, NULL);
	}

	return cid;
}

void
smfAtExitCleanUp(void)
{
#ifdef NOT_OUR_JOB
/* Since we didn't open the sendmail access database, its not our
 * responsibility to close it.
 */
	smdbClose(smdbAccess);
	VectorDestroy(smdbAliases);
#endif
	(void) pthread_mutex_unlock(&smfMutex);
	(void) pthread_mutex_destroy(&smfMutex);

	/* Unlink the named socket ourselves since it appears that
	 * libsmf doesn't do this. See Sendmail 8.13.0 change log.	 
	 */
	if (strncmp(smfDesc->socket, "unix:", 5) == 0) 
		(void) unlink(smfDesc->socket + 5);
	else if (strncmp(smfDesc->socket, "local:", 6) == 0)
		(void) unlink(smfDesc->socket + 6);
		
	(void) unlink("pid");
}

void
smfSignalExit(int signum)
{
	(void) signal(signum, SIG_IGN);
	syslog(LOG_ERR, "signal %d received, program exit", signum);
	exit(0);
}

int
smfSetProcessOwner(smfInfo *smf)
{
	struct group *gr;
	struct passwd *pw;

	/* If we're running as root user and/or root group, we want to 
	 * change the process ownership to some benign user and group
	 * specified at compile time that has permissions to read the 
	 * sendmail.cf, access and aliases databases. There are 
	 * intentionally NO runtime options for this. This is a secuirty
	 * thang.
	 */
	if ((pw = getpwnam(smf->user)) == NULL) {
		syslog(LOG_ERR, "user \"%s\" not found", smf->user);
		return -1;
	}

	if ((gr = getgrnam(smf->group)) == NULL) {
		syslog(LOG_ERR, "group \"%s\" not found", smf->group);
		return -1;
	}

	/* Set the process ownership now to something else if we're root. */
	if (getuid() == 0 && pw->pw_uid != 0 && gr->gr_gid != 0) {
		(void) setgid(gr->gr_gid);
		(void) setuid(pw->pw_uid);
	}

	syslog(LOG_DEBUG, "process uid=%d gid=%d", getuid(), getgid());
	
	return 0;
}

/*
 * @global-out pthread_mutex_t smfMutex;
 *
 * @side-effects umask changed
 * @side-effects QUIT handler set
 * @side-effects process ownership changed
 * @side-effects working directory changed
 * @side-effects closes standard file pointers
 */
int
smfMainStart(smfInfo *smf, struct smfiDesc desc)
{
	FILE *fp;
	sigset_t set;

	smfDesc = smf;

	if (smfLogDetail & SMF_LOG_LIBMILTER)
		smfi_setdbg(100);

	if (signal(SIGQUIT, smfSignalExit) == SIG_ERR) {
		syslog(LOG_ERR, "failed to set SIGQUIT handler: %s (%d)", strerror(errno), errno);
		return 1;
	}

	(void) sigemptyset(&set);
	(void) sigaddset(&set, SIGQUIT);
	if (pthread_sigmask(SIG_BLOCK, &set, NULL)) {
		syslog(LOG_ERR, "failed to mask SIGQUIT from other threads: %s (%d)", strerror(errno), errno);
		return 1;
	}

	if (getuid() == 0)
		smfSetProcessOwner(smf);
	
	umask(S_IXUSR | S_IRWXG | S_IRWXO);

	/* Set the working directory to some place we should have the
	 * permissions to write to. Especially important if you want
	 * to get core dumps while debugging.
	 */
	if (chdir(smf->stateDir)) {
		syslog(LOG_ERR, "failed to set working directory to %s: %s (%d)", smf->stateDir, strerror(errno), errno);
		return 1;
	}

	if (pthread_mutex_init(&smfMutex, NULL)) {
		syslog(LOG_ERR, "failed to initialise mutex\n");
		return 1;
	}

	if (strncmp(smf->socket, "unix:", 5) == 0)
		(void) unlink(smf->socket + 5);
	else if (strncmp(smf->socket, "local:", 6) == 0)
		(void) unlink(smf->socket + 6);

	if (smfi_setconn((char *) smf->socket) == MI_FAILURE) {
		syslog(LOG_ERR, "smfi_setconn(%s) failed\n", smf->socket);
		return 1;
	}

	if (smfi_register(desc) == MI_FAILURE) {
		syslog(LOG_ERR, "smfi_register() failed\n");
		return 1;
	}

	/* Don't need these and they just use up resources. There are
	 * two schools of throught here: leave the standard files open,
	 * but redirected to /dev/null to avoid possible errors should
	 * the milter or C library read or write on standard I/O; the 
	 * other is to close them to free up the file descriptors, which
	 * may be required in an I/O intensive milter.
	 */
	switch (smf->standardIO) {
	case SMF_STDIO_CLOSE:
		fclose(stderr);
		fclose(stdout);
		fclose(stdin);
		break;
	case SMF_STDIO_IGNORE:
		(void) freopen("/dev/null", "a", stderr);
		(void) freopen("/dev/null", "a", stdout);
		(void) freopen("/dev/null", "r", stdin);
		break;
	}

	/* Create this in the state directory. */
	if ((fp = fopen("pid", "w")) == NULL) {
		syslog(LOG_WARNING, "failed to create pid file \"%s/pid\": %s (%d)", smf->stateDir, strerror(errno), errno);
	} else {	
		(void) fprintf(fp, "%d\n", getpid());
		(void) fclose(fp);
	}

	/* Write version information to log. */
	syslog(LOG_INFO, "%s %s", smf->version, smf->copyright);
	syslog(LOG_INFO, LIBSNERT_VERSION " " LIBSNERT_COPYRIGHT);

#if defined(SMFI_VERSION)
	syslog(LOG_INFO, "Sendmail libmilter Protocol version %d", SMFI_VERSION);
#endif

#ifdef LIBSNERT_WITH_BERKELEY_DB
# if DB_VERSION_MAJOR > 1
	syslog(LOG_INFO, "%s", db_version((int *) 0, (int *) 0,(int *) 0));
# else
	syslog(LOG_INFO, "Built with Berkeley DB");
# endif
#else 
	syslog(LOG_INFO, "Built without Berkeley DB support");
#endif
	return smfi_main();
}
