/*
 * Time.c
 *
 * Fine resolution clock.
 *
 * Copyright 2002, 2004 by Anthony Howe.  All rights reserved.
 */

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#define UNKNOWN_API		0
#define ANSI_API		1	/* time(), seconds 10^1 */
#define BSD_API			2	/* gettimeofday(), micro-seconds 10^-6 */
#define POSIX_API		3	/* clock_gettime(), nano-seconds 10^-9 */

#include <stdlib.h>
#include <sys/types.h>

#include <com/snert/lib/version.h>
#include <com/snert/lib/sys/Time.h>

#if ! defined(TIME_API) || TIME_API == unknown
#error "No fine resolution clock support defined."
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

Time *
TimeGetNow()
{
	Time *t;

	if ((t = malloc(sizeof *t)) != (Time *) 0)
		TimeSetNow(t);

	return t;
}

void
TimeSetNow(Time *t)
{
#if TIME_API == ANSI_API
	t->s = (unsigned long) time((time_t *) 0);
	t->ns = 0;
#endif
#if TIME_API == BSD_API
{
	struct timezone gmt = { 0, 0 };
	gettimeofday((struct timeval *) t, &gmt);
	t->ns *= UNIT_MILLI;
}
#endif
#if TIME_API == POSIX_API
	(void) clock_gettime(CLOCK_REALTIME, (struct timespec *) t);
#endif
}

Time *
TimeClone(Time *orig)
{
	Time *copy;

	if ((copy = malloc(sizeof *copy)) != (Time *) 0)
		*copy = *orig;

	return copy;
}

void
TimeDestroy(Time *tp)
{
	if (tp != (Time *) 0)
		free(tp);
}

void
TimeAdd(Time *acc, Time *b)
{
	if (acc != (Time *) 0 && b != (Time *) 0) {
		acc->s += b->s;
		acc->ns += b->ns;

		if (UNIT_NANO <= acc->ns) {
			acc->ns -= UNIT_NANO;
			acc->s++;
		}
	}
}

void
TimeSub(Time *acc, Time *b)
{
	if (acc != (Time *) 0 && b != (Time *) 0) {
		if (acc->ns < b->ns) {
			acc->ns += UNIT_NANO;
			acc->s--;
		}
		acc->ns -= b->ns;
		acc->s -= b->s;
	}
}

int
TimeIsZero(Time *acc)
{
	return acc->s == 0 && acc->ns == 0;
}

#if defined(HAVE_GMTIME) && ! defined(HAVE_GMTIME_R)
/***
 *** This is cover function and NOT thread safe.
 ***/

struct tm *
gmtime_r(const time_t *clock, struct tm *result)
{
	struct tm *static_storage;
	
	if ((static_storage = gmtime(clock)) == (struct tm *) 0)
		return (struct tm *) 0;
		
	*result = *static_storage;
	
	return result;
}
#endif

#if defined(HAVE_LOCALTIME) && ! defined(HAVE_LOCALTIME_R)
/***
 *** This is cover function and NOT thread safe.
 ***/

struct tm *
localtime_r(const time_t *clock, struct tm *result)
{
	struct tm *static_storage;
	
	if ((static_storage = localtime(clock)) == (struct tm *) 0)
		return (struct tm *) 0;
		
	*result = *static_storage;
	
	return result;
}
#endif

long
TimeStampAdd(char *buffer, long bufferLength)
{
	int offset;
	time_t now;
	long length;
	struct tm gmt;
	struct tm local;

	/* Build a Date header. */
	(void) time(&now);
	
	if (gmtime_r(&now, &gmt) == (struct tm *) 0)
		return 0;

	if (localtime_r((const time_t *) &now, &local) == (struct tm *) 0)
		return 0;

	length = strftime(buffer, bufferLength, "%a, %d %b %Y %H:%M:%S +0000", (const struct tm *) &local);

	/* Reformat the numerical time zone. */
	buffer += length - 5;

	/* Make sure the string was not truncated. */	
	if (length < 5|| *buffer != '+')
		return 0;

	/* Taken from sendmail/arpadate.c */
	offset = (local.tm_hour - gmt.tm_hour) * 60 + local.tm_min - gmt.tm_min;

	/* Assume that offset isn't more than a day. */
	if (local.tm_year < gmt.tm_year)
		offset -= 24 * 60;
	else if (local.tm_year > gmt.tm_year)
		offset += 24 * 60;
	else if (local.tm_yday < gmt.tm_yday)
		offset -= 24 * 60;
	else if (local.tm_yday > gmt.tm_yday)
		offset += 24 * 60;

	if (offset < 0) {
		offset = -offset;
		*buffer++ = '-';
	} else {
		*buffer++ = '+';
	}

	*buffer++ = (char)((offset / 600) + '0');
	*buffer++ = (char)((offset / 60) % 10 + '0');
	offset %= 60;
	*buffer++ = (char)((offset / 10) + '0');
	*buffer++ = (char)((offset % 10) + '0');
	
	return length;
}
