/*
 * Buf.h
 *
 * Variable length buffer.
 *
 * Copyright 2001, 2004 by Anthony Howe. All rights reserved.
 */

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include <com/snert/lib/io/posix.h>
#include <com/snert/lib/util/Buf.h>

#define BUF_GROWTH	100

struct buf {
	unsigned char *bytes;
	long capacity;
	long length;
};

static void
bounds(struct buf *a, long *offp, long *lenp)
{
	long off = *offp;
	long len = *lenp;

	if (off < 0)
		off = 0;
	else if (a->length <= off)
		off = a->length - 1;

	if (len < 0 || a->length < off + len)
		len = a->length - off;

	*offp = off;
	*lenp = len;
}

static int
enlarge(struct buf *a, long length)
{
	long capacity;
	unsigned char *bytes;

	if (a->capacity < length) {
		capacity = (length / BUF_GROWTH + 1) * BUF_GROWTH;
		if ((bytes = realloc(a->bytes, capacity)) == 0)
			return -1;

		a->capacity = capacity;
		a->bytes = bytes;
	}

	return 0;
}


Buf
BufCreate(long capacity)
{
	struct buf *a;

	if (capacity < 0)
		return (Buf) 0;

	a = malloc(sizeof *a);
	if (a == (struct buf *) 0)
		return (Buf) 0;

	a->bytes = malloc((size_t) capacity);
	if (a->bytes == (unsigned char *) 0) {
		free(a);
		return (Buf) 0;
	}

	a->capacity = capacity;
	a->length = 0;

	return (Buf) a;
}

Buf
BufFromBytes(unsigned char *bytes, long off, long len)
{
	struct buf *copy;

	if (len < 0)
		return (Buf) 0;

	copy = (struct buf *) BufCreate(len + BUF_GROWTH);
	if (copy == (struct buf *) 0)
		return (Buf) 0;

	memcpy(copy->bytes, bytes + off, len);
	copy->bytes[len] = 0;
	copy->length = len;

	return (Buf) copy;
}

Buf
BufFromBuf(Buf a, long off, long len)
{
	bounds((struct buf *) a, &off, &len);
	return BufFromBytes(((struct buf *) a)->bytes, off, len);
}

Buf
BufFromString(const char *s)
{
	return BufFromBytes((unsigned char *) s, 0, (long) strlen(s));
}

void
BufDestroy(Buf a)
{
	if (a != (Buf) 0) {
		free(((struct buf *) a)->bytes);
		free(a);
	}
}

unsigned char *
BufAsBytes(Buf a)
{
	unsigned char *bytes;

	bytes = ((struct buf *) a)->bytes;
	free(a);

	return bytes;
}

long
BufLength(Buf a)
{
	return ((struct buf *) a)->length;
}

void
BufSetLength(Buf a, long len)
{
	struct buf *aa = (struct buf *) a;

	if (len < 0)
		aa->length = 0;
	else if (!enlarge(aa, len + 1))
		aa->length = len;

	/* Keep the buffer null terminated. See BufAddByte(). */
	aa->bytes[aa->length] = 0;
}

long
BufCapacity(Buf a)
{
	return ((struct buf *) a)->capacity;
}

unsigned char *
BufBytes(Buf a)
{
	return ((struct buf *) a)->bytes;
}

int
BufCompareBuf(Buf a, long ax, Buf b, long bx, long len)
{
	struct buf *aa = (struct buf *) a;
	struct buf *bb = (struct buf *) b;

	if (len <= 0)
		return 0;

	if (aa == (struct buf *) 0 && bb == (struct buf *) 0)
		/* null == null */
		return 0;

	if (aa == (struct buf *) 0 && bb != (struct buf *) 0)
		/* null < b */
		return -1;

	if (aa != (struct buf *) 0 && bb == (struct buf *) 0)
		/* a > null */
		return 1;

	if (ax < 0)
		ax = 0;
	else if (aa->length < ax)
		ax = aa->length;

	if (bx < 0)
		bx = 0;
	else if (bb->length < bx)
		bx = bb->length;

	if (aa->length < ax + len && bx + len <= bb->length)
		/* Region A too short. */
		return -1;

	if (ax + len <= aa->length && bb->length < bx + len)
		/* Region B too short. */
		return 1;

	/* Region A and B within bounds, now compare them. */
	return memcmp(aa->bytes + ax, bb->bytes + bx, len);
}

int
BufCompare(Buf a, Buf b)
{
	struct buf *aa = (struct buf *) a;
	struct buf *bb = (struct buf *) b;

	if (aa->length != bb->length)
		return aa->length - bb->length;

	return BufCompareBuf(a, 0, b, 0, aa->length);
}

void
BufReverse(Buf a, long off, long len)
{
	unsigned char *x, *y, byte;

	bounds((struct buf *) a, &off, &len);

	x = y = ((struct buf *) a)->bytes + off;

	for (y += len; x < --y; ++x) {
		byte = *y;
		*y = *x;
		*x = byte;
	}

}

void
BufToLower(Buf a, long off, long len)
{
	unsigned char *x;

	bounds((struct buf *) a, &off, &len);

	for (x = ((struct buf *) a)->bytes + off; 0 < len; --len, ++x) {
		if (isupper(*x))
			*x = (unsigned char) tolower(*x);
	}
}

void
BufToUpper(Buf a, long off, long len)
{
	unsigned char *x;

	bounds((struct buf *) a, &off, &len);

	for (x = ((struct buf *) a)->bytes + off; 0 < len; --len, ++x) {
		if (islower(*x))
			*x = (unsigned char) toupper(*x);
	}
}

unsigned char
BufGetByte(Buf a, long off)
{
	long len = 1;
	bounds((struct buf *) a, &off, &len);
	return ((struct buf *) a)->bytes[off];
}

unsigned char *
BufGetBytes(Buf a, long off, long len)
{
	unsigned char *bytes;

	bounds((struct buf *) a, &off, &len);

	bytes = malloc((size_t) len);
	memcpy(bytes, ((struct buf *) a)->bytes + off, len);

	return bytes;
}

char *
BufToString(Buf a)
{
	char *string;
	struct buf *aa = (struct buf *) a;

	string = malloc(aa->length + 1);
	if (string != (char *) 0) {
		memcpy(string, aa->bytes, aa->length);
		string[aa->length] = '\0';
	}

	return string;
}

void
BufSetByte(Buf a, long off, unsigned char byte)
{
	long len = 1;

	bounds((struct buf *) a, &off, &len);
	((struct buf *) a)->bytes[off] = byte;
}

void
BufSetBytes(Buf a, long ax, unsigned char *b, long bx, long len)
{
	bounds((struct buf *) a, &ax, &len);
	memcpy(((struct buf *) a)->bytes + ax, b + bx, len);
}

int
BufInsertBytes(Buf a, long target, unsigned char *bytes, long source, long len)
{
	struct buf *aa = (struct buf *) a;

	if (len <= 0)
		return 0;

	/* Make sure we have enough room for the data and a null byte. */
	if (enlarge(aa, aa->length + len + 1) < 0)
		return -1;
		
	memmove(aa->bytes + target + len, aa->bytes + target, aa->length - target);

	memcpy(aa->bytes + target, bytes + source, len);
	aa->length += len;

	/* Keep the buffer null terminated. See BufAddByte(). */
	aa->bytes[aa->length] = 0;

	return 0;
}

int
BufAddByte(Buf a, int byte)
{
	struct buf *aa = (struct buf *) a;

	/* Make sure we have enough room for the byte and a null byte. */
	if (enlarge(aa, aa->length + 2) < 0)
		return -1;

	aa->bytes[aa->length++] = (unsigned char) byte;

	/* Keep the buffer null terminated so that we can use it with
	 * functions that work on C strings. The null byte is NOT part
	 * of the data nor does it affect the buffer length. Note though
	 * that if the data contains a null byte, then C string functions
	 * will have undefined results, but then you should know that.
	 */
	aa->bytes[aa->length] = 0;

	return 0;

}

int
BufAddBytes(Buf a, unsigned char *bytes, long off, long len)
{
	struct buf *aa = (struct buf *) a;

	if (len <= 0)
		return 0;

	/* Make sure we have enough room for the data and a null byte. */
	if (enlarge(aa, aa->length + len + 1) < 0)
		return -1;

	memcpy(aa->bytes + aa->length, bytes + off, len);
	aa->length += len;

	/* Keep the buffer null terminated. See BufAddByte(). */
	aa->bytes[aa->length] = 0;

	return 0;
}

int
BufAddString(Buf a, char *s)
{
	return BufAddBytes(a, s, 0, strlen(s));
}

int
BufAddBuf(Buf a, Buf b, long off, long len)
{
	bounds((struct buf *) b, &off, &len);
	return BufAddBytes(a, ((struct buf *) b)->bytes, off, len);
}

int
BufAddInputLine(Buf a, FILE *fp, long max)
{
	int byte;
	long start = BufLength(a);

	while (max < 0 || BufLength(a) < max) {
		if ((byte = fgetc(fp)) == EOF) {
			if (start == BufLength(a))
				/* EOF or error. */
				return -1;
			/* EOF before newline. */
			break;
		}

		if (byte == '\n') {
			struct buf *aa = (struct buf *) a;

			if (BufGetByte(a, BufLength(a)-1) == '\r')
				BufSetLength(a, BufLength(a)-1);

			/* Keep the buffer null terminated. See BufAddByte(). */
			aa->bytes[aa->length] = 0;

			break;
		}

		BufAddByte(a, byte);
	}

	return BufLength(a) - start;
}

int
BufAddReadLine(Buf a, int fd, long max)
{
	unsigned char byte;
	long start = BufLength(a);

	while (max < 0 || BufLength(a) < max) {
		if (read(fd, &byte, 1) != 1) {
			if (start == BufLength(a))
				/* EOF or error. */
				return -1;
			/* EOF before newline. */
			break;
		}

		if (byte == '\n') {
			struct buf *aa = (struct buf *) a;

			if (BufGetByte(a, BufLength(a)-1) == '\r')
				BufSetLength(a, BufLength(a)-1);

			/* Keep the buffer null terminated. See BufAddByte(). */
			aa->bytes[aa->length] = 0;

			break;
		}

		BufAddByte(a, byte);
	}

	return BufLength(a) - start;
}

