/*
 * DebugMemory.c
 *
 * Copyright 2003, 2004 by Anthony Howe.  All rights reserved.
 */

/*
 * Default debug memory pool size in kilobytes. Can be controlled
 * at runtime via an environment variable of the same name.
 */
#ifndef DEBUG_MALLOC_POOL_SIZE
#define DEBUG_MALLOC_POOL_SIZE		(2 * 1024)
#endif

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#define DEBUG_MALLOC_POOL_SIZE_LENGTH	(sizeof ("DEBUG_MALLOC_POOL_SIZE=") - 1)

#include <com/snert/lib/version.h>

#include <errno.h>
#include <stdlib.h>
#include <string.h>

extern char **environ;

#include <com/snert/lib/io/Log.h>
#include <com/snert/lib/io/posix.h>
#include <com/snert/lib/util/Memory.h>

#ifdef _REENTRANT
# if defined(HAVE_PTHREAD_MUTEX_INIT)
#  include <pthread.h>

# elif defined(__WIN32__)
#  include <windows.h>

typedef HANDLE pthread_mutex_t;
typedef void pthread_mutexattr_t;

#define PTHREAD_MUTEX_INITIALIZER 0

static int 
pthread_mutex_init(pthread_mutex_t *handle, const pthread_mutexattr_t *attr)
{
	if ((*(volatile pthread_mutex_t *) handle = CreateMutex(NULL, FALSE, NULL)) == (HANDLE) 0)
		return -1;
		
	return 0;
}

static int
pthread_mutex_lock(pthread_mutex_t *handle)
{
	if (*(volatile pthread_mutex_t *) handle == (HANDLE) 0) {
		if (pthread_mutex_init(handle, NULL))
			return -1;
	}

	switch (WaitForSingleObject(*(volatile pthread_mutex_t *)handle, INFINITE)) {
	case WAIT_TIMEOUT:
	case WAIT_OBJECT_0:
	case WAIT_ABANDONED:
		return 0;
	case WAIT_FAILED:
		return -1;
	}

	return 0;
}

static int
pthread_mutex_unlock(pthread_mutex_t *handle)
{
	if (*(volatile pthread_mutex_t *)handle == (HANDLE) 0 
	|| ReleaseMutex(*(volatile pthread_mutex_t *)handle) == 0)
		return -1;
			
	return 0;
}

static int
pthread_mutex_destroy(pthread_mutex_t *handle)
{
	if (*(volatile pthread_mutex_t *) handle != (HANDLE) 0)
		CloseHandle(*(volatile pthread_mutex_t *) handle);

	return 0;
}

# endif

static pthread_mutex_t memory_mutex = PTHREAD_MUTEX_INITIALIZER;
static int cleanup_registered; 

static void
DebugMallocAtExitCleanup(void)
{
	pthread_mutex_destroy((pthread_mutex_t *) &memory_mutex);
}

#endif /* _REENTRANT */

static void *sbrk_list[10];
static void *DebugMallocHeader;
static volatile int already_in_malloc;

void  
DebugFree(void *chunk, char *file, long line)
{	
	void **list, **stop;

#ifdef _REENTRANT
	if (pthread_mutex_lock(&memory_mutex)) {
		syslog(LOG_ERROR, "DebugFree(%lx): pthread_mutex_lock() failed at %s:%d", chunk, file, line);
		abort();
	}
#endif
	stop = &sbrk_list[sizeof sbrk_list/sizeof sbrk_list[0]];

	for (list = sbrk_list; list < stop; list++) {
		if (*list == chunk) {
			*list = (void *) 0;
			goto done;
		}
	}

	if (MemoryFree(DebugMallocHeader, chunk)) {
		syslog(LOG_ERROR, "DebugFree(%lx): MemoryFree() failed at %s:%d", chunk, file, line);
		abort();
	}
done:
#ifdef _REENTRANT
	if (pthread_mutex_unlock(&memory_mutex)) {
		syslog(LOG_ERROR, "DebugFree(%lx): pthread_mutex_unlock() failed at %s:%d", chunk, file, line);
		abort();
	}
#endif
}

void *
DebugMalloc(long size, char *file, long line)
{
	void *chunk;
	
	if (already_in_malloc) {
		/* Keep track of sbrk() pointers for when we try and free
		 * them. This space is considered lost, but accounted for
		 * since this space should only be allocated during the
		 * initial application start up and freed on exit.
		 */
		void **list, **stop;
		
		stop = &sbrk_list[sizeof sbrk_list/sizeof sbrk_list[0]];
		
		for (list = sbrk_list; list < stop; list++) {
			if (*list == (void *) 0)
				return *list = sbrk(size);
		}
		
		errno = ENOMEM;
		
		return (void *) 0;
	}

#ifdef _REENTRANT
	if (!cleanup_registered) {
		atexit(DebugMallocAtExitCleanup);
		cleanup_registered = 1;
	}

	if (pthread_mutex_lock(&memory_mutex)) {
		syslog(LOG_ERROR, "DebugMalloc(): pthread_mutex_lock() failed at %s:%d", file, line);
		abort();
	}
#endif	
	already_in_malloc = 1;

	if (DebugMallocHeader == (void *) 0) {
		char **env;
		long pool_size;

		/* We have to wait for environ to be initialised before we
		 * can search and allocate a prefered pool size.
		 */
		if (environ == (char **) 0) {
			chunk = DebugMalloc(size, file, line);
			goto done;
		}

		pool_size = DEBUG_MALLOC_POOL_SIZE;
		for (env = environ; *env != (char *) 0; env++) {
			if (strncmp(*env, "DEBUG_MALLOC_POOL_SIZE=", DEBUG_MALLOC_POOL_SIZE_LENGTH) == 0) {
				pool_size = strtol(*env+DEBUG_MALLOC_POOL_SIZE_LENGTH, (char **) 0, 10);
				if (pool_size <= 0)
					pool_size = DEBUG_MALLOC_POOL_SIZE;
				break;
			}
		}	

		DebugMallocHeader = MemoryCreate((void *) 0, pool_size * 1024);
		
		if (DebugMallocHeader == (void *) 0) {
			syslog(LOG_ERROR, "DebugMalloc(): MemoryCreate() failed at %s:%d", file, line);
			abort();
		}
	}
	
	chunk = MemoryAllocate(DebugMallocHeader, size);	
done:
	if (chunk == (void *) 0 && errno == EFAULT) {
		syslog(LOG_ERROR, "DebugMalloc(): MemoryVerifySize() failed at %s:%d", file, line);
		abort();
	}

	already_in_malloc = 0;
	
#ifdef _REENTRANT
	if (pthread_mutex_unlock(&memory_mutex)) {
		syslog(LOG_ERROR, "DebugMalloc(): pthread_mutex_unlock() failed at %s:%d", file, line);
		abort();
	}
#endif	
	return chunk;
}

void *
DebugCalloc(long n, long size, char *file, long line)
{
	void *chunk;
	
	if ((chunk = DebugMalloc(n * size, file, line)) != (void *) 0)
		memset(chunk, 0, n * size);
		
	return chunk;
}

void *
DebugRealloc(void *chunk, long size, char *file, long line)
{
	void *replace;
	
	if (DebugMallocHeader == (void *) 0)
		return DebugMalloc(size, file, line);

#ifdef _REENTRANT
	if (pthread_mutex_lock(&memory_mutex)) {
		syslog(LOG_ERROR, "DebugRealloc(): pthread_mutex_lock() failed at %s:%d", file, line);
		abort();
	}
#endif
	replace = MemoryReallocate(DebugMallocHeader, chunk, size);
	
	if (replace == (void *) 0 && errno == EFAULT) {
		syslog(LOG_ERROR, "DebugRealloc(): MemoryVerifySize() failed at %s:%d", file, line);
		abort();
	}

#ifdef _REENTRANT
	if (pthread_mutex_unlock(&memory_mutex)) {
		syslog(LOG_ERROR, "DebugRealloc(): pthread_mutex_unlock() failed at %s:%d", file, line);
		abort();
	}
#endif	
	return replace;
}


#ifndef DEBUG_MALLOC_TRACE_CALLER
void 
free(void *p)
{
	DebugFree(p, __FILE__, __LINE__);
}

void *
malloc(size_t n)
{
	return DebugMalloc((long) n, __FILE__, __LINE__);
}

void *
calloc(size_t m, size_t n)
{
	return DebugCalloc(m, (long) n, __FILE__, __LINE__);
}

void *
realloc(void *p, size_t n)
{
	return DebugRealloc(p, (long) n, __FILE__, __LINE__);
}
#endif
