/*
 * Text.c
 *
 * Copyright 2001, 2004 by Anthony Howe.  All rights reserved.
 */

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include <com/snert/lib/util/Buf.h>
#include <com/snert/lib/util/Text.h>
#include <com/snert/lib/io/posix.h>

typedef int (*TextCompareFunction)(const void *, const void *);
typedef long (*TextLongCompareFunction)(const void *, const void *);

/***********************************************************************
 *** 
 ***********************************************************************/

char *
TextDup(const char *orig)
{
	char *copy;

	copy = malloc(strlen(orig) + 1);
	if (copy == (char *) 0)
		return (char *) 0;

	return strcpy(copy, orig);
}

void
boundOffsetLength(long slength, long *offset, long *length)
{
	/* Correct offset if out of bounds. */
	if (*offset < 0)
		*offset = 0;
	else if (slength < *offset)
		*offset = slength;

	/* Correct length if out of bounds. */
	if (*length < 0)
		*length = slength;
	if (slength < *offset + *length)
		*length = slength - *offset;
}

char *
TextSubstring(const char *orig, long offset, long length)
{
	char *copy;

	if (orig == (const char *) 0)
		return (char *) 0;
	
	boundOffsetLength(strlen((char *) orig), &offset, &length);

	copy = malloc(length + 1);
	if (copy == (char *) 0)
		return (char *) 0;

	(void) strncpy(copy, orig+offset, length);
	copy[length] = '\0';
	
	return copy;
}

void
TextLower(char *s, long offset, long length)
{
	char *stop;

	if (s == (char *) 0)
		return;
	
	boundOffsetLength(strlen(s), &offset, &length);

	for (stop = s + offset + length; s < stop; s++) {
		*s = (char) tolower(*s);
	}
}

void
TextUpper(char *s, long offset, long length)
{
	char *stop;

	if (s == (char *) 0)
		return;
	
	boundOffsetLength(strlen(s), &offset, &length);

	for (stop = s + offset + length; s < stop; s++) {
		*s = (char) toupper(*s);
	}
}

void
TextReverse(char *s, long offset, long length)
{
	char ch, *x, *y;

	if (s == (char *) 0)
		return;
	
	boundOffsetLength(strlen(s), &offset, &length);

	/* Reverse segment of string. */
	for (x = y = s + offset, y += length; x < --y; ++x) {
		ch = *y;
		*y = *x;
		*x = ch;
	}
}

void
TextInvert(char *s, long offset, long length)
{
	char *stop;

	if (s == (char *) 0)
		return;
	
	boundOffsetLength(strlen(s), &offset, &length);

	for (stop = s + offset + length; s < stop; s++) {
		*s = (char) (isupper(*s) ? tolower(*s) : toupper(*s));
	}
}

#ifdef DROPPED
/**
 * <p>
 * Copy at most length bytes or until a null byte is seen from
 * the source C string to the target C string. The target string
 * will be null terminated if there is enough room for the null
 * byte.
 * </p>
 *
 * @param t
 *	The target string.
 *
 * @param s
 *	The source string.
 *
 * @param length
 *	The max. number of bytes to copy.
 *
 * @return
 *	The number of bytes actually copied, excluding the
 *	null terminator. This will be the length of s or
 *	length if the end of the target string buffer was
 *	reached.
 *
 * @see 
 *	memcpy(), snprintf(), strncpy()
 */
long
TextCopyN(char *t, char *s, long length)
{
	char *start;

	for (start = s; *s != '\0' && 0 < length--; )
		*t++ = *s++;

	if (0 < length)
		*t = '\0';

	return s - start;
}
#endif

/**
 * <p> 
 * Copy up to tlength - 1 octets from the null terminated C source string
 * into the target string buffer, null terminating the result. If tlength
 * is zero, then nothing was copied and the contents of the target string
 * buffer remain unaltered.
 * </p>
 *
 * @param t
 *	The target string buffer.
 *
 * @param tlength
 *	The physical length of target string buffer.
 *
 * @param s
 *	The source string.
 *
 * @return
 *	The number of bytes copied, excluding the terminating null.
 *
 * @see 
 *	memcpy(), snprintf(), strncpy(), strlcpy()
 */
long
TextCopy(char *t, long tlength, char *s)
{
	char *stop;
	
	if (tlength <= 0)
		return 0;

	for (stop = t + tlength - 1; t < stop && *s != '\0'; )
		*t++ = *s++;

	*t = '\0';
	
	return tlength - (stop - t) - 1 ;
}

/**
 * <p>
 * Compute the hash value of a string.
 * </p>
 *
 * @param hash
 *	A previously computed hash value.
 *
 * @param s
 *	A string to hash with the previous hash.
 *
 * @return
 *	A new hash value.
 */
long
TextHash(long hash, const char *s)
{
	while (*s != '\0')
		hash = 37 * hash + *s++;

	return hash;
}

/**
 * <p>
 * Is the string empty or blank?  This is equivalent to the
 * Perl regular expression:
 * </p>
 * <pre>
 *	m!^\s*!.
 * </pre>
 *
 * @param s
 *	A string to scan.
 *
 * @return
 *	True if the string is empty (zero length) or contains
 *	only whitespace.
 */
int
TextIsEmpty(const char *s)
{
	return s[strcspn(s, " \t\r\n\f")] == '\0';
}

/**
 * <p>
 * Does the string consist only of digits in the given radix?
 * </p>
 *
 * @param s
 *	A string to scan.
 *
 * @param radix
 *	The expected radix.
 *
 * @return
 *	True if the string conatins only characters for the given
 *	integer in the given radix.
 */
int
TextIsInteger(const char *s, int radix)
{
	char *digit;
	static char digits[] = "0123456789abcdefghijklmnopqrstuvwxyz";

	if (*s == '\0' || radix < 2 || 36 < radix)
		return 0;

	for ( ; *s != '\0'; ++s) {
		digit = strchr(digits, *s);
		if (digit == (char *) 0 || radix <= digit - digits)
			return 0;
	}

	return 1;
}

/**
 * <p>
 * Given the character following a backslash, return the
 * the character's ASCII escape value.
 * </p>
 * <pre>
 *   bell            \a	0x07
 *   backspace       \b	0x08
 *   escape          \e	0x1b
 *   formfeed        \f	0x0c
 *   linefeed        \n	0x0a
 *   return          \r	0x0d
 *   space           \s	0x20
 *   tab             \t	0x09
 *   vertical-tab    \v	0x0b
 * </pre>
 *
 * @param ch
 *	A character that followed a backslash.
 *
 * @return
 *	The ASCII value of the escape character or the character itself.
 */
int
TextBackslash(char ch)
{
	switch (ch) {
	case 'a': return '\007';
	case 'b': return '\010';
	case 'e': return '\033';
	case 'f': return '\014';
	case 'n': return '\012';
	case 'r': return '\015';
	case 's': return '\040';
	case 't': return '\011';
	case 'v': return '\013';
	}

	return ch;
}

/**
 * <p>
 * Parse the string for the next token. A token consists of characters
 * not found in the set of delimiters. It may contain backslash-escape
 * sequences, which shall be converted into literals or special ASCII
 * characters. It may contain single or double quoted strings, in which
 * case the quotes shall be removed, though any backslash escape
 * sequences within the quotes are left as is.
 * </p>
 *
 * @param string
 *	A quoted string.
 *
 * @param stop
 *	A pointer to a C string pointer. The point were the scan stops
 *	is passed back through this argument. This pointer can be NULL.
 *
 * @param delims
 *	A set of delimiter characters.
 * 
 * @param returnEmptyToken
 *	If false then a run of one or more delimeters is treated as a
 *	single delimeter separating tokens. Otherwise each delimeter
 *	separates a token that may be empty.
 *
 *	string		true		false
 *	-------------------------------------------
 *	[a,b,c]		[a] [b] [c]	[a] [b] [c]
 *	[a,,c]		[a] [] [c]	[a] [c]
 *	[a,,]		[a] [] [] 	[a]
 *	[,,]		[] [] []	(null)
 *	[]		[]		(null)
 *
 * @return
 *	An allocated token.
 *
 * @see #TextBackslash(char)
 */
char *
TextToken(const char *string, const char **stop, const char *delims, int returnEmptyToken)
{
	char *token, *t;
	const char *s;
	int quote = 0, escape = 0, length;

	if (string == (char *) 0) {
		if (stop != (const char **) 0)
			*stop = (char *) 0;
		return (char *) 0;
	}
	
	/* Skip leading delimiters? */
	if (!returnEmptyToken) {
		/* Find start of next token. */
		string += strspn(string, delims);
	
		if (*string == '\0') {
			if (stop != (const char **) 0)
				*stop = (char *) 0;
			return (char *) 0;
		}
	}
	
	/* Find end of token. */
	for (s = string; *s != '\0'; ++s) {
		if (escape) {
			escape = 0;
			continue;
		}

		switch (*s) {
		case '"': case '\'':
			quote = *s == quote ? 0 : *s;
			continue;
		case '\\':
			escape = 1;
			if (quote == 0) continue;
			continue;
		}

		if (quote == 0 && strchr(delims, *s) != (char *) 0)
			break;
	}

	token = malloc((s - string) + 1);
	if (token == (char *) 0)
		return (char *) 0;

	/* Copy token, removing quotes and backslashes. */
	for (t = token; string < s; ++string) {
		if (escape) {
			if (quote == 0)
				*t++ = (char) TextBackslash(*string);
			else
				*t++ = *string;
			escape = 0;
			continue;
		}

		switch (*string) {
		case '"': case '\'':
			quote = *string == quote ? 0 : *string;
			continue;
		case '\\':
			escape = 1;
			if (quote == 0) continue;
			break;
		}

		if (quote == 0 && strchr(delims, *string) != (char *) 0)
			break;

		*t++ = *string;
	}
	*t = '\0';

	if (*s == '\0') {
		/* Token found and end of string reached. 
		 * Next iteration should return no token.
		 */
		s = (char *) 0;
	} else {
		length = strspn(s, delims);
		if (returnEmptyToken) {
			/* Consume only a single delimter. */
			s += length <= 0 ? 0 : 1;
		} else {
			/* Consume one or more delimeters. */
			s += length;
		}
	}
	
	if (stop != (const char **) 0)
		*stop = s;

	return token;
}

/**
 * <p>
 * The given string contains a list of substrings separated by the
 * specified delimiter characters. The substrings may contain quoted
 * strings and/or contain backslash-escaped characters. The common
 * backslash escape sequences are supported and return their ASCII
 * values.
 * </p>
 *
 * @param string
 *	A list represented as a string.
 *
 * @param delims
 *	A set of delimiter characters.
 *
 * @param returnEmptyTokens
 *	If false then a run of one or more delimeters is treated as a
 *	single delimeter separating tokens. Otherwise each delimeter
 *	separates a token that may be empty.
 *
 *	string		true		false
 *	-------------------------------------------
 *	[a,b,c]		[a] [b] [c]	[a] [b] [c]
 *	[a,,c]		[a] [] [c]	[a] [c]
 *	[a,,]		[a] [] [] 	[a]
 *	[,,]		[] [] []	(empty vector)
 *	[]		[]		(empty vector)
 *
 * @return
 *	A vector of C strings.
 */
Vector 
TextSplit(const char *string, const char *delims, int returnEmptyTokens)
{
	char *token;
	Vector list;
	
	if ((list = VectorCreate(5)) == NULL)
		return NULL;

	VectorSetDestroyEntry(list, free);

	while ((token = TextToken(string, &string, delims, returnEmptyTokens)) != (char *) 0)
		VectorAdd(list, token);

	return list;
}

/**
 * <p>
 * Join a vector of strings into a string with the given delimiter.
 * </p>
 *
 * @param delim
 *	A delimiter of one or more characters.
 *
 * @param strings
 *	A list of strings to be joined.
 *
 * @return
 *	A string comprised of all the objects in string representation
 *	each separated by the string delimiter.
 */
char *
TextJoin(const char *delim, Vector strings)
{
	long i;
	char *s;
	Buf collector = BufCreate(100);

	for (i = 0; i < VectorLength(strings); ++i) {
		BufAddString(collector, VectorGet(strings, i));
		BufAddString(collector, (char *) delim);
	}

	BufSetLength(collector, BufLength(collector) - strlen(delim));
	s = BufToString(collector);
	BufDestroy(collector);

	return s;
}

/**
 * <p>
 * Expand all the tab characters into spaces. Tabstops are assumed
 * to be every 8th column.
 * </p>
 *
 * @param s
 *	A string buffer containing tab characters to be expanded.
 *
 * @param col
 *	Initial column position
 *
 * @return
 * 	A string buffer where all tab characters have been converted
 *	to spaces.
 */
Buf
TextExpand(Buf s, long col)
{
	long i;
	unsigned char byte;
	Buf t = BufCreate(100);

	for (i = 0; i < BufLength(s); ++i, ++col) {
		byte = BufGetByte(s, i);
		if (byte == '\t') {
			do {
				BufAddByte(t, ' ');
			} while ((++col & 7) != 0);
		} else {
			BufAddByte(t, byte);
		}
	}

	return t;
}

/**
 * <p>
 * Convert a byte array into a hex encoded string.
 * </p>
 *
 * @param b
 *	A byte array.
 *
 * @return
 *	A string representing the byte array as hexdecimal values.
 */
char *
TextHexEncode(Buf b)
{
	long i;
	Buf s = BufCreate(BufLength(b) * 2 + 1);
	static char xdigits[] = "0123456789ABCDEF";

	for (i = 0; i < BufLength(b); ++i) {
		unsigned char a = BufGetByte(b, i);
		BufAddByte(s, xdigits[(a >> 4) & 0xf]);
		BufAddByte(s, xdigits[a & 0xf]);
	}

	BufAddByte(s, '\0');

	return (char *) BufAsBytes(s);
}

int
TextInsensitiveCompareN(const void *xp, void const *yp, long len)
{
	int diff;
	char *x = (char *) xp;
	char *y = (char *) yp;

	if (x == (char *) 0 && y != (char *) 0)
		return 1;
	if (x != (char *) 0 && y == (char *) 0)
		return -1;
	if (x == (char *) 0 && y == (char *) 0)
		return 0;

	for ( ; *x != '\0' && *y != '\0'; ++x, ++y) {
		if (0 <= len && len-- == 0)
			return 0;

		if (*x != *y) {
			diff = tolower(*x) - tolower(*y);
			if (diff != 0)
				return diff;
		}
	}

	if (len == 0)
		return 0;

	return *x - *y;
}

int
TextInsensitiveCompare(const void *xp, void const *yp)
{
	return TextInsensitiveCompareN(xp, yp, -1);
}

/**
 * <p>
 * Compare the leading part of the string in a case insensitive manner.
 * </p>
 *
 * @param text
 *	The string to check.
 *
 * @param prefix
 *	The string prefix to match.
 *
 * @return
 * 	Return -1 on no match, otherwise the length of the matching prefix.
 */
long
TextInsensitiveStartsWith(const char *text, const char *prefix)
{
	const char *start = text;

	for ( ; *prefix != '\0'; ++text, ++prefix) {
		if (tolower(*text) != tolower(*prefix))
			return -1;
	}

	return (long) (text - start);
}

/**
 * <p>
 * Compare the trailing part of the string in a case sensitive manner.
 * </p>
 *
 * @param text
 *	The string to check.
 *
 * @param suffix
 *	The string suffix to match.
 *
 * @return
 * 	Return -1 on no match, otherwise the length of the matching suffix.
 */
long
TextInsensitiveEndsWith(const char *text, const char *suffix)
{
	long length = strlen(suffix);
	const char *etext = text + strlen(text);
	const char *esuffix = suffix + length;

	while (suffix < esuffix) {
		if (etext <= text)
			return -1;

		--etext;
		--esuffix;

		if (tolower(*etext) != tolower(*esuffix))
			return -1;
	}

	return length;
}

int
TextSensitiveCompareN(const void *xp, void const *yp, long len)
{
	char *x = (char *) xp;
	char *y = (char *) yp;

	if (x == (char *) 0 && y != (char *) 0)
		return 1;
	if (x != (char *) 0 && y == (char *) 0)
		return -1;
	if (x == (char *) 0 && y == (char *) 0)
		return 0;

	if (len < 0)
		return strcmp(x, y);

	return strncmp(x, y, len);
}

int
TextSensitiveCompare(const void *xp, void const *yp)
{
	return TextSensitiveCompareN(xp, yp, -1);
}

/**
 * <p>
 * Compare the leading part of the string in a case sensitive manner.
 * </p>
 *
 * @param text
 *	The string to check.
 *
 * @param prefix
 *	The string prefix to match.
 *
 * @return
 * 	Return -1 on no match, otherwise the length of the matching prefix.
 */
long
TextSensitiveStartsWith(const char *text, const char *prefix)
{
	const char *start = text;

	for ( ; *prefix != '\0'; ++text, ++prefix) {
		if (*text != *prefix)
			return -1;
	}

	return (long) (text - start);
}

/**
 * <p>
 * Compare the trailing part of the string in a case sensitive manner.
 * </p>
 *
 * @param text
 *	The string to check.
 *
 * @param suffix
 *	The string suffix to match.
 *
 * @return
 * 	Return -1 on no match, otherwise the length of the matching suffix.
 */
long
TextSensitiveEndsWith(const char *text, const char *suffix)
{
	long length = strlen(suffix);
	const char *etext = text + strlen(text);
	const char *esuffix = suffix + length;

	while (suffix < esuffix) {
		if (etext <= text)
			return -1;
		if (*--etext != *--esuffix)
			return -1;
	}

	return length;
}

/**
 * <p>
 * Find the first occurence of a substring.
 * </p>
 *
 * @param text
 *	The string to search.
 *
 * @param sub
 *	The substring to find.
 *
 * @return
 * 	Return -1 on no match, otherwise the offset of the substring.
 */
long
TextSensitiveFind(const char *text, const char *sub)
{
	const char *t, *stop;
	long textLength, subLength;

	/* Impossible to match undefined strings. */
	if (text == (char *) 0 || sub == (char *) 0)
		return -1;

	subLength = strlen(sub);
	textLength = strlen(text);
	
	/* Empty string matches end of string. */
	if (subLength == 0)
		return textLength;
	
	stop = &text[textLength - subLength + 1];
	
	for (t = text; t < stop; t++) {
		if (TextSensitiveCompareN(t, sub, subLength) == 0)
			return (long) (t - text);
	}
	
	/* No match found. */
	return -1;
}

/**
 * <p>
 * Find the first occurence of a case-insensitive substring.
 * </p>
 *
 * @param text
 *	The string to search.
 *
 * @param sub
 *	The substring to find.
 *
 * @return
 * 	Return -1 on no match, otherwise the offset of the substring.
 */
long
TextInsensitiveFind(const char *text, const char *sub)
{
	const char *t, *stop;
	long textLength, subLength;

	/* Impossible to match undefined strings. */
	if (text == (char *) 0 || sub == (char *) 0)
		return -1;

	subLength = strlen(sub);
	textLength = strlen(text);
	
	/* Empty string matches end of string. */
	if (subLength == 0)
		return textLength;
	
	stop = &text[textLength - subLength + 1];
	
	for (t = text; t < stop; t++) {
		if (TextInsensitiveCompareN(t, sub, subLength) == 0)
			return (long) (t - text);
	}
	
	/* No match found. */
	return -1;
}

/**
 * <p>
 * Read a line of input until a newline (CRLF or LF) is read or the buffer
 * is filled. The newline is removed from the input and the buffer is always
 * null terminated.
 * </p>
 *
 * @param fp
 *	The FILE * from which to read input.
 *
 * @param line
 *	The input buffer.
 *
 * @param size
 *	The size of the input buffer.
 *
 * @return
 * 	Return the length of the input buffer. otherwise -1 on error.
 */
long
TextInputLine(FILE *fp, char *line, long size)
{
	long i;

	for (i = 0, --size; i < size; ++i) {
		line[i] = (char) fgetc(fp);

		if (feof(fp) || ferror(fp))
			return -1;

		if (line[i] == '\n') {
			line[i] = '\0';
			if (0 < i && line[i-1] == '\r')
				line[--i] = '\0';
			break;
		}
	}

	line[i] = '\0';

	return i;
}

/**
 * <p>
 * Read a line of input until a newline (CRLF or LF) is read or the buffer
 * is filled. The newline is removed from the input and the buffer is always
 * null terminated.
 * </p>
 *
 * @param fd
 *	The file handle from which to read input.
 *
 * @param line
 *	The input buffer.
 *
 * @param size
 *	The size of the input buffer.
 *
 * @return
 * 	Return the length of the input buffer. otherwise -1 on error.
 */
long
TextReadLine(int fd, char *line, long size)
{
	long i;

	for (i = 0, --size; i < size; ++i) {
		if (read(fd, &line[i], 1) <= 0)
			return -1;

		if (line[i] == '\n') {
			line[i] = '\0';
			if (0 < i && line[i-1] == '\r')
				line[--i] = '\0';
			break;
		}
	}

	line[i] = '\0';

	return i;
}

int
TextCountOccurences(const char *s1, const char *s2)
{
	const char *t1, *t2;
	int count = 0;

	for ( ; *s1 != '\0'; ++s1) {
		for (t1 = s1, t2 = s2; *t2 != '\0'; ++t1, ++t2) {
			if (*t1 != *t2)
				break;
		}
		count += (*t2 == '\0');
	}

	return count;
}

/***********************************************************************
 *** Testing
 ***********************************************************************/

#ifdef TEST
#include <stdio.h>

#ifdef USE_DEBUG_MALLOC
# include <com/snert/lib/util/DebugMalloc.h>
#endif

void
TestTextSplit(char *string, char *delims, int empty)
{
	int i;
	char *s;
	Vector v;

	printf("string=[%s] delims=[%s] return-empty-tokens=%d\n", string, delims, empty);
	v = TextSplit(string, delims, empty);

	if (v == NULL) {
		printf("vector is null\n");
		return;
	}

	printf("  length=%ld ", VectorLength(v));
	for (i = 0; i < VectorLength(v); i++) {
		s = VectorGet(v, i);
		printf("[%s]", s == (char *) 0 ? "" : s);
	}
	
	VectorDestroy(v);
	printf("\n");
}

void
TestFunction(TextLongCompareFunction func, const char *a, const char *b, long expect)
{
	long result = (*func)(a, b);
	
	printf("a=[%s] b=[%s] expect=%ld, result=%ld %s\n", a, b, expect, result, expect == result ? "OK" : "FAIL");
}

void
TestCompareFunction(TextCompareFunction func, const void *a, const void *b, int expect)
{
	int result;
	
	result = (*func)(a, b);
	if (result < 0)
		result = -1;
	else if (0 < result)
		result = 1;
	else 
		result = 0;
	
	printf("a=[%s] b=[%s] expect=%d, result=%d %s\n", (char *) a, (char *) b, expect, result, expect == result ? "OK" : "FAIL");
}


int
main(int argc, char **argv) 
{
	printf("\n--TextInsensitiveCompare--\n");
	TestCompareFunction(TextInsensitiveCompare, "a", "A", 0);
	TestCompareFunction(TextInsensitiveCompare, "b", "A", 1);
	TestCompareFunction(TextInsensitiveCompare, "a", "B", -1);

	TestCompareFunction(TextInsensitiveCompare, "ab", "AB", 0);
	TestCompareFunction(TextInsensitiveCompare, "ab", "A", 1);
	TestCompareFunction(TextInsensitiveCompare, "a", "AB", -1);

	TestCompareFunction(TextInsensitiveCompare, "ORDB.org", "ordborg", -1);
	TestCompareFunction(TextInsensitiveCompare, "ORDB.org", "ordb.org", 0);
	TestCompareFunction(TextInsensitiveCompare, "bitbucket@ORDB.org", "bitbucket@ordb.org", 0);

	printf("\n--TextInsensitiveEndsWith--\n");
	TestFunction((LongCompareFunction) TextInsensitiveEndsWith, "", "ordb.org", -1);
	TestFunction((LongCompareFunction) TextInsensitiveEndsWith, ".org", "ordb.org", -1);
	TestFunction((LongCompareFunction) TextInsensitiveEndsWith, "ORDB.org", "ordb.org", sizeof("ordb.org")-1);
	TestFunction((LongCompareFunction) TextInsensitiveEndsWith, "bitbucket@ORDB.org", "ordb.org", sizeof("ordb.org")-1);

	printf("\n--TextSplit--\n");

	/* Empty string and empty delimeters. */
	TestTextSplit("", "", 1);		/* length=1 [] */
	TestTextSplit("", "", 0);		/* length=0 */

	/* Empty string. */
	TestTextSplit("", ",", 1);		/* length=1 [] */
	TestTextSplit("", ",", 0);		/* length=0 */
	
	/* Empty delimiters. */
	TestTextSplit("a,b,c", "", 1);		/* length=1 [a,b,c] */
	TestTextSplit("a,b,c", "", 0);		/* length=1 [a,b,c] */

	/* Assorted combinations of empty tokens. */
	TestTextSplit(",", ",", 1);		/* length=2 [][] */
	TestTextSplit(",", ",", 0);		/* length=0 */
	TestTextSplit("a,,", ",", 1);		/* length=3 [a][][] */
	TestTextSplit("a,,", ",", 0);		/* length=1 [a] */
	TestTextSplit(",b,", ",", 1);		/* length=3 [][b][] */
	TestTextSplit(",b,", ",", 0);		/* length=1 [b] */
	TestTextSplit(",,c", ",", 1);		/* length=3 [][][c] */
	TestTextSplit(",,c", ",", 0);		/* length=1 [c] */
	TestTextSplit("a,,c", ",", 1);		/* length=3 [a][][c] */
	TestTextSplit("a,,c", ",", 0);		/* length=2 [a][c] */
	TestTextSplit("a,b,c", ",", 1);		/* length=3 [a][b][c] */
	TestTextSplit("a,b,c", ",", 0);		/* length=3 [a][b][c] */
	
	/* Quoting of tokens. */
	TestTextSplit("a,b\\,c", ",", 1);	/* length=2 [a][b,c] */
	TestTextSplit("a,b\\,c", ",", 0);	/* length=2 [a][b,c] */
	TestTextSplit("a,'b,c'", ",", 1);	/* length=2 [a][b,c] */
	TestTextSplit("a,'b,c'", ",", 0);	/* length=2 [a][b,c] */
	TestTextSplit("\"a,b\",c", ",", 1);	/* length=2 [a,b][c] */
	TestTextSplit("\"a,b\",c", ",", 0);	/* length=2 [a,b][c] */
	TestTextSplit("a,'b,c'd,e", ",", 1);	/* length=3 [a][b,cd][e] */
	TestTextSplit("a,'b,c'd,e", ",", 0);	/* length=3 [a][b,cd][e] */
	TestTextSplit("a,'',e", ",", 1);	/* length=3 [a][][e] */
	TestTextSplit("a,'',e", ",", 0);	/* length=3 [a][][e] */
	TestTextSplit("a,b''d,e", ",", 1);	/* length=3 [a][bd][e] */
	TestTextSplit("a,b''d,e", ",", 0);	/* length=3 [a][bd][e] */

	printf("\n--DONE--\n");

	return 0;
}
#endif
