/*
 * mailout.c
 *
 * Copyright 2002, 2003 by Anthony Howe.  All rights reserved.
 */

#include <com/snert/lib/version.h>

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <com/snert/lib/io/Log.h>
#include <com/snert/lib/io/posix.h>
#include <com/snert/lib/io/Socket.h>
#include <com/snert/lib/mail/MailSpan.h>
#include <com/snert/lib/util/getopt.h>
#include <com/snert/lib/util/Text.h>

#define INPUT_LINE_SIZE		1000
#define HOST_NAME_SIZE		128
#define MAIL_ADDRESS_SIZE	256

#define NON_BLOCKING_WRITE
#define NON_BLOCKING_READ

int smtpPort = 25;
char *smtpHost = "127.0.0.1";
long socketTimeout = 120000;

char usage[] =
"usage: mailout [-h host][-p port][-f[from]] to ... <message\n"
"\n"
"-h host\t\tSMTP host to contact [default localhost]\n"
"-p port\t\tSMTP port to use [default 25]\n"
"-f[from]\tEmail address of sender [default postmaster@localhost]\n"
"to ...\t\tOne or more destination email addresses\n"
"message\t\tThe message to send is read from standard input.\n"
"\n"
"mailout/1.10 Copyright 2002, 2003 by Anthony Howe.  All rights reserved.\n"
;

char line[INPUT_LINE_SIZE];
char hostname[HOST_NAME_SIZE];
char postmaster[MAIL_ADDRESS_SIZE];

int
printline(Socket s, char *line)
{
	int error;
	int rc = 0;

	LogDebug("> %s", line);
	
#if defined(NON_BLOCKING_WRITE)
# if ! defined(NON_BLOCKING_READ)
	/* Do not block on sending to the SMTP server just yet. */
	(void) SocketSetNonBlocking(s, 1);
# endif

	if (SocketPrint(s, line) < 0) {
		error = SocketGetError(s);
		LogError("! SocketPrint() error: %s (%d)", strerror(error), error);
		goto error0;
	}

	/* Now wait for the output to be sent to the SMTP server. */
	if (SocketWaitForOutput(s, socketTimeout) != 0) {
		LogError("! timeout before output sent to SMTP server");
		goto error0;
	}
#else
	if (SocketPrint(s, line) < 0) {
		error = SocketGetError(s);
		LogError("! SocketPrint() error: %s (%d)", strerror(error), error);
		goto error0;
	}
#endif
	rc = 1;
error0:
#if ! defined(NON_BLOCKING_READ) && defined(NON_BLOCKING_WRITE)
	(void) SocketSetNonBlocking(s, 0);
#endif
	return rc;
}

int
printlines(Socket s, char **lines)
{
        for ( ; *lines != (char *) 0; lines++) {
        	if (!printline(s, *lines))
        		return 0;
        }
        
        return 1;
}      

int
getSmtpResponse(Socket s, int code, char *line, int linesize)
{
	int error;
	char *stop;
	long length, value;

	if (linesize < sizeof (long))
		return 0;
	
#if defined(NON_BLOCKING_READ) && ! defined(NON_BLOCKING_WRITE)
	/* Do not block on reading from the SMTP server just yet. */
	(void) SocketSetNonBlocking(s, 1);
#endif

	/* Ideally we should collect _all_ the response lines into a variable
	 * length buffer (see com/snert/lib/util/Buf.h), but I don't see any
	 * real need for it just yet.
	 */

	do {
		/* Reset this return code for each input line of a
		 * multiline response in case there is a read error.
		 */
		value = 450;
		stop = line;

		/* Erase the first 4 bytes of the line buffer, which
		 * corresponds with the 3 ASCII digit status code
		 * followed by either a ASCII hyphen or space.
		 */
		line[0] = line[1] = line[2] = line[4] = '\0';

		error = SocketWaitForInput(s, socketTimeout);

		if (error == ETIMEDOUT) {
			LogError("! timeout before input from SMTP server");
			goto error0;
		}

		if (error != 0) {
			LogError("! input error from SMTP server: %s (%d)", strerror(error), error);
			goto error0;
		}

		length = SocketReadLine(s, line, linesize);

		if ((error = SocketGetError(s)) != 0 || length == -1) {
			LogError("! read error from SMTP server: %s (%d)", strerror(error), error);
			goto error0;
		}

		if (length == -2) {
			LogError("! unexpected EOF from SMTP server");
			goto error0;
		}

		/* Did we read sufficient characters for a response code? */
		if (length < 4) {
			LogError("! truncated response from SMTP server, length=%ld", length);
			goto error0;
		}

		LogDebug("< %s", line);

		if ((value = strtol(line, &stop, 10)) == 421) {
			break;
		}
	} while (line[3] == '-');
error0:
#if defined(NON_BLOCKING_READ) && ! defined(NON_BLOCKING_WRITE)
	(void) SocketSetNonBlocking(s, 0);
#endif

	return value == code && line + 3 == stop;
}

int
main(int argc, char **argv)
{
	int length;
	Socket client;
	char *from = (char *) 0;

#if defined(__BORLANDC__)
	_fmode = O_BINARY;
#endif

	LogSetLevel(LOG_DEBUG);
	LogSetProgramName("mailout");
	if (LogOpen("mailout.log")) {
		fprintf(stderr, "Failed to open \"maillog.log\": %s (%d)\n", strerror(errno), errno);
		return 2;
	}

	while(getopt(argc, argv, "f;h:p:") != -1) {
		switch (optopt) {
		case 'f':
			from = optarg;
			if (from == (char *) 0)
				break;

			length = MailSpanLocalPart(optarg);
			if (from[length] != '@') {
				LogError("-f argument \"%s\" not valid", from);
				LogError(usage);
				return 2;
			}

			optarg = &from[length + 1];
			length = MailSpanDomainName(optarg, 1);
			if (optarg[length] != '\0') {
				LogError("-f argument \"%s\" not valid", from);
				LogError(usage);
				return 2;
			}
			break;
		case 'h':
			smtpHost = optarg;
			break;
		case 'p':
			smtpPort = strtol(optarg, (char **) 0, 10);
			break;
		default:
			fprintf(stderr, usage);
			fprintf(stderr, LIBSNERT_VERSION " " LIBSNERT_COPYRIGHT "\n");
			return 2;
		}
	}

	if (argc < optind + 1) {
		fprintf(stderr, usage);
		fprintf(stderr, LIBSNERT_VERSION " " LIBSNERT_COPYRIGHT "\n");
		return 2;
	}

	if (from == (char *) 0) {
		/* When no from address supplied, default to postmaster. */
		length = MailSpanIPv4(smtpHost);
		sprintf(
			postmaster,
			smtpHost[length] == '\0'
				? "postmaster@[%s]"
				: "postmaster@%s",
			smtpHost
		);
		from = postmaster;
	}

	if (SocketInit()) {
		LogError("SocketInit()");
		return 1;
	}

	LogInfo("connecting to host=%s port=%d", smtpHost, smtpPort);

	client = SocketOpenTcpClient(smtpHost, smtpPort);
	if (client == (Socket) 0) {
		LogError("SocketOpenTcpClient()");
		return 1;
	}

/*	SocketSetTcpNoDelay(client, 1); */

#if defined(NON_BLOCKING_READ) && defined(NON_BLOCKING_WRITE)
	/* Set non-blocking I/O once for both getSmtpResponse() and
	 * printline() and leave it that way. 
	 */
	if (SocketSetNonBlocking(client, 1)) {
		LogError("! internal error: SocketSetNonBlocking() failed: %s (%d)", strerror(errno), errno);
		return 1;
	}
#endif

	getSmtpResponse(client, 220, line, INPUT_LINE_SIZE);

	(void) gethostname(hostname, HOST_NAME_SIZE);
	sprintf(line, "HELO %s\r\n", hostname);
	printline(client, line);
	getSmtpResponse(client, 250, line, INPUT_LINE_SIZE);

	LogInfo("from=<%s>", from);
	sprintf(line, "MAIL FROM:<%s>\r\n", from);
	printline(client, line);
	getSmtpResponse(client, 250, line, INPUT_LINE_SIZE);

	for ( ; optind < argc; ++optind) {
		LogInfo("to=<%s>", argv[optind]);
		sprintf(line, "RCPT TO:<%s>\r\n", argv[optind]);
		printline(client, line);
		getSmtpResponse(client, 250, line, INPUT_LINE_SIZE);
	}

	printline(client, "DATA\r\n");
	getSmtpResponse(client, 354, line, INPUT_LINE_SIZE);

	while (0 <= (length = TextReadLine(FILENO_STDIN, line, INPUT_LINE_SIZE))) {
		if (line[0] == '.') {
			if (line[1] == '\0')
				break;
			SocketPrint(client, ".");
		}

		line[length++] = '\r';
		line[length++] = '\n';
		line[length] = '\0';

		printline(client, line);
	}

	printline(client, ".\r\n");
	getSmtpResponse(client, 250, line, INPUT_LINE_SIZE);

	printline(client, "QUIT\r\n");
#ifdef DONT_DO_THIS
	(void) SocketReadLine(client, line, INPUT_LINE_SIZE);
#endif
	SocketClose(client);

	LogDebug("done");

	return 0;
}


